/*
	$Id: GroupTransfer.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2007 Jari Häkkinen, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.clients.migrate;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Group;
import net.sf.basedb.core.Quota;
import net.sf.basedb.core.User;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

/**
<dl>
<dt><b>Base 1 tables</b></dt>
<dd>UserGroup</dd>
<dt><b>Unused columns</b></dt>
<dd>owner, addedDate</dd>
<dt><b>Notes</b></dt>
<dd>
<ul>
<li>Users are added to each group.</li>
</ul>
</dd>
</dl>

@author Gregory, Martin
@version 2.0
@base.modified $Date: 2009-04-06 14:52:39 +0200 (Mon, 06 Apr 2009) $
*/

public class GroupTransfer extends Transfer
{
	/**
	Logger for this class
	 */
	private static final org.apache.log4j.Logger log = org.apache.log4j.Logger.getLogger("net.sf.basedb.clients.migrate.GroupTransfer");
	
	/**
	Maps base 1 ids to Base 2 Group objects
	 */
	private UserTransfer userT = null;
	private QuotaTransfer quotaT = null;

	GroupTransfer()
	{}
	
	/**
	@param userT Depends on a finnished UserTransfer.
	 */
	GroupTransfer(UserTransfer userT, QuotaTransfer quotaT)
	{
		log.debug("Checking that necessary transfers have been done before proceeding with Groups");
		if (userT.isDone() && quotaT.isDone())
		{
			this.userT = userT;
			this.quotaT = quotaT;
		}
		else
		{
			if (!userT.isDone()) log.debug("Dependency: Users haven't been transfered");
			if (!quotaT.isDone()) log.debug("Dependency: Quotas haven't been transfered");
			dependencyFailure();
		}
	}

	/**
	Starts a Group transfer. Stops the application if users have not yet been
	transfered.
	 */
	public void start()
	{
		if (userT == null)
		{
			log.error("Can't continue with transferring groups, the users need to be transfered before groups");
			stop();
		}
		else
		{
			log.info("Transferring Groups...");
			int count = getInt("countGroups");
			initIdMap(count);
			startProgress(count, "Groups");
			run(prepareStatementFromFile("selectGroups"));
			done();
			log.info("Transferring Groups...DONE");
		}
	}

	/**
	@param dc {@link DbControl}
	@param rs {@link ResultSet}
	@return A {@link Group} created from the current row in the ResultSet
	 */
	Group createItem(DbControl dc, ResultSet rs) throws BaseException,
			SQLException
	{
		Group group = Group.getNew(dc);
		/*
	We save the base 1 group id in the name so that we later can map it
	to a BASE 2 group id.
		 */
		group.setName(rs.getString(2));
		group.setDescription(rs.getString(3));
		group.setRemoved(rs.getBoolean(4));
		long byteQuota = rs.getLong(6);
		if (rs.wasNull()) byteQuota = Quota.UNLIMITED;
		group.setQuota(quotaT.getQuotaForTotal(dc, byteQuota));		
		if (log.isDebugEnabled()) 
		{
			log.debug("Transfered  group: " + group.getName() + "\tBASE1 id("+ rs.getInt(1) + ")");
		}
		// Add users to this group
		PreparedStatement usersSql = prepareStatementFromFile("selectUsersByGroup");
		int groupId = rs.getInt(1);
		usersSql.setInt(1, groupId);
		ResultSet usersRs = usersSql.executeQuery();
		log.info("adding users to group");
		while (usersRs.next())
		{
			User user = userT.getUserById(dc, usersRs.getInt(1));
			if (user == null)
			{
				log.warn("User with base 1 id " + usersRs.getInt(1)
						+ " is missing from the User Transfer");
			}
			else
			{
				group.addUser(user);
				if (log.isDebugEnabled())
				{
					log.debug("User " + user.getLogin() + " added to group " + group.getName());
				}
			}
		}
		return group;
	}

	public Group getGroupById(DbControl dbc, int base1Id)
	{
		Integer id = getBase2Id(base1Id);
		return id == null ? null : Group.getById(dbc, id);
	}
}
