/*
	$Id: SampleTissueTransfer.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Gregory Vincic
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2007 Jari Häkkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.clients.migrate;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.HashMap;

/**
This class provides methods for transferring SampleTissue data from base 1 to
BASE 2. SampleTissue's are not available in BASE 2 as objects. The
information is stored as part of the description for MeasuredBioMaterials.

@author Gregory Vincic, Martin Svensson
@version 2.0
@base.modified $Date: 2009-04-06 14:52:39 +0200 (Mon, 06 Apr 2009) $
*/

public class SampleTissueTransfer extends Transfer
{
	/**
	Logger for this class
	 */
	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.Logger.getLogger("net.sf.basedb.clients.migrate.SampleTissueTransfer");
	
	private HashMap<Integer, String> tissue = null;

	SampleTissueTransfer()
	{
		tissue = new HashMap<Integer, String>();
	}

	public void start()
	{
		log.info("Transferring SampleTissue...");
		int count = getInt("countSampleTissues");
		initIdMap(count);
		startProgress(count, "SampleTissues");
		run(prepareStatementFromFile("selectSampleTissues"), 0, getBatchSize());
		done();
		log.info("Transferring SampleTissue...DONE");
	}

	/**
	Recursive run over a SQL query defined by a prepared statement and LIMIT
	values from and size. If this method is called directly from a child
	transfer, remember to call {@link #startProgress(int, String)} before you call
	this method
	
	@param ps <code>PreparedStatement</code> which includes a "LIMIT ?,?"
	       at the end.
	@param from <code>int</code> indicating start position of the
	       ResultSet.
	@param size <code>int</code> indicating number of rows to include.
	@see Transfer#prepareStatementFromFile(String)
	 */
	protected void run(PreparedStatement ps, int from, int size)
	{
		try
		{
			int parameters = ps.getParameterMetaData().getParameterCount();
			// Set LIMIT ?, ? which should be the last two parameters in the
			// statement
			ps.setInt(parameters - 1, from);
			ps.setInt(parameters, size);
			ResultSet rs = ps.executeQuery();
			DbControl dc = newDbControl();
			int rowCount = 0;
			while (rs.next())
			{
				rowCount++;
				createItem(dc, rs);
				progress.increase();
			}
			dc.close();
			rs.close();
			if (rowCount > 0 && rowCount == size)
			{
				// Fetch next batch
				run(ps, from + size, size);
			}
			else
			{
				progress.stop();
			}
		}
		catch (SQLException e)
		{
			log.error("FAILED when executing SQL statement", e);
			stop();
		}
		catch (BaseException e)
		{
			log.error("FAILED when transferring SampleTissues", e);
			stop();
		}
	}

	BasicItem createItem(DbControl b2c, ResultSet rs) throws BaseException,
			SQLException
	{
		String tissueString = rs.getString(3);
		if (rs.getInt(2) != 0)
		{
			tissueString = tissue.get(new Integer(rs.getInt(2))) + "->"
					+ tissueString;
		}
		tissue.put(new Integer(rs.getInt(1)), tissueString);
		log.debug("Transfered SampleTissue: " + tissueString);
		return null;
	}

	protected String getTissue(int tissueId)
	{
		return tissue.get(new Integer(tissueId));
	}
}
