/*
	$Id: ExtensionsFile.java 4512 2008-09-11 20:05:50Z jari $

	Copyright (C) Nicklas Nordborg
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.extensions;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipInputStream;

import javax.servlet.ServletContext;

import net.sf.basedb.core.plugin.About;
import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.JarClassLoader;
import net.sf.basedb.util.extensions.Extension;
import net.sf.basedb.util.extensions.ExtensionPoint;
import net.sf.basedb.util.extensions.Registry;
import net.sf.basedb.util.extensions.xml.XmlLoader;

/**
	Represents a single extensions XML file or a JAR file containing 
	extensions. Each file has an individual {@link XmlLoader} to
	keep track of the extensions/extension points loaded from the file.

	@author nicklas
	@version 2.7
	@base.modified $Date:2008-03-20 12:15:25 +0100 (Thu, 20 Mar 2008) $
*/
public class ExtensionsFile
	implements Comparable<ExtensionsFile>
{
	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.clients.web.extensions.ExtensionsFile");

	private final ExtensionsDirectory directory;
	private final File file;
	private final File resourcesDir;
	private final XmlLoader loader;
	private final boolean isJar;
	
	private JarClassLoader jarLoader;
	private boolean hasValidated;
	private boolean isValid;
	private Throwable validationError;
	
	private long lastModified;
	private long size;
	private About about;
	private boolean hasError;
	
	private Map<String, ServletWrapper> servlets;
	
	/**
		Create a new object.
		@param directory The directory with extensions;
		@param file The XML or JAR file containting the extension definition
		@param resourcesDir The directory where resources are extracted
		@param loader The XML loader to use
	*/
	public ExtensionsFile(ExtensionsDirectory directory, File file, 
		File resourcesDir, XmlLoader loader)
	{
		this.directory = directory;
		this.file = file;
		this.resourcesDir = resourcesDir;
		this.loader = loader;
		this.isJar = file.getName().endsWith(".jar");
	}
	
	/*
		From the Comparable interface
		-----------------------------
	*/
	@Override
	public int compareTo(ExtensionsFile other)
	{
		return this.file.compareTo(other.file);
	}
	// ------------------------------
	
	/**
		Get the name of the file.
	*/
	public String getName()
	{
		return file.getName();
	}

	/**
		If the file name ends with '.jar' the file is considered a JAR file.
	 	Otherwise it is considered an XML file. The file name check is case
	 	sensitive.
		@return TRUE if the file is a JAR file, false if it is an XML file
	*/
	public boolean isJar()
	{
		return isJar;
	}
	
	/**
		Check if the file really exists on the disk.
		@return See {@link File#exists()}
	*/
	public boolean exists()
	{
		return file.exists();
	}
	
	/**
		Check if the file has been modified since the last call to
		{@link #resetModified()}. We consider the file to be modified if
		the timestamp ({@link File#lastModified()}) or size ({@link File#length()})
		has changed, or if the underlying class loader has detected a change
		in any files it depends on (see {@link JarClassLoader#hasChanged(boolean)}).
		@return TRUE if the file has been modified, FALSE otherwise
		@see #resetModified()
	*/
	public boolean isModified()
	{
		return lastModified != file.lastModified() || size != file.length() ||
			(jarLoader != null && jarLoader.hasChanged(true)) || 
			(isJar && jarLoader == null);
	}

	/**
		Resets the remembered timestamp and size attributes of the file.
		@see #isModified()
	*/
	void resetModified()
	{
		lastModified = file.lastModified();
		size = file.length();
	}
	
	/**
		If the file is a new file not previously registered with the
		extension system. Thew 'new' status is changed when {@link #resetModified()}
		is called which usually happens when all extensions has been registered.
		@return TRUE if the file is a new file, FALSE otherwise
	*/
	public boolean isNew()
	{
		return lastModified == 0;
	}
	
	/**
		Checks if the file is a valid extensions definition file. This means that
		the file must be either:
		<ul>
		<li>An extensions definition XML file 
		<li>A JAR file with an extensions definition XML file at
			<code>META-INF/extensions.jar</code>.
		</ul>
		
		@return TRUE if the file is valid, FALSE otherwise
		@see #getValidationError()
	*/
	public boolean isValid()
	{
		if (!hasValidated || isModified())
		{
			try
			{
				validate();
			}
			catch (Exception ex)
			{}
		}
		return isValid;
	}
	
	/**
		Get more information about the error that caused the validation
		to fail.
		@return An exception or null if the validation succeeded
	*/
	public Throwable getValidationError()
	{
		return validationError;
	}
	
	/**
		If there was an error when registering the extensions in this file.
		This property is only set if the file has been determined to be a 
		valid extensions file by the {@link #isValid()} method. Extensions that
		has an error are automatically disabled.
		
		@return TRUE if there was some error, FALSE if everything is ok
	*/
	public boolean hasError()
	{
		return hasError;
	}
	
	/**
		Sets the error status.
	*/
	void setError(boolean error)
	{
		this.hasError = error;
	}
	
	/**
		Get an iterator returning all extension points that
		are defined by this file. The returned iterator doesn't support 
		removal and will not reflect changes made by the same or other 
		threads after this call has returned.
	*/
	public Iterator<ExtensionPoint<?>> getExtensionPoints()
	{
		// Create a copy, or the iterator will fail if new
		// files are added by another thread
		List<ExtensionPoint<?>> copy = 
			new ArrayList<ExtensionPoint<?>>(loader.getExtensionPoints());
		return Collections.unmodifiableList(copy).iterator();	
	}

	/**
		Get an iterator returning all extensions that
		are defined by this file. The returned iterator doesn't support 
		removal and will not reflect changes made by the same or other 
		threads after this call has returned.
	*/
	public Iterator<Extension<?>> getExtensions()
	{
		// Create a copy, or the iterator will fail if new
		// files are added by another thread
		List<Extension<?>> copy = 
			new ArrayList<Extension<?>>(loader.getExtensions());
		return Collections.unmodifiableList(copy).iterator();	
	}
	
	/**
		Get information about this extensions file.
		@return An About object or null if no information is available
	*/
	public About getAbout()
	{
		return about;
	}
	
	/**
		Get a XML-like string representation of the parameters that was
		used to initialise a factory. The factory must belong to an
		extension or extension point defined by this file.
		
		@param factory The factory object
		@return The parameters as found in the XML file
	*/
	public String getFactoryParameters(Object factory)
	{
		return loader.getFactoryParameters(factory);
	}
	
	/**
		Load the extension definitions from the XML or JAR file.
	*/
	void loadExtensions()
	{
		if (!loader.hasValidFile()) validate();
		if (isJar)
		{
			about = loadJar();
		}
		else
		{
			about = loadXml();
		}
	}
	
	private void validate()
	{
		hasValidated = true;
		isValid = false;
		validationError = null;
		try
		{
			if (isJar)
			{
				validateJar();
			}
			else
			{
				validateXml();
			}
		}
		catch (RuntimeException ex)
		{
			validationError = ex.getCause();
			throw ex;
		}
		catch (Throwable t)
		{
			validationError = t;
			throw new RuntimeException(t);
		}
		isValid = true;
	}
	
	/**
		Try to validate an XML file with extensions. The XML file is valid
		if it matches the extension definition XML file format.
		@return The global about information or null if not present
	*/
	private About validateXml()
	{
		log.info("Validating extensions in XML file: " + file.getName());
		About about = null;
		try
		{
			about = loader.validateXmlFile(new FileInputStream(file), file.getName());
		}
		catch (Exception ex)
		{
			log.error("Error validating extensions in XML file: " + file, ex);
			throw new RuntimeException(file + " is not a valid extensions file", ex);
		}
		log.info("Extensions in XML file are valid: " + file.getName());
		return about;
	}
	
	/**
		Try to validate a JAR file with extensions. The JAR file is valid
		if it contains a file 'META-INF/extensions.xml' which is a valid
		extension definition XML file.
		@return The global about information or null if not present
	*/
	private About validateJar()
	{
		log.info("Validating extensions in JAR file: " + file.getName());
		InputStream in = null;
		ZipFile zipFile = null;
		About about = null;
		try
		{
			zipFile = new ZipFile(file);
			ZipEntry zipEntry = zipFile.getEntry("META-INF/extensions.xml");
			if (zipEntry == null)
			{
				throw new FileNotFoundException("META-INF/extensions.xml");
			}
			in = zipFile.getInputStream(zipEntry);
			about = loader.validateXmlFile(in, file.getName());
		}
		catch (Exception ex)
		{
			log.error("Error validating extensions in JAR file: " + file, ex);
			throw new RuntimeException(file + " is not a valid extensions file", ex);
		}
		finally
		{
			try
			{
				in.close();
				zipFile.close();
			}
			catch (IOException ex)
			{}
		}
		log.info("Extensions in JAR file are valid: " + file.getName());
		return about;
	}
	
	/**
		Load extensions from a validated XML file.
		@return The global about information or null if not present
	*/
	private About loadXml()
	{
		log.info("Loading extensions from XML file: " + file.getName());
		About about = null;
		try
		{
			about = loader.loadLastValidatedFile(null, true);
		}
		catch (Exception ex)
		{
			log.error("Error loading extensions from XML file: " + file, ex);
			throw new RuntimeException(ex);
		}
		log.info("Extensions loaded from XML file: " + file.getName());
		return about;
	}
	
	/**
		Load extensions from a validated JAR file.
		@return The global about information or null if not present
	*/
	private About loadJar()
	{
		log.info("Loading extensions from JAR file: " + file.getName());
		About about = null;
		try
		{
			jarLoader = (JarClassLoader)JarClassLoader.getInstance(file.getAbsolutePath(), true);
			about = loader.loadLastValidatedFile(jarLoader, true);
		}
		catch (Exception ex)
		{
			log.error("Error loading extensions from JAR file: " + file, ex);
			throw new RuntimeException(ex);
		}
		log.info("Extensions loaded from JAR file: " + file.getName());
		return about;
	}
	
	
		
	/**
		Load servlet definitions from a JAR file.
	*/
	int loadServlets(ServletContext context)
	{
		if (!isJar) return 0;
		log.info("Loading servlets from JAR file: " + file.getName());
		InputStream in = null;
		ZipFile zipFile = null;
		int numServlets = 0;
		try
		{
			zipFile = new ZipFile(file);
			ZipEntry zipEntry = zipFile.getEntry("META-INF/servlets.xml");
			if (zipEntry == null)
			{
				log.info("META-INF/servlets.xml not found in file: " + file.getName());
				return 0;
			}
			
			if (servlets == null) servlets = new HashMap<String, ServletWrapper>();
			servlets.clear();
			in = zipFile.getInputStream(zipEntry);
			
			ClassLoader jarLoader = JarClassLoader.getInstance(file.getAbsolutePath(), true);
			
			ServletLoader servletLoader = new ServletLoader();
			servletLoader.loadXmlFile(in, file.getName(), jarLoader, true);
			for (ServletWrapper wrapper : servletLoader.getServlets())
			{
				wrapper.setServletContext(context);
				servlets.put(wrapper.getServletName(), wrapper);
				numServlets++;
			}
		}
		catch (Exception ex)
		{
			log.error("Error loading servlets from JAR file: " + file, ex);
			throw new RuntimeException("Error loading servlets from '" + file + "': " + 
				ex.getMessage(), ex);
		}
		finally
		{
			try
			{
				if (in != null) in.close();
				if (zipFile != null) zipFile.close();
			}
			catch (IOException ex)
			{}
		}
		log.info(numServlets + " servlets loaded from JAR file: " + file.getName());
		return numServlets;
	}
	
	
	/**
		Extract resources from a JAR file to a directory. This method does nothing
		if the file is an XML file. 

		@param filter A regular expression that must match all files
			that should be extracted, or null to every file
		@param replacement A replacement pattern used to convert the filename
			in the JAR file to the filename on disk, or null to not change
			the filename. See {@link Matcher#replaceAll(String)}.
		@param forceOverwrite If TRUE, overwrite files even if they have the same
			size and timestamp as the file in the JAR file
		@return The number of extracted files
	*/
	int extractResources(Pattern filter, String replacement, boolean forceOverwrite)
	{
		if (!isJar) return 0;
		
		log.info("Extracting resources from " + file.getName() + 
				" to " + resourcesDir.getAbsolutePath() + "; forceOverwrite=" + forceOverwrite);
		log.debug("filter=" + filter + "; replacement=" + replacement);
		
		int numExtracted = 0;
		try
		{
			ZipInputStream zipStream = new ZipInputStream(new FileInputStream(file));
			ZipEntry zipEntry = null;
			while ((zipEntry = zipStream.getNextEntry()) != null)
			{
				String name = zipEntry.getName();
				// Ignore directories
				if (zipEntry.isDirectory()) 
				{
					log.debug("File '" + name + "' is a directory. Ignored.");
					continue;
				}

				// Get the name of the zipped file, but strip first slash
				if (name.startsWith("/")) name = name.substring(1);
				log.debug("File '" + name + "' is a file. Continuing.");
				
				// Check with the filter if the file should be extracted
				boolean extract = false;
				String extractionName = name;
				if (filter == null)
				{
					extract = true;
				}
				else
				{
					Matcher matcher = filter.matcher(name);
					extract = matcher.matches();
					// Get the new filename
					if (extract && replacement != null)
					{
						extractionName = matcher.replaceAll(replacement);
					}
				}
				
				// Now, extract the file
				if (!extract)
				{
					log.debug("File '" + name + "' doesn't match filter. Skipped.");
				}
				else
				{
					long time = zipEntry.getTime();
					File extractTo = new File(resourcesDir, extractionName);
					
					boolean modified = extractTo.lastModified() != time || 
						extractTo.length() != zipEntry.getSize();
					if (modified || forceOverwrite)
					{
						log.info("Extracting '" + name + "' to " + extractTo.getAbsolutePath());
						extractTo.getParentFile().mkdirs();
						OutputStream out = new FileOutputStream(extractTo);
						FileUtil.copy(zipStream, out);
						out.close();
						extractTo.setLastModified(zipEntry.getTime());
						numExtracted++;
					}
					else
					{
						log.debug("File '" + name + "' has same time and size. Ignored.");
					}
				}
				
			}
		}
		catch (Exception ex)
		{
			throw new RuntimeException(ex);
		}
		return numExtracted;
	}
	
	/**
		Remove all extracted resources. This method simply removes all
		files and directories it can find in the home directory of the
		extensions. Files from two different extension packages should
		thus never be mixed in the same directory.
		@return The number of removed files
	*/
	int removeResources()
	{
		log.info("Removing resource files from " + resourcesDir.getAbsolutePath());
		int numRemoved = 0;
		
		List<File> directories = new ArrayList<File>();
		directories.add(resourcesDir);

		while (directories.size() > 0)
		{
			int index = directories.size() - 1;
			File lastDir = directories.get(index);
			File[] all = lastDir.listFiles();
			log.debug("Entering directory " + lastDir.getAbsolutePath());
			boolean hasSubDirectories = false;
			if (all != null)
			{
				for (File f : all)
				{
					if (f.isDirectory())
					{
						directories.add(f);
						hasSubDirectories = true;
					}
					else
					{
						log.debug("Removing file " + f.getName());
						f.delete();
						numRemoved++;
					}
				}
			}
			if (!hasSubDirectories) 
			{
				log.debug("Removing directory " + lastDir.getAbsolutePath());
				lastDir.delete();
				directories.remove(index);
			}
		}
		log.info("Removed " + numRemoved + " resource files from " + resourcesDir.getAbsolutePath());
		return numRemoved;
	}
	
	/**
		Register all extensions with a registry.
		@param registry The registry
		@param forceUpdate TRUE to force an update of already registered
			extensions, FALSE to only update if the extension package has 
			been modified
		@return The number of registered or updated extensions
	*/
	int registerExtensions(Registry registry, boolean forceUpdate)
	{
		for (Extension ext : loader.getExtensions())
		{
			directory.addRegistered(ext.getId(), this);
		}
		return loader.registerExtensions(registry, forceUpdate || isModified());
	}

	/**
		Register all extension points with a registry.
		@param registry The registry
		@param forceUpdate TRUE to force an update of already registered
			extension points, FALSE to only update if the extension package
			has been modified
		@return The number of registered or updated extensions
	*/
	int registerExtensionPoints(Registry registry, boolean forceUpdate)
	{
		for (ExtensionPoint ep : loader.getExtensionPoints())
		{
			directory.addRegistered(ep.getId(), this);
		}
		return loader.registerExtensionPoints(registry, forceUpdate || isModified());
	}
	
	/**
		Unregister extensions and extension points that have been removed
		from this package after an update.
	
		@param registry The registry
		@return The number of unregistered extensions and extension points
	*/
	int unregisterMissing(Registry registry)
	{
		// TODO - implement
		return 0;
	}
	
	/**
		Unregister all extensions and extension points.
	
		@param registry The registry
		@return The number of unregistered extensions and extension points
	*/
	int unregisterAll(Registry registry)
	{
		for (Extension ext : loader.getExtensions())
		{
			directory.removeRegistered(ext.getId(), this);
		}
		for (ExtensionPoint ep : loader.getExtensionPoints())
		{
			directory.removeRegistered(ep.getId(), this);
		}
		int numExt = loader.unregisterExtensionPoints(registry);
		numExt += loader.unregisterExtensions(registry);
		return numExt;
	}
	
	/**
		Get a servlet that has been defined in this JAR file.
		@param servletName The name of the servlet, specified by the
			&lt;servlet-name&gt; tag in the XML file.
		@return A wrapper that can be used to invoke the servlet,
			or null if not servlet with the given name exists
	*/
	public ServletWrapper getServlet(String servletName)
	{
		return servlets == null ? null : servlets.get(servletName);
	}
	
}
