/**
	$Id:JspContext.java 4187 2008-03-20 11:15:25Z nicklas $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.extensions;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import javax.servlet.jsp.JspWriter;
import javax.servlet.jsp.PageContext;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.util.extensions.ActionFactory;
import net.sf.basedb.util.extensions.ClientContext;
import net.sf.basedb.util.extensions.Extension;
import net.sf.basedb.util.extensions.InvokationContext;
import net.sf.basedb.util.extensions.RendererFactory;

/**
	Context object for the web application. Instances of this class
	are always used for passing client context information to web extensions.
	This means that it is always safe to cast the client context to this
	class in the {@link ActionFactory#prepareContext(InvokationContext)} 
	and {@link ActionFactory#getActions(InvokationContext)} methods
	and other methods that has a context parameter.
	
	<pre class="code">
&#64;Override
public boolean prepareContext(InvokationContext context) 
{
	JspContext jspContext = (JspContext)context.getClientContext();
	jspContext.addStylesheet("/base/include/styles/mystylesheet.css");
	// Do other things
	return true;
}
</pre>

	<p>
	This context exposes the underlying JSP {@link PageContext},
	which can be used to get information about the current request
	among other things. Render factories usually needs access to
	the response output stream, {@link #getOut()}, so that renderers 
	can write the generated HTML to the response.

	@author nicklas
	@version 2.7
	@base.modified $Date:2008-03-20 12:15:25 +0100 (Thu, 20 Mar 2008) $
*/
public class JspContext
	extends ClientContext
{
	
	private final ExtensionsDirectory directory;
	private final PageContext pageContext;
	private final GuiContext guiContext;
	
	private Set<String> scripts;
	private Set<String> stylesheets;
	
	JspContext(ExtensionsDirectory directory, SessionControl sc, DbControl dc, 
		PageContext pageContext, GuiContext guiContext, Object item)
	{
		super(sc, dc, item);
		this.directory = directory;
		this.pageContext = pageContext;
		this.guiContext = guiContext;
	}

	/**
		Get the JSP Page context object for the current request.
	*/
	public PageContext getPageContext()
	{
		return pageContext;
	}

	/**
		Get a writer object that can be used to return generated text
		to the response that is sent back to the browser.
		<p>
		
		Note! The writer is only intended to be used by render
		objects when rendering the extensions. Using the writer
		from the {@link ActionFactory#prepareContext(InvokationContext)} or 
		{@link RendererFactory#prepareContext(InvokationContext)}
		method may produce unpredictable results.
		
		@return A {@link JspWriter} object
	*/
	public JspWriter getOut()
	{
		return pageContext.getOut();
	}
	
	/**
		Get the current {@link GuiContext}.
		@return A GuiContext object or null if it is not applicable
			to the extension point
		@since 2.12
	*/
	public GuiContext getGuiContext()
	{
		return guiContext;
	}
	
	/**
		Get the URL to the web application root as a string. The URL
		only includes the path information, not the server name or protocol.
	*/
	public String getRoot()
	{
		return pageContext.getServletContext().getContextPath();
	}
	
	/**
		Get the URL to the home directory for a given extension. The URL
		only includes the path information, not the server name or protocol.
		@param extension The extension
	*/
	public String getHome(Extension extension)
	{
		ExtensionsFile file = directory.getFileByExtensionId(extension.getId());
		return file == null ? null : directory.getResourcesUrl(file);
	}
	
	/**
		Add an absolute reference to a script that should be included on the
		generated page with a <code>&lt;script&gt;</code> tag. The 
		path must be an absolute path. 

		<p>
		Note 1! This method must be called from the {@link 
		ActionFactory#prepareContext(InvokationContext)} or 
		{@link RendererFactory#prepareContext(InvokationContext)}
		methods. Changes to the context after that will not be reflected
		in the output.
		
		<p>
		Note 2! Not all extension points supports custom scripts. Check
		the documentation for the extension point you are extending.
		
		@param absolutePath The absolute path
	*/
	public void addScript(String absolutePath)
	{
		if (scripts == null) scripts = new HashSet<String>();
		scripts.add(absolutePath);
	}
	
	/**
		Add an absolute reference to a CSS stylesheet that should be included on the
		generated page with a <code>&lt;link&gt;</code> tag. The 
		path must be the absolute path. 

		<p>
		Note 1! This method must be called from the {@link 
		ActionFactory#prepareContext(InvokationContext)} or 
		{@link RendererFactory#prepareContext(InvokationContext)}
		methods. Changes to the context after that will not be reflected
		in the output.
		
		<p>
		Note 2! Not all extension points supports custom stylesheets. Check
		the documentation for the extension point you are extending.

		@param absolutePath The absolute path
	*/
	public void addStylesheet(String absolutePath)
	{
		if (stylesheets == null) stylesheets = new HashSet<String>();
		stylesheets.add(absolutePath);
	}

	/**
		Get all scripts that has been added to this context.
		@return A collection of scripts
	*/
	public Collection<String> getScripts()
	{
		return scripts;
	}
	
	/**
		Get all stylesheets that has been added to this context.
		@return A collection of stylesheets
	*/
	public Collection<String> getStylesheets()
	{
		return stylesheets;
	}
	
}
