/**
	$Id: PermissionButtonFactory.java 4872 2009-04-02 07:22:17Z nicklas $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.extensions.toolbar;

import net.sf.basedb.clients.web.extensions.AbstractJspActionFactory;
import net.sf.basedb.core.AccessControlled;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.util.extensions.ClientContext;
import net.sf.basedb.util.extensions.InvokationContext;
import net.sf.basedb.util.extensions.xml.PathSetter;
import net.sf.basedb.util.extensions.xml.VariableSetter;

/**
	A button factory that can hide or disable button items based 
	on the logged in user's permissions. It is possible to set
	different text, tooltip, icon, etc. for a disabled button.
	<p>
	
	If the {@link ClientContext} has a current item that is
	an {@link AccessControlled} item permissions are checked against 
	that item:
	{@link AccessControlled#hasPermission(Permission)}. If no item
	type has been specified with {@link #setItem(String)} the factory will
	accept any kind of item, otherwise it will only generate button
	actions if the item type matches.
	<p>
	
	If the <code>ClientContext</code> doesn't have a current item,
	the factory checks the role-based permissions with
	{@link SessionControl#hasPermission(Permission, Item)}.
	
	<p>
	Permission levels can be
	set for when the button should be:
	
	<ul>
	<li>visible and enabled
	<li>visible but disabled
	<li>hidden
	</ul>
	
	If the button is hidden this is indicated by returning
	false from the {@link #prepareContext(InvokationContext)} method. Most
	properties can have different values for the enabled/disabled state.
	For example:
	<ul>
	<li>{@link #setTitle(String)} sets the title for both states
	<li>{@link #setDisabledTitle(String)} sets the title for the disabled state
	<li>{@link #setEnabledTitle(String)} sets the title for the enabled state
	</ul>
	
	<p>
	The default settings will accept all items that the user has read access 
	to and make the button visible and enabled.
	
	<p>
	If the extension point supports scripts and stylesheets, use the
	{@link #setScript(String)} and {@link #setStylesheet(String)} methods.
	Each call to these methods will add the the argument to a set. Use
	the {@link #getScripts()} and {@link #getStylesheets()} methods to gain
	access to the sets, for example, to remove added items.
	
	@author nicklas
	@version 2.7
	@base.modified $Date:2008-03-20 12:15:25 +0100 (Thu, 20 Mar 2008) $
*/
public class PermissionButtonFactory 
	extends AbstractJspActionFactory<ButtonAction>	
{
	private String id;
	private String onClick;

	private Permission visiblePermission;
	private Permission enabledPermission;
	private Item itemToCheck;
	
	private String enabledClazz;
	private String disabledClazz;
	
	private String enabledIcon;
	private String disabledIcon;
	
	private String enabledStyle;
	private String disabledStyle;
	
	private String enabledTitle;
	private String disabledTitle;
	
	private String enabledTooltip;
	private String disabledTooltip;
	
	/**
		Create a new factory instance.
	*/
	public PermissionButtonFactory()
	{
		visiblePermission = Permission.READ;
		enabledPermission = Permission.READ;
	}
	
	/*
		From the ActionFactory interface
		--------------------------------
	*/
	@Override
	public ButtonAction[] getActions(InvokationContext context) 
	{
		ClientContext cc = context.getClientContext();
		ButtonBean action = null;
		if (!hasPermission(cc, itemToCheck, visiblePermission))
		{
			// Button should not be visible
			return null;
		}
		else if (!hasPermission(cc, itemToCheck, enabledPermission))
		{
			// Button should be visible but disabled
			action = new ButtonBean();
			action.setClazz(disabledClazz);
			action.setEnabled(false);
			action.setVisible(true);
			action.setIcon(disabledIcon);
			action.setId(id);
			action.setOnClick(onClick);
			action.setStyle(disabledStyle);
			action.setTitle(disabledTitle);
			action.setTooltip(disabledTooltip);
		}
		else
		{
			// Button is visible and enabled
			action = new ButtonBean();
			action.setClazz(enabledClazz);
			action.setEnabled(true);
			action.setVisible(true);
			action.setIcon(enabledIcon);
			action.setId(id);
			action.setOnClick(onClick);
			action.setStyle(enabledStyle);
			action.setTitle(enabledTitle);
			action.setTooltip(enabledTooltip);
		}
		return new ButtonAction[] { action };
	}
	// ------------------------------------

	/**
		Set the minimum permission that is required to make the
		button enabled.
		@param permission One of {@link Permission} constants
	*/
	public void setEnabledPermission(String permission)
	{
		this.enabledPermission = Permission.valueOf(permission);
	}
	
	/**
		Set the minimum permission that is required to make the
		button visible.
		@param permission One of {@link Permission} constants
	*/
	public void setVisiblePermission(String permission)
	{
		this.visiblePermission = Permission.valueOf(permission);
	}
	
	/**
		Set the item type to check the permission for.
		@param item One of {@link Item} constants or null to
			accept all items
	*/
	public void setItem(String item)
	{
		this.itemToCheck = item == null ? null : Item.valueOf(item);
	}

	/**
		Set the class to use when the button is enabled.
	*/
	public void setEnabledClazz(String clazz) 
	{
		this.enabledClazz = clazz;
	}

	/**
		Set the class to use when the button is disabled.
	*/
	public void setDisabledClazz(String clazz) 
	{
		this.disabledClazz = clazz;
	}
	
	/**
		Set the class to use in all cases.
	*/
	public void setClazz(String clazz) 
	{
		this.enabledClazz = clazz;
		this.disabledClazz = clazz;
	}

	/**
		Set the icon to use when the button is enabled.
	*/
	@VariableSetter
	@PathSetter
	public void setEnabledIcon(String icon)
	{
		this.enabledIcon = icon;
	}
	/**
		Set the icon to use when the button is disabled.
	*/
	@VariableSetter
	@PathSetter
	public void setDisabledIcon(String icon)
	{
		this.disabledIcon = icon;
	}
	/**
		Set the icon to use in all cases.
	*/
	@VariableSetter
	@PathSetter
	public void setIcon(String icon) 
	{
		this.enabledIcon = icon;
		this.disabledIcon = icon;
	}
	
	/**
		Set the ID of the button.
	*/
	public void setId(String id) 
	{
		this.id = id;
	}

	@VariableSetter
	public void setOnClick(String onClick) 
	{
		this.onClick = onClick;
	}

	/**
		Set the style to use when the button is enabled.
	*/
	public void setEnabledStyle(String style)
	{
		this.enabledStyle = style;
	}
	/**
		Set the style to use when the button is disabled.
	*/
	public void setDisabledStyle(String style)
	{
		this.disabledStyle = style;
	}
	/**
		Set the style to use in all cases.
	*/
	public void setStyle(String style) 
	{
		this.enabledStyle = style;
		this.disabledStyle = style;
	}
	
	/**
		Set the title of the button when it is enabled.
	*/
	public void setEnabledTitle(String title)
	{
		this.enabledTitle = title;
	}
	/**
		Set the title of the button when it is disabled.
	*/
	public void setDisabledTitle(String title)
	{
		this.disabledTitle = title;
	}
	/**
		Set the title of the button in all cases.
	*/
	public void setTitle(String title) 
	{
		this.enabledTitle = title;
		this.disabledTitle = title;
	}

	/**
		Set the tooltip of the button when it is enabled.
	*/
	public void setEnabledTooltip(String tooltip)
	{
		this.enabledTooltip = tooltip;
	}
	/**
		Set the tooltip of the button when it is disabled.
	*/
	public void setDisabledTooltip(String tooltip)
	{
		this.disabledTooltip = tooltip;
	}
	/**
		Set the tooltip of the button in all cases.
	*/
	public void setTooltip(String tooltip) 
	{
		this.enabledTooltip = tooltip;
		this.disabledTooltip = tooltip;
	}
	
	private boolean hasPermission(ClientContext cc, Item accepted, Permission permission)
	{
		boolean hasPermission = false;
		Object currentItem = cc.getCurrentItem();
		if (currentItem instanceof AccessControlled)
		{
			AccessControlled acItem = (AccessControlled)currentItem;
			if (accepted == null || accepted.getItemClass().isInstance(acItem))
			{
				hasPermission = acItem.hasPermission(permission);
			}
		}
		else if (accepted != null)
		{
			hasPermission = cc.getSessionControl().hasPermission(permission, itemToCheck);
			if (permission == Permission.DENIED) hasPermission = !hasPermission;
		}
		return hasPermission;
	}

}
