/*
	$Id: Head.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Gregory Vincic

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.taglib;

import net.sf.basedb.core.SessionControl;
import net.sf.basedb.clients.web.Base;
import net.sf.basedb.util.Values;

import java.util.LinkedHashSet;
import java.util.Arrays;

import javax.servlet.jsp.JspException;
import javax.servlet.jsp.JspTagException;
import javax.servlet.jsp.tagext.TagSupport;

/**
	This tag defines the header section of a page. It must be located inside
	a {@link Page &lt;base:page&gt;} ag.

	<p>
	<b>Syntax:</b>
	<pre>
   &lt;base:head
      styles=...
      scripts=...
   &gt;
</pre>
	<p>

	<table border=1 cellspacing=0 cellpadding=2>
	<tr>
		<td><b>Attribute</b></td>
		<td><b>Default value</b></td>
		<td><b>Required</b></td>
		<td><b>Description</b></td>
	</tr>
	<tr>
		<td>styles</td>
		<td>-</td>
		<td>no</td>
		<td>
			A comma-separated list of stylesheet filenames that is needed by the page.
			The <code>main.css</code> are <code>menu.css</code> automatically included.
			The stylesheets must be located in the <code>/include/styles</code> folder.
			Do not include the folder names in the stylesheet filename.
		</td>
	</tr>
	<tr>
		<td>scripts</td>
		<td>-</td>
		<td>no</td>
		<td>
			A comma-separated list of javascript filenames that is needed by the page.
			The <code>main.js</code> are <code>menu.js</code> automatically included.
			The scripts must be located in the <code>/include/scripts</code> folder.
			Do not include the folder names in the scripts filename.
		</td>
	</tr>
	</table>


	@author Nicklas
	@version 2.0
*/
public class Head
	extends TagSupport
{

	/**
		
	*/
	private static final long serialVersionUID = 948719986796860767L;

	/**
		The parent &lt;base:page&gt; tag.
	*/
	private Page page;
	
	/**
		A comma-separated list of stylesheets filenames.
	*/
	private String styles = null;
	
	/**
		A comma-separated list of JavaScript filenames.
	*/
	private String scripts = null;
	
	/*
		Taglib initialization methods
	*/
	public void setStyles(String styles)
	{
		this.styles = styles;
	}
	public String getStyles()
	{
		return styles;
	}
	
	public void setScripts(String scripts)
	{
		this.scripts = scripts;
	}
	public String getScripts()
	{
		return scripts;
	}

	private void appendStyles(StringBuilder sb, SessionControl sc, boolean hideLongTexts)
	{
		LinkedHashSet<String> allStyles = new LinkedHashSet<String>();
		allStyles.add("main.css");
		if (page.getTypeCode() == Page.PAGE_TYPE_DEFAULT) allStyles.add("menu.css");
		if (getStyles() != null) allStyles.addAll(Arrays.asList(getStyles().split(",")));
		String fontsize = Values.getString(sc != null ? sc.getUserClientSetting("appearance.fontsize") : null, "size_m.css");
		allStyles.add(fontsize);
		if (hideLongTexts) allStyles.add("long_texts.css");
		for (String style : allStyles)
		{
			sb.append("\t<link rel=\"stylesheet\" type=\"text/css\" href=\"");
			sb.append(page.getRoot()).append("include/styles/").append(style).append("\">\n");
		}
	}

	private void appendScripts(StringBuilder sb, boolean scanForLongTexts)
	{
		SessionControl sc = page.getSessionControl();
		LinkedHashSet<String> allScripts = new LinkedHashSet<String>();
		allScripts.add("main.js");
		if (page.getTypeCode() == Page.PAGE_TYPE_DEFAULT) allScripts.add("menu.js");
		if (getScripts() != null) allScripts.addAll(Arrays.asList(getScripts().split(",")));
		for (String script : allScripts)
		{
			sb.append("\t<script language=\"JavaScript\" src=\"");
			sb.append(page.getRoot()).append("include/scripts/").append(script).append("\" type=\"text/javascript\"></script>\n");
		}
		sb.append("\t<script language=\"JavaScript\" type=\"text/javascript\">\n");
		sb.append("\t\tfunction getRoot()\n");
		sb.append("\t\t{\n");
		sb.append("\t\t\treturn '").append(page.getRoot()).append("';\n");
		sb.append("\t\t}\n");
		sb.append("\t\tfunction getScale()\n");
		sb.append("\t\t{\n");
		sb.append("\t\t\treturn ").append(Base.getScale(sc)).append(";\n");
		sb.append("\t\t}\n");
		sb.append("\t\tfunction getSessionId()\n");
		sb.append("\t\t{\n");
		sb.append("\t\t\treturn '").append(sc == null ? "" : sc.getId()).append("';\n");
		sb.append("\t\t}\n");
		if (page.getMaxUrlLength() != null)
		{
			sb.append("\t\tfunction getMaxUrlLength()\n");
			sb.append("\t\t{\n");
			sb.append("\t\t\treturn '").append(page.getMaxUrlLength()).append("';\n");
			sb.append("\t\t}\n");
		}
		sb.append("\t</script>\n");
	}

	public int doStartTag()
		throws JspException
	{
		if (!(getParent() instanceof Page)) throw new JspTagException("Tag <base:head> must be inside a <base:page> tag");
		page = (Page)getParent();
		
		if (page.getTypeCode() == Page.PAGE_TYPE_INCLUDE) return SKIP_BODY;

		SessionControl sc = page.getSessionControl();
		String longTexts = Values.getString(sc != null ? sc.getUserClientSetting("text.long") : null, "display");

		StringBuilder sb = new StringBuilder();
		sb.append("<head>\n");
		sb.append("\t<link rel=\"SHORTCUT ICON\" href=\"").append(page.getRoot()).append("favicon.ico\">\n");
		appendStyles(sb, sc, !"display".equals(longTexts));
		appendScripts(sb, "click".equals(longTexts));
		
		switch (page.getTypeCode())
		{
		case Page.PAGE_TYPE_FRAMESET:
			if(page.getTitle() != null) sb.append("\t<title>").append(page.getBaseVersion()).append(" @ ").append(page.getServerName()).append(" -- ").append(page.getTitle()).append("</title>\n");
			break;

		case Page.PAGE_TYPE_POPUP:
		default:
			if(page.getTitle() != null) sb.append("\t<title>").append(page.getTitle()).append("</title>\n");
			break;
		}

		try
		{
			pageContext.getOut().print(sb.toString());
		}
		catch (Exception ex)
		{
			throw new JspTagException(ex.getMessage());
		}
		return EVAL_BODY_INCLUDE;
	}

	public int doEndTag() 
		throws JspException
	{
		if (page.getTypeCode() != Page.PAGE_TYPE_INCLUDE)
		{
			StringBuilder sb = new StringBuilder();
			sb.append("</head>\n");
			try
			{
				pageContext.getOut().print(sb.toString());
			}
			catch (Exception ex)
			{
				throw new JspTagException(ex.getMessage());
			}
		}
		return EVAL_PAGE;
	}
}



