/**
	$Id: BioPlate.java 4704 2008-12-11 15:27:45Z martin $

	Copyright (C) 2008 Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.BioPlateData;
import net.sf.basedb.core.data.BioWellData;
import net.sf.basedb.core.data.HardwareData;
import net.sf.basedb.core.data.PlateCoordinate;

import java.util.Set;

/**
 	This class is represents physical bio-plates. A bio-plate must have
	a {@link PlateGeometry}
	<p>
	A bio-plate consists of {@link BioWell}:s which are organised into rows and
	columns as specified by the plate geometry.	
  	@author martin
  	@version 2.10
  	@base.modified $Date: 2008-12-11 16:27:45 +0100 (Thu, 11 Dec 2008) $
 	
 */
public class BioPlate 
	extends AnnotatedItem<BioPlateData>
{
	/**
		The type of item represented by this class.
		@see Item#BIOPLATE
		@see #getType()
	*/
	public static final Item TYPE = Item.BIOPLATE;

	/**
		The maximum length of the barcode that can be stored in the
		database. Check the length against this value before calling the
		{@link #setBarcode(String)} method to avoid exceptions.
	*/
	public static final int MAX_BARCODE_LENGTH = BioPlateData.MAX_BARCODE_LENGTH;
	
	/**
	 	Creates a new bioplate with a specific plate geometry 
	  	@param dc DbControl used for database access
	  	@param plateGeometry The geometry the new plate should have 
	  	@return A new BioPlate object
	  	@throws BaseException If anything goes wrong when creating.
	  	@throws InvalidDataException If plateGemoetry is null
	 */
	public static BioPlate getNew(DbControl dc, PlateGeometry plateGeometry)
		throws BaseException, InvalidDataException
	{
		BioPlate bp = dc.newItem(BioPlate.class);
		bp.setPlateGeometry(plateGeometry);
		bp.setName("New bioplate");		
		return bp;
	}
	
	/**
	 	Gets a BioPlate when knowing the item's id
	  	@param dc DbControl used for database access
	  	@param id The item's id.
	  	@return <code>BioPlate</code> object
	  	@throws ItemNotFoundException If no item with the given id was found
	  	@throws PermissionDeniedException If logged in user doesn't have enough 
	  		permission to the requested item
	  	@throws BaseException If there is another error.
	 */
	public static BioPlate getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		BioPlate p = dc.loadItem(BioPlate.class, id);
		if (p == null) throw new ItemNotFoundException("BioPlate[id="+id+"]");
		return p;
	}
	
	/**
	 	Gets query that selects all available bioplates
	  	@return <code>ItemQuery</code> object
	 */
	public static ItemQuery<BioPlate> getQuery()
	{
		return new ItemQuery<BioPlate>(BioPlate.class);
	}

	BioPlate(BioPlateData annotatedData)
	{
		super(annotatedData);
	}
	
	/**
	 	Gets parent plates, always returns null
	 */
	@Override
	public Set<Annotatable> getAnnotatableParents() 
		throws BaseException
	{
		return null;
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}	
	//-------------------------------------------
	
	/*
	 	From the BasicItem class
	 	--------------------------------------
	 */
	/**
	 	Check if there are any <code>MeasuredBioMaterial</code> on
	 	this plate
	 */
	@Override
	public boolean isUsed()
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
				"GET_BIOMATERIAL_ON_PLATE", "count(*)");
		/*
		 	SELECT {1}
		 	FROM MeasuredBioMaterialData mbm
		 	JOIN mbm.bioWell bw
		 	WHERE bw.bioPlate = :bioPlate
		 */
		query.setEntity("bioPlate", this.getData());
		return HibernateUtil.loadData(Long.class, query) > 0 || super.isUsed();
	}
	
	/**
	 	Gets all <code>MeasuredBioMaterial</code> located on this plate
	 */
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
				"GET_BIOMATERIAL_ON_PLATE", "mbm");
		/*
		 	SELECT {1}
		 	FROM MeasuredBioMaterialData mbm
		 	JOIN mbm.BioWell bw
		 	WHERE bw.bioPlate = :bioPlate
		 */
		query.setEntity("bioPlate", this.getData());
		addUsingItems(using, query);
		return using;
	}
	//-------------------------------------------

	/**
		Get the {@link PlateGeometry} of this bioplate.
		@return The <code>PlateGeometry</code> item
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#READ} permission for the plate geometry
		@throws BaseException If there is another error
	*/
	public PlateGeometry getPlateGeometry()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(PlateGeometry.class, getData().getPlateGeometry()); 
	}
	
	private void setPlateGeometry(PlateGeometry plateGeometry)
		throws PermissionDeniedException, InvalidDataException
	{
		if (plateGeometry == null) throw new InvalidUseOfNullException("plateGeometry");
		plateGeometry.checkPermission(Permission.USE);
		checkPermission(Permission.WRITE);
		getData().setPlateGeometry(plateGeometry.getData());		
	}

	/**
	 	Sets the external id of this bio plate
	  	@param externalId The external id to set. Null is allowed.
	  		Maximum length {@link BioPlateData#MAX_EXTERNAL_ID_LENGTH} characters.
	  	@throws PermissionDeniedException If logged in user doesn't have 
	  		{@link Permission#WRITE} on the plate
	  	@throws InvalidDataException If maximum length of externalId is exceeded.
	 */
	public void setExternalId(String externalId)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setExternalId(StringUtil.setNullableString(externalId, "externalId", BioPlateData.MAX_EXTERNAL_ID_LENGTH));
	}
	
	/**
	 	Gets the external id of this bio plate
	  	@return A String object or null
	 */
	public String getExternalId()
	{
		return getData().getExternalId();
	}
	
	/**
	 	Sets the bio plates barcode
	  	@param barcode The barcode to set. Null is allowed.
	  		Maximum length {@link BioPlateData#MAX_BARCODE_LENGTH} characters.
	  	@throws PermissionDeniedException If logged in user doesn't have 
	  		{@link Permission#WRITE} on the plate.
	  	@throws InvalidDataException If maximum length of barcode is exceeded.
	 */
	public void setBarcode(String barcode)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setBarcode(StringUtil.setNullableString(barcode, "barcode", BioPlateData.MAX_BARCODE_LENGTH));
	}
	
	/**
	 	Gets the barcode of this bio plate.
	  	@return A String object or null.
	 */
	public String getBarcode()
	{
		return getData().getBarcode();
	}
	
	/**
	 	Sets the freezer where this bio plate is stored
	  	@param freezer The freezer to set. Null is allowed.
	  	@throws PermissionDeniedException If logged in user doesn't have
	  		{@link Permission#WRITE} on this bio plate or 
	  		{@link Permission#USE} on the freezer.
	 */
	public void setFreezer(Hardware freezer)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		HardwareData hw = null;
		if (freezer != null)
		{
			freezer.checkPermission(Permission.USE);
			hw = freezer.getData();
		}
		getData().setFreezer(hw);
	}
	
	/**
	 	Gets the freezer associated with this bio plate
	  	@return A {@link Hardware} item or null.
	 */
	public Hardware getFreezer()
	{
		return getDbControl().getItem(Hardware.class, getData().getFreezer());
	}
	
	/**
	 	Gets a query that returns all <code>BioWell</code>s 
	 	on this plate.
	  	@return an {@link ItemQuery} object.
	 */
	public ItemQuery<BioWell> getBioWells()
	{
		return BioWell.getQuery(this);
	}

	/**
	 	Gets bio well for specific location.
	  	@param row Row index where the bio well is located.
	  	@param column Column index where the bio well is located
	  	@return A {@link BioWell} object or 
	  		null if no bio well can be found
	 */
	public BioWell getBioWell(int row, int column)
	{
		BioWellData wellData = getData().getBioWells().get(new PlateCoordinate(row, column));
		return getDbControl().getItem(BioWell.class, wellData);
	}
}