/*
	$Id: FileType.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Date;
import java.util.Set;

import net.sf.basedb.core.data.FileTypeData;

/**
	This class is used to represent the type of
	{@link File} items in BASE.

	@author enell, Nicklas
	@version 2.0
	@base.modified $Date: 2009-04-06 14:52:39 +0200 (Mon, 06 Apr 2009) $
	@see File File
*/
public class FileType
	extends BasicItem<FileTypeData>
	implements Nameable, SystemItem, Registered
{
	
	/**
		The type of item represented by this class.
		@see Item#FILETYPE
		@see #getType()
	*/
	public static final Item TYPE = Item.FILETYPE;
	
	/**
		The id for the <code>FileType</code> item representing a
		protocol.
	*/
	public static final String PROTOCOL = "net.sf.basedb.core.FileType.PROTOCOL";

	/**
		The id for the <code>FileType</code> item representing a
		raw data file.
	*/
	public static final String RAW_DATA = "net.sf.basedb.core.FileType.RAW_DATA";

	/**
		The id for the <code>FileType</code> item representing a
		plate mapping file.
	*/
	public static final String PLATE_MAPPING = "net.sf.basedb.core.FileType.PLATE_MAPPING";

	/**
		The id for the <code>FileType</code> item representing a
		reporter map file.
	*/
	public static final String REPORTER_MAP = "net.sf.basedb.core.FileType.REPORTER_MAP";
	
	/**
		The id for the <code>FileType</code> item representing a
		print map file.
	*/
	public static final String PRINT_MAP = "net.sf.basedb.core.FileType.PRINT_MAP";

	/**
		The id for the <code>FileType</code> item representing an
		image file.
	*/
	public static final String IMAGE = "net.sf.basedb.core.FileType.IMAGE";
	
	/**
		The id for the <code>FileType</code> item representing a
		spot images file.
	*/
	public static final String SPOT_IMAGES = "net.sf.basedb.core.FileType.SPOT_IMAGES";
	
	/**
		The id for the <code>FileType</code> item representing a
		reporter file.
	 */
	public static final String REPORTER = "net.sf.basedb.core.FileType.REPORTER";

	/**
		The id for the <code>FileType</code> item representing a
		plate file.
	*/
	public static final String PLATE = "net.sf.basedb.core.FileType.PLATE";

	/**
		The id for the <code>FileType</code> item representing a
		file containing analysed spot data.
		@since 2.8
	*/
	public static final String SPOT_DATA = "net.sf.basedb.core.FileType.SPOT_DATA";

	/**
		The id for the <code>FileType</code> item representing a
		file containing analysed extra value data.
		@since 2.8
	*/
	public static final String EXTRAVALUE_DATA = "net.sf.basedb.core.FileType.EXTRAVALUE_DATA";

	
	/**
		Get a <code>FileType</code> object when you know the ID.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>FileType</code> item
		@throws ItemNotFoundException If an item with the specified ID is not found
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ} permission to the items
		@throws BaseException If there is another error
	*/
	public static FileType getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		FileType ft = dc.loadItem(FileType.class, id);
		if (ft == null) throw new ItemNotFoundException("FileType[id="+id+"]");
		return ft;
	}
	
	/**
		Get a query configured to retrieve <code>FileType</code> items.
		
		@return An {@link ItemQuery} object
	 	@throws BaseException If getting the query fails. 
	*/
	public static ItemQuery<FileType> getQuery()
		throws BaseException
	{
		return new ItemQuery<FileType>(FileType.class);
	}

	/**
		Create a <code>PropertyFilter</code> object that can be used to
		filter files by their file type: file.fileType = FileType
		@param id The id of the file type.
		@return A <code>PropertyFilter</code> object
		@see ItemContext#setPropertyFilter(PropertyFilter)
	*/
	public static PropertyFilter getPropertyFilter(String id)
	{
		return new PropertyFilter("fileType", Operator.EQ, 
			Integer.toString(SystemItems.getId(id)), Type.INT);
	}
	
	/**
		Creates a new filetype item from the given data.
	
		@param data the data.
	*/
	FileType(FileTypeData data)
	{
		super(data);
	}
	
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	public String getName()
	{
		return getData().getName();
	}
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}
	public String getDescription()
	{
		return getData().getDescription();
	}
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------
	/*
		From the SystemItem interface
		-------------------------------------------
	*/
	public String getSystemId()
	{
		return getData().getSystemId();
	}
	public boolean isSystemItem()
	{
		return getData().getSystemId() != null;
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	// -------------------------------------------
	/*
	From the BasicItem class
	-------------------------------------------
	*/
	/**
		Check if:
		<ul>
		<li>any file has been assigned this type
		</ul>
		
		@return TRUE if this FileType is used.
		@throws BaseException If there is an error 
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_FILES_FOR_FILETYPE", "count(*)");
		/*
			SELECT {1}
			FROM FileData f
			WHERE f.fileType = :filetype
		*/
		query.setEntity("filetype", this.getData());
		return HibernateUtil.loadData(Long.class, query) > 0 || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link File}:s of this type
		<ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_FILES_FOR_FILETYPE", "f.id");
			/*
				SELECT {1}
				FROM FileData f
				WHERE f.fileType = :filetype
			*/
		query.setEntity("filetype", this.getData());
		addUsingItems(using, Item.FILE, query);
		return using;
	}
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		if (isSystemItem()) 
		{
			denied |= Permission.deny(Permission.DELETE, Permission.CREATE);
		}
		super.initPermissions(granted, denied);
	}
	// -------------------------------------------
}
