/*
	$Id: PredefinedQuery.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Johan Enell, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.util.XMLUtil;

import java.net.URL;
import java.util.Arrays;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.xpath.XPath;

/**
	This class is used for reading XML files with information
	about predefined queries. The file should be located at the path
	returned by {@link Application#getQueryFile() Application.getQueryFile}.

	The format of the XML file is forced by a DTD. In summary:

	<pre class="code">
&lt;?xml version="1.0" ?&gt;
&lt;!DOCTYPE predefined-queries SYSTEM "predefined.dtd" &gt;
&lt;predefined-queries&gt;
   &lt;query id="GET_GROUPS_FOR_USER" type="HQL"&gt;
      &lt;sql&gt;
         SELECT ug.groupId
         FROM UserGroups ug
         WHERE ug.userId = :user_id
      &lt;/sql&gt;
      &lt;description&gt;
         Loads the group ID for all groups where the user is a member
      &lt;/description&gt;
   &lt;/query&gt;
&lt;/predefined-queries&gt;
</pre>

	@author Nicklas
	@version 2.0
*/
class PredefinedQuery
{

	/**
		Log core events.
	*/
	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.core.PredefinedQuery");

	/**
		Holds the XML document containing the common queries.
	*/
	private static Document commonDom = null;

	/**
		Holds the XML document containing the dialect-specific queries.
	*/
	private static Document dialectDom = null;

	/**
		The path to the query file with common queries.
	*/
	private static final String commonQueryFile = "/common-queries.xml";

	/**
		The path to the query file with dialect-specific queries.
	*/
	private static String dialectQueryFile = null;
	
	/**
		The DTD which is used to validate the XML file.
	*/
	private static final URL dtdFile = PredefinedQuery.class.getResource("/net/sf/basedb/core/dtd/predefined-queries.dtd");
	
	private static boolean isInitialised = false;

	/**
		Load and parse the file with predefined queries.
	*/
	static synchronized void init()
		throws BaseException
	{
		if (isInitialised) return;
		dialectQueryFile = Application.getQueryFile();
		try
		{
			commonDom = XMLUtil.getValidatedXml(PredefinedQuery.class.getResource(commonQueryFile), dtdFile);
			log.info("Loaded predefined queries from file: " + commonQueryFile);
			if (dialectQueryFile != null)
			{
				dialectDom = XMLUtil.getValidatedXml(PredefinedQuery.class.getResource(dialectQueryFile), dtdFile);
				log.info("Loaded predefined queries from file: " + dialectQueryFile);
			}
		}
		catch (Exception ex)
		{
			log.warn("Exception while loading predefined queries", ex);
			throw new BaseException(ex);
		}
		isInitialised = true;
	}
	
	/**
		Unload all settings.
	*/
	static synchronized void unload()
	{
		isInitialised = false;
		dialectDom = null;
		dialectQueryFile = null;
	}

	/**
		Get the query string for the predefined query with
		the specified name.
		@throws BaseException If the query is not found or there 
			is any other error
	*/
	static String getQueryString(String name, String... replacements)
		throws BaseException
	{
		Element query = null;
		try
		{
			XPath findQuery = XPath.newInstance("//query[@id='"+name+"']");
	
			if (dialectDom != null) query = (Element)findQuery.selectSingleNode(dialectDom);
			if (query == null) query = (Element)findQuery.selectSingleNode(commonDom);
		}
		catch (JDOMException ex)
		{
			throw new BaseException(ex);
		}
		if (query == null)
		{
			throw new BaseException("Node '<query id=\""+name+
				"\">' was not found in file '"+commonQueryFile+"'" +
				(dialectQueryFile != null ? " or in file '"+dialectQueryFile+"'" : ""));
		}
		String sql = query.getChildText("sql").trim();
		if (replacements != null && replacements.length > 0)
		{
			String original = sql;
			for (int i = replacements.length; i > 0; --i)
			{
				sql = sql.replaceAll("\\{" + (i) + "\\}", replacements[i-1]);
			}
			assert !sql.equals(original) : "No replacements made to SQL: " + sql + 
				"; replacements=" + Arrays.asList(replacements);
		}
		return sql;
	}

}
