/*
  $Id $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

  BASE is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 3
  of the License, or (at your option) any later version.

  BASE is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Set;

import net.sf.basedb.core.data.QuantityData;
import net.sf.basedb.core.data.UnitData;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.units.UnitCache;

/**
	This class is used to represent a quantity for specifying units
	to annotations.

	@author Nicklas
	@version 2.9
	@base.modified $Date: 2009-04-06 14:15:21 +0200 (Mon, 06 Apr 2009) $
	@see Unit
	@see <a href="http://physics.nist.gov/cuu/Units/units.html">Essentials of 
		the SI: Base &amp; derived units</a>
*/
public class Quantity
	extends BasicItem<QuantityData>
	implements Nameable, SystemItem, Removable
{

	/**
		The type of item represented by this class.
		@see Item#QUANTITY
		@see #getType()
	*/
	public static final Item TYPE = Item.QUANTITY;
	
	/**
		The id for the <code>Quantity</code> object representing length.
		@see "http://physics.nist.gov/cuu/Units/meter.html"
	*/
	public static final String LENGTH = "net.sf.basedb.core.Quantity.LENGTH";

	/**
		The id for the <code>Quantity</code> object representing mass.
		@see "http://physics.nist.gov/cuu/Units/kilogram.html"
	*/
	public static final String MASS = "net.sf.basedb.core.Quantity.MASS";

	/**
		The id for the <code>Quantity</code> object representing time.
		@see "http://physics.nist.gov/cuu/Units/second.html"
	*/
	public static final String TIME = "net.sf.basedb.core.Quantity.TIME";

	/**
		The id for the <code>Quantity</code> object representing temperature.
		@see "http://physics.nist.gov/cuu/Units/kelvin.html"
	*/
	public static final String TEMPERATURE = "net.sf.basedb.core.Quantity.TEMPERATURE";

	/**
		The id for the <code>Quantity</code> object representing electric current.
		@see "http://physics.nist.gov/cuu/Units/ampere.html"
	*/
	public static final String CURRENT = "net.sf.basedb.core.Quantity.CURRENT";

	/**
		The id for the <code>Quantity</code> object representing "amount
		of substance".
		@see "http://physics.nist.gov/cuu/Units/mole.html"
	*/
	public static final String SUBSTANCE = "net.sf.basedb.core.Quantity.SUBSTANCE";

	/**
		The id for the <code>Quantity</code> object representing area.
		@see "http://en.wikipedia.org/wiki/Area"
	*/
	public static final String AREA = "net.sf.basedb.core.Quantity.AREA";

	/**
		The id for the <code>Quantity</code> object representing volume.
		@see "http://en.wikipedia.org/wiki/Volume"
	*/
	public static final String VOLUME = "net.sf.basedb.core.Quantity.VOLUME";

	/**
		The id for the <code>Quantity</code> object representing velocity.
		@see "http://en.wikipedia.org/wiki/Speed"
	*/
	public static final String VELOCITY = "net.sf.basedb.core.Quantity.VELOCITY";

	/**
		The id for the <code>Quantity</code> object representing angle.
		@see "http://en.wikipedia.org/wiki/Radian"
	*/
	public static final String ANGLE = "net.sf.basedb.core.Quantity.ANGLE";

	/**
		The id for the <code>Quantity</code> object representing force.
		@see "http://en.wikipedia.org/wiki/Newton"
	*/
	public static final String FORCE = "net.sf.basedb.core.Quantity.FORCE";
	
	/**
		The id for the <code>Quantity</code> object representing frequency.
		@see "http://en.wikipedia.org/wiki/Newton"
	*/
	public static final String FREQUENCY = "net.sf.basedb.core.Quantity.FREQUENCY";

	/**
		The id for the <code>Quantity</code> object representing density.
		@see "http://en.wikipedia.org/wiki/Density"
	*/
	public static final String DENSITY = "net.sf.basedb.core.Quantity.DENSITY";

	/**
		The id for the <code>Quantity</code> object representing energy.
		@see "http://en.wikipedia.org/wiki/Joule"
	*/
	public static final String ENERGY = "net.sf.basedb.core.Quantity.ENERGY";

	/**
		The id for the <code>Quantity</code> object representing power.
		@see "http://en.wikipedia.org/wiki/Watt"
	*/
	public static final String POWER = "net.sf.basedb.core.Quantity.POWER";

	/**
		The id for the <code>Quantity</code> object representing pressure.
		@see "http://en.wikipedia.org/wiki/Pascal_(unit)"
	*/
	public static final String PRESSURE = "net.sf.basedb.core.Quantity.PRESSURE";

	/**
		The id for the <code>Quantity</code> object representing electrical potential.
		@see "http://en.wikipedia.org/wiki/Volt"
	*/
	public static final String POTENTIAL = "net.sf.basedb.core.Quantity.POTENTIAL";

	/**
		The id for the <code>Quantity</code> object representing electrical resistance.
		@see "http://en.wikipedia.org/wiki/Ohm"
	*/
	public static final String RESISTANCE = "net.sf.basedb.core.Quantity.RESISTANCE";

	/**
		The id for the <code>Quantity</code> object representing electrical charge.
		@see "http://en.wikipedia.org/wiki/Coulomb"
	*/
	public static final String CHARGE = "net.sf.basedb.core.Quantity.CHARGE";

	/**
		The id for the <code>Quantity</code> object representing capacitance.
		@see "http://en.wikipedia.org/wiki/Farad"
	*/
	public static final String CAPACITANCE = "net.sf.basedb.core.Quantity.CAPACITANCE";
	
	/**
		The id for the <code>Quantity</code> object representing magnetic field.
		@see "http://en.wikipedia.org/wiki/Tesla_(unit)"
	*/
	public static final String MAGNETIC_FIELD = "net.sf.basedb.core.Quantity.MAGNETIC_FIELD";

	/**
		The id for the <code>Quantity</code> object representing magnetic flux.
		@see "http://en.wikipedia.org/wiki/Weber_(unit)"
	*/
	public static final String MAGNETIC_FLUX = "net.sf.basedb.core.Quantity.MAGNETIC_FLUX";

	/**
		The id for the <code>Quantity</code> object representing absorbed dose.
		@see "http://en.wikipedia.org/wiki/Gray_(unit)"
	*/
	public static final String ABSORBED_DOSE = "net.sf.basedb.core.Quantity.ABSORBED_DOSE";

	/**
		The id for the <code>Quantity</code> object representing equivalent dose.
		@see "http://en.wikipedia.org/wiki/Sievert"
	*/
	public static final String EQUIVALENT_DOSE = "net.sf.basedb.core.Quantity.EQUIVALENT_DOSE";

	/**
		The id for the <code>Quantity</code> object representing radioactivity.
		@see "http://en.wikipedia.org/wiki/Becquerel"
	*/
	public static final String RADIOACTIVITY = "net.sf.basedb.core.Quantity.RADIOACTIVITY";

	/**
		The id for the <code>Quantity</code> object representing fractions.
		@see "http://en.wikipedia.org/wiki/Parts-per_notation"
	*/
	public static final String FRACTION = "net.sf.basedb.core.Quantity.FRACTION";

	/**
		The id for the <code>Quantity</code> object representing acceleration.
		@see "http://en.wikipedia.org/wiki/Acceleration"
	*/
	public static final String ACCELERATION = "net.sf.basedb.core.Quantity.ACCELERATION";

	/**
		The id for the <code>Quantity</code> object representing torque.
		@see "http://en.wikipedia.org/wiki/Torque"
	*/
	public static final String TORGUE = "net.sf.basedb.core.Quantity.TORQUE";

	/**
		The id for the <code>Quantity</code> object representing momentum.
		@see "http://en.wikipedia.org/wiki/Momentum"
	*/
	public static final String MOMENTUM = "net.sf.basedb.core.Quantity.MOMENTUM";

	/**
		The id for the <code>Quantity</code> object representing flow rate.
		@see "http://en.wikipedia.org/wiki/Volumetric_flow_rate"
	*/
	public static final String FLOW_RATE = "net.sf.basedb.core.Quantity.FLOW_RATE";

	/**
		The id for the <code>Quantity</code> object representing thermal conductivity.
		@see "http://en.wikipedia.org/wiki/Thermal_conductivity"
	*/
	public static final String THERMAL_CONDUCTIVITY = "net.sf.basedb.core.Quantity.THERMAL_CONDUCTIVITY";

	/**
		The id for the <code>Quantity</code> object representing storage sizes.
		@see "http://en.wikipedia.org/wiki/Byte"
	*/
	public static final String STORAGE_SIZE = "net.sf.basedb.core.Quantity.STORAGE_SIZE";
	
	/**
		Create a new <code>Quantity</code> item.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param referenceUnit The reference unit which all units must
			relate to
		@return The new <code>Quantity</code> item
		@throws BaseException If there is an error
	*/
	public static Quantity getNew(DbControl dc, String referenceUnit)
		throws BaseException
	{
		Quantity q = dc.newItem(Quantity.class);
		q.setName("New quantity");
		q.setReferenceUnit(referenceUnit);
		return q;
	}
	
	/**
		Get a <code>Quantity</code> object when you know the ID.

		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@param id The system ID of the item to load.
		@return The <code>Quantity</code> item
		@throws ItemNotFoundException This exception is thrown if an
			item with the specified ID is not found
		@throws PermissionDeniedException This exception is thrown if
			the logged in user doesn't have read
			permission to the item
		@throws BaseException This exception is thrown if there is another
			error
	*/
	public static Quantity getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Quantity q = dc.loadItem(Quantity.class, id);
		if (q == null) throw new ItemNotFoundException("Quantity[ID="+id+"]");
		return q;
	}
	
	/**
		Get a query configured to retrieve quantities.
		
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<Quantity> getQuery()
	{
		return new ItemQuery<Quantity>(Quantity.class);
	}
	
	/**
		Creates a quantity item.
		@param data the data
	*/
	Quantity(QuantityData data)
	{
		super(data);
	}

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check if:
		<ul>
		<li>Units for this quantity exists.
		<li>Annotation types are using this quantity
		</ul>
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_UNITS_FOR_QUANTITY", "count(*)");
		/*
			SELECT {1}
			FROM Unit unt
			WHERE unt.quantity = :quantity
		*/
		query.setEntity("quantity", this.getData());
		boolean used = HibernateUtil.loadData(Long.class, query) > 0;
		if (!used)
		{
			query = HibernateUtil.getPredefinedQuery(session, 
				"GET_ANNOTATIONTYPES_FOR_QUANTITY", "count(*)");
			/*
				SELECT {1}
				FROM AnnotationTypeData at
				WHERE at.quantity = :quantity
			*/
			query.setEntity("quantity", this.getData());
			used = HibernateUtil.loadData(Long.class, query) > 0;
		}
		return used || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link Unit}:s for this quantity
		<li>{@link AnnotationType}:s using this quantity
		<ul>
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		
		// Units
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
				"GET_UNITS_FOR_QUANTITY", "unt.id");
		/*
			SELECT {1}
			FROM Unit unt
			WHERE unt.quantity = :quantity
		*/
		query.setEntity("quantity", this.getData());
		addUsingItems(using, Item.UNIT, query);
		
		// Annotation types
		query = HibernateUtil.getPredefinedQuery(session, 
				"GET_ANNOTATIONTYPES_FOR_QUANTITY", "at.id");
		/*
			SELECT {1}
			FROM AnnotationTypeData at
			WHERE at.quantity = :quantity
		*/
		query.setEntity("quantity", this.getData());
		addUsingItems(using, Item.ANNOTATIONTYPE, query);
		return using;
	}

	/**
		If this is a system quantity, delete and create permissions are revoked.
	*/
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		if (isSystemItem()) 
		{
			denied |= Permission.deny(Permission.DELETE, Permission.CREATE);
		}
		super.initPermissions(granted, denied);
	}
	// -------------------------------------------
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	//  -------------------------------------------
	/*
		From the Removable interface
		-------------------------------------------
	*/
	public boolean isRemoved()
	{
		return getData().isRemoved();
	}
	public void setRemoved(boolean removed)
		throws PermissionDeniedException
	{
		checkPermission(removed ? Permission.DELETE : Permission.WRITE);
		getData().setRemoved(removed);
	}
	// -------------------------------------------
	/*
		From the SystemItem interface
		-------------------------------------------
	*/
	public String getSystemId()
	{
		return getData().getSystemId();
	}

	public boolean isSystemItem()
	{
		return getSystemId() != null;
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	public String getName()
	{
		return getData().getName();
	}

	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}

	public String getDescription()
	{
		return getData().getDescription();
	}

	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------

	/**
		The maximum length of the reference unit that can be stored in the database.
		@see #setReferenceUnit(String)
	*/
	public static final int MAX_REFERENCE_UNIT_LENGTH = QuantityData.MAX_REFERENCE_UNIT_LENGTH;

	/**
		The reference unit describes the unit that all other units must
		relate to. When setting {@link Unit#getReferenceFactor()} and 
		{@link Unit#getReferenceOffset()} for a unit the following formula
		should be used:
		<pre class="code">
1 reference unit * factor + offset = 1 of the unit
</pre>
		For example:
	<pre class="code">
1 meter * 0.001 + 0 = 1 millimeter
1 kelvin * 1 + 273.15 = 1 °Celsius
</pre>
		@see Unit#getReferenceFactor()
		@see Unit#getReferenceOffset()
	*/
	public String getReferenceUnit()
	{
		return getData().getReferenceUnit();
	}

	/**
		Sets the reference unit.
		@param referenceUnit
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission
		@throws InvalidDataException If the value is null, or longer than
			{@link #MAX_REFERENCE_UNIT_LENGTH}
	*/
	public void setReferenceUnit(String referenceUnit)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setReferenceUnit(StringUtil.setNotNullString(referenceUnit, 
				"referenceUnit", MAX_REFERENCE_UNIT_LENGTH));
	}
	
	/**
		Get a query that returns all units of this quantity.
		@return An {@link ItemQuery} object
	*/
	public ItemQuery<Unit> getUnits()
	{
		ItemQuery<Unit> query = Unit.getQuery();
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("quantity"), 
				Hql.entity(this)
			)
		);
		return query;
	}
	
	/**
		Get a a unit by symbol, also checking among the aliases. 
		Note! This method always results in a database query. If 
		you need to lookup lots of symbols, use a {@link UnitCache}
		instead which has better performance.
		@param symbol The symbol
	*/
	public Unit getUnit(String symbol)
	{
		UnitData unit = getUnit(getDbControl().getHibernateSession(), symbol);
		if (unit == null)
		{
			throw new ItemNotFoundException("Unit[" + getName() +"=" + symbol+ "]");
		}
		return getDbControl().getItem(Unit.class, unit);
	}
	
	/**
		Get the unit with a given symbol, or null if no unit is found.
	*/
	UnitData getUnit(org.hibernate.Session session, String symbol)
	{
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(
				session, "GET_UNIT_WITH_SYMBOL_FOR_QUANTITY");
		query.setEntity("quantity", this.getData());
		query.setString("symbol", symbol);
		UnitData unit = HibernateUtil.loadData(UnitData.class, query);
		return unit;
	}
	
}
