/*
	$Id: Update.java 4938 2009-05-19 09:24:55Z nicklas $

	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.sql.SQLException;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.hibernate.mapping.Table;

import net.sf.basedb.core.data.ArrayDesignData;
import net.sf.basedb.core.data.DataCubeData;
import net.sf.basedb.core.data.ExperimentData;
import net.sf.basedb.core.data.FileData;
import net.sf.basedb.core.data.FormulaData;
import net.sf.basedb.core.data.JobData;
import net.sf.basedb.core.data.MeasuredBioMaterialData;
import net.sf.basedb.core.data.PlateData;
import net.sf.basedb.core.data.PlateMappingData;
import net.sf.basedb.core.data.PluginConfigurationData;
import net.sf.basedb.core.data.PluginDefinitionData;
import net.sf.basedb.core.data.ProjectData;
import net.sf.basedb.core.data.RawBioAssayData;
import net.sf.basedb.core.data.RoleData;
import net.sf.basedb.core.data.RoleKeyData;
import net.sf.basedb.core.data.SchemaVersionData;
import net.sf.basedb.core.data.TransformationData;
import net.sf.basedb.core.data.keyring.PluginKeys;
import net.sf.basedb.core.dbengine.DbEngine;
import net.sf.basedb.core.dbengine.TableInfo;
import net.sf.basedb.core.hibernate.ExecuteUpdateWork;
import net.sf.basedb.core.hibernate.TableInfoWork;

/**
	This class contains static methods used to update items already in the database.
	
	<table>
	<tr>
		<th>Schema version</th>
		<th>Description</th>
	</tr>
	
	<tr>
		<td>1</td>
		<td>The original schema as released in RC1.</td>
	</tr>
	
	<tr>
		<td>2</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.FormulaData}.
			<li>Added {@link net.sf.basedb.core.data.JobData#getActiveProjectId()} property. 
				The update must set all null values to 0.
			<li>Added {@link SchemaVersionData}.
			</ul>
		</td>
	</tr>
	
	<tr>
		<td>3</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.ExperimentData#getExperimentalFactors()}. 
				No need to update existing values.
			</ul>
		</td>
	</tr>

	<tr>
		<td>4</td>
		<td>
			<ul>
			<li>Removed unique index for {@link net.sf.basedb.core.data.HelpData#getExternalId()}. 
			</ul>
		</td>
	</tr>

	<tr>
		<td>5</td>
		<td>
			<ul>
			<li>Removed the AffyFeatures table and AffyFeatureData item. The table
				has to be removed manually.
			</ul>
		</td>
	</tr>
	
	<tr>
		<td>6</td>
		<td>
			<ul>
			<li>Added label and description property to 
				{@link net.sf.basedb.core.data.ParameterValueData}
			<li>Added {@link net.sf.basedb.core.data.AnyToAnyData}
			</ul>
		</td>
	</tr>

	<tr>
		<td>7</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.RawBioAssayData#getHasData()}.
				The update must set all null values to true or false depending on
				the number of spots.
			<li>Removed {@link net.sf.basedb.core.data.FileAttachableData} from
				<code>RawBioAssayData</code>.
			<li>Delete all {@link FormulaData} with affymetrix raw data type.
			</ul>
		</td>
	</tr>
	
	<tr>
		<td>8</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.HardwareData} to BioMaterialEvents and PlateEvents.
			</ul>
		</td>
	</tr>
	
	<tr>
		<td>9</td>
		<td>
			<ul>
			<li>Added removable property to {@link net.sf.basedb.core.data.HardwareTypeData}
				The update of existing data is done before the database is initiated.
			</ul>
		</td>
	</tr>

	<tr>
		<td>10</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.PluginDefinitionData#getSupportsConfigurations()}
				and {@link net.sf.basedb.core.data.PluginDefinitionData#getRequiresConfiguration()}
				The update of existing data is done before the database is initiated.
			</ul>
		</td>
	</tr>
	
	<tr>
		<td>11</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.AnnotationTypeCategoryData}
			</ul>
		</td>
	</tr>
	
	<tr>
		<td>12</td>
		<td>
			<ul>
			<li>Fixed incorrect {@link PlateMapping}:s. The 96-to-384 and 384-to-1536 were
				wrong. The fix renames and flags the old ones as removed.
			</ul>
		</td>
	</tr>

	<tr>
		<td>13</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.JobAgentData} and related classes.
				The update sets the {@link net.sf.basedb.core.data.PluginDefinitionData#isTrusted()}
				property to true for all existing plugins.
			</ul>
		</td>
	</tr>

	<tr>
		<td>14</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.PluginPermission} and related classes.
				The update sets the {@link net.sf.basedb.core.data.PluginDefinitionData#getUsePermissions()}
				property to true for all existing plugins.
			</ul>
		</td>
	</tr>
	
	<tr>
		<td>15</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.JobData#getSendMessage()} and 
				{@link net.sf.basedb.core.data.JobData#getStackTrace()} properties.
				The update sets the send message to true for all existing jobs.
			</ul>
		</td>
	</tr>
	
	<tr>
		<td>16</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.ExtraValueData#getJob()}.
			</ul>
		</td>
	</tr>
	
	<tr>
		<td>17</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.MimeTypeData#getFileType()}.
			</ul>
		</td>
	</tr>
	
	<tr>
		<td>18</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.ExperimentData#getDirectory()} and 
				{@link net.sf.basedb.core.data.ExperimentData#getTitle()}.
			</ul>
		</td>
	</tr>
	
	<tr>
		<td>19</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.PluginConfigurationData#getParameterVersion()}
			<li>Added {@link net.sf.basedb.core.data.JobData#getParameterVersion()}
			<li>Changed {@link net.sf.basedb.core.data.PluginConfigurationData#getConfigurationValues()}.
			</ul>
			The update sets the parameter version for existing jobs and configurations to 1
			and copies data from the old table (PluginConfigurationValues) to the new table
			(VersionedPluginConfigurationValues)
		</td>
	</tr>
	
	<tr>
		<td>20</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.FormulaData#getColoring()}
			<li>Added {@link net.sf.basedb.core.data.ExtraValueTypeData#getColoring()}
			</ul>
			The update sets all color information to false and null for
			existing items. The install may later add coloring to some of the formulas.
		</td>
	</tr>
	
	<tr>
		<td>21</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.PluginDefinitionData#getAllowImmediateExecution()}
			</ul>
			The update sets the value to false for all existing plugins with a null value. 
			The installation may later change the value to true for some plugins (ie. export 
			plugins).
		</td>
	</tr>
	
	<tr>
		<td>22</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.ProtocolData#getParameters()}
			<li>Added {@link net.sf.basedb.core.data.AnnotationTypeData#isProtocolParameter()}
			</ul>
			The update sets the <code>isProtocolParameter</code> value to false for all existing 
			annotation types with a null value.
		</td>
	</tr>
	
	<tr>
		<td>23</td>
		<td>
			<ul>
			<li>Made {@link net.sf.basedb.core.data.BioAssaySetData} and 
				{@link net.sf.basedb.core.data.BioAssayData} annotatable.
			</ul>
		</td>
	</tr>
	
	<tr>
		<td>24</td>
		<td>
			<ul>
			<li>File parameters are stored as item parameters. The update will
				move existing file parameter into the ItemValues table.
			</ul>
		</td>
	</tr>

	<tr>
		<td>25</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.DiskUsageData#getItemType()}.
				The update will first set all null values to the correct item type
				by checking the diskconsumable items.
			</ul>
		</td>
	</tr>
	
	<tr>
		<td>26</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.AnnotationTypeData#getExternalId()}.
			</ul>
		</td>
	</tr>
	
	<tr>
		<td>27</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.ProjectData#getProjectDefaults()}.
			</ul>
		</td>
	</tr>
	
	<tr>
		<td>28</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.DataCubeData#getMaxRawMappingsForSpot()}.
			</ul>
			The update must query the dynamic database and set this value for
			all existing data cubes.
		</td>
	</tr>

	<tr>
		<td>29</td>
		<td>
			No changes at all except increasing the schema version number. 
			The veryfication of remaining quantity is moved	to next update of schema version. 
		</td>
	</tr>
	
	<tr>
		<td>30</td>
		<td>
			No schema change as such, but we need to verify and possibly update the
			{@link net.sf.basedb.core.data.MeasuredBioMaterialData#getRemainingQuantity()}
			since a bug may have caused it to have an incorrect value.
		</td>
	</tr>
	
	<tr>
		<td>31</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.FileData#isWriteProtected()}.
			<li>Added {@link net.sf.basedb.core.data.FileData#getLastUpdate()}.
			</ul>
			The update sets the <code>isWriteProtected</code> value to false for all existing 
			files with a null value and <code>lastUpdate</code> to the date and time
			the internal file last was modified for existing files.
		</td>
	</tr>
	
	<tr>
		<td>32</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.GroupData#isDefault()}.
			<li>Added {@link net.sf.basedb.core.data.RoleData#isDefault()}.
			</ul>
			The update sets the <code>isDefault</code> value to false for all existing 
			groups and roles with a null value.
		</td>
	</tr>
	
	<tr>
		<td>33</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.JobData#getRemoveJobWhenFinished()}.
			</ul>
			The update sets the <code>removeJobWhenFinished</code> value to false for all existing jobs.
		</td>
	</tr>

	<tr>
		<td>34</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.ContextData#getItemId()}.
			</ul>
			The update sets the <code>itemId</code> value to 0 for all existing contexts.
		</td>
	</tr>
	
	<tr>
		<td>35</td>
		<td>
			Check and update the size of File items, since they may be incorrect
			due to a bug in the File class. The MD5 value in that case also
			incorrect and is set to null.
			See ticket http://base.thep.lu.se/ticket/575
		</td>
	</tr>
	
	<tr>
		<td>36</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.FormulaData#getAverageMethod()}.
			<li>Added {@link net.sf.basedb.core.data.ExtraValueTypeData#getAverageMethod()}.
			</ul> 
			The update sets the <code>averageMethod</code> to {@link Formula.AverageMethod#ARITHMETIC_MEAN}.
			See ticket http://base.thep.lu.se/ticket/648
		</td>
	</tr>
	
	<tr>
		<td>37</td>
		<td>
			<ul>
				<li>Implement {@link net.sf.basedb.core.data.ExtendableData} to {@link net.sf.basedb.core.data.UserData}</li>
			</ul>
			Only the schemaVersion number is increased in this class, the real update work is done when the database is re-build.
		</td>
	</tr> 

	<tr>
		<td>38</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.JobData#getExperiment()}.
			</ul>
			The update sets the experiment for jobs that belong to a single
			transformation.
		</td>
	</tr>
	
	<tr>
		<td>39</td>
		<td>
			Remove AnyToAny links created by older versions of the 
			Base1PluginExecuter plug-in.
		</td>
	</tr>
	
	<tr>
		<td>40</td>
		<td>
			Used to "Change invalid property filter on 'guiContexts.itemType' to 
			'$ctx.itemType'", but since the source problem never was fixed this is now
			done in 43.
		</td>
	</tr>
	
	<tr>
		<td>41</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.MimeTypeData#getAutoCompress()}
			<li>Added {@link net.sf.basedb.core.data.FileData#isCompressed()}. 
			<li>Added {@link net.sf.basedb.core.data.DirectoryData#getAutoCompress()}. 
			</ul>
			The update sets all values on existing items to false.
		</td>
	</tr>
	
	<tr>
		<td>42</td>
		<td>
			No longer used.
		</td>
	</tr>

	<tr>
		<td>43</td>
		<td>
			Change invalid property filter on 'guiContexts.itemType' to 
			'$ctx.itemType'". Was originally fixed by 40, but since the source
			problem never was fixed it must be done again.
		</td>
	</tr>

	<tr>
		<td>44</td>
		<td>
			<ul>
			<li>Added {@link Platform}, {@link DataFileType} and
				a lot of related classes.
			</ul>
			The update will set a platform for all array designs and
			raw bioassays. For Affymetrix data it will move the CEL and CDF files
			from AnyToAny links into FileSet:s.
		</td>
	</tr>

	<tr>
		<td>45</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.PluginDefinitionData#getUseInternalJobQueue()}
			</ul>
			The update sets all values on existing items to true (if no job agent has been 
			configured for the plug-in) or false (if a job agent is configured).
		</td>
	</tr>
	
	<tr>
		<td>46</td>
		<td>
			<ul>
			<li>Made {@link net.sf.basedb.core.data.PluginDefinitionData}
				and {@link net.sf.basedb.core.data.PluginConfigurationData} 
				annotatable.
			</ul>
			No special database update is needed. Only increase the schema version.
		</td>
	</tr>
	<tr>
		<td>47</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.JobData#getSignalTransporter()}.
			</ul>
			No special database update is needed. Only increase the schema version.
		</td>
	</tr>
	<tr>
		<td>48</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.ReporterData#getLastSource()}.
			</ul>
			No special database update is needed. Only increase the schema version.
		</td>
	</tr>
	<tr>
		<td>49</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.FeatureData#getExternalId()}
			<li>Added {@link net.sf.basedb.core.data.ArrayDesignData#getFeatureIdentificationMethod()}
			</ul>
			The update sets the identification method to COORDINATES for all existing
			array designs, and deletes the unique index on row, column, block on
			the Features table.
		</td>
	</tr>
	<tr>
		<td>50</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.DataFileTypeData#getValidatorJarPath()}
			<li>Added {@link net.sf.basedb.core.data.DataFileTypeData#getMetadataReaderJarPath()}
			</ul>
			No special database update is needed. Only increase the schema version.
		</td>
	</tr>
	<tr>
		<td>51</td>
		<td>
			External plug-in paths should be relative to the plugins.dir setting
			in base.config. The update loads all PluginDefinitions with a JAR path
			and convert the path with {@link PluginDefinition#convertToRelative(String)}.
		</td>
	</tr>
	
	<tr>
		<td>52</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.ArrayDesignData#getNumArrays()}. The update
				sets the value to 1 for all existing array designs.
			<li>Added {@link net.sf.basedb.core.data.HybridizationData#getNumArrays()}. The update
				sets the value to 1 for all existing hybridizations.
			</ul>
		</td>
	</tr>

	<tr>
		<td>53</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.ArrayDesignBlockData#getFeatureDiameter()}. The update
				sets the value to 0 for all existing array designs.
			</ul>
		</td>
	</tr>
	
	<tr>
		<td>54</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.RawBioAssayData#getJob()}. 
			</ul>
			No special database update is needed. Only increase the schema version.
		</td>
	</tr>
	
	<tr>
		<td>55</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.ArrayDesignData#getJob()}. 
			</ul>
			No special database update is needed. Only increase the schema version.
		</td>
	</tr>
	<tr>
		<td>56</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.ReporterListData#getSize()}. 
			</ul>
			The update needs to calculate the number of reporters in existing
			lists.
		</td>
	</tr>
	
	<tr>
		<td>57</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.JobData#getJobAgentId()}. 
			</ul>
			The update grants {@link Permission#SELECT_JOBAGENT} to the
			administrators, power users and job agent roles.
		</td>
	</tr>
	<tr>
		<td>58</td>
		<td>
			<ul>
			<li>Added {@link net.sf.basedb.core.data.DirectoryData#getAutoShare()}. 
			</ul>
			The update sets the value to false for all directories.
		</td>
	</tr>
	<tr>
		<td>59</td>
		<td>
			Added:
			<ul>
			<li>{@link net.sf.basedb.core.data.BioAssaySetData#getFileSet()}. 
			<li>{@link net.sf.basedb.core.data.BioAssaySetData#getNumFileReporters()}. 
			<li>{@link net.sf.basedb.core.data.BioAssaySetData#getNumFileSpots()}. 
			<li>{@link net.sf.basedb.core.data.BioAssayData#getNumFileSpots()}. 
			<li>{@link net.sf.basedb.core.data.ExtraValueData#getFileSet()}.
			<li>{@link net.sf.basedb.core.data.ExtraValueData#getNumFileValues()}.
			</ul>
			The update sets the spot/reporter/value count to 0 for all existing bioassay
			sets and extra values.
		</td>
	</tr>
	<tr>
		<td>60</td>
		<td>
			Not a schema change as such. We need to update incorrect quota values
			for experiments. See http://base.thep.lu.se/ticket/1056
		</td>
	</tr>
	<tr>
		<td>61</td>
		<td>
			Added:
			<ul>
			<li>{@link net.sf.basedb.core.data.JobData#getPluginVersion()}. 
			</ul>
			No special database update is needed. Only increase the schema version.
		</td>
	</tr>
	<tr>
		<td>62</td>
		<td>
			Added:
			<ul>
			<li>{@link net.sf.basedb.core.data.JobData#getContextItem()}. 
			<li>{@link net.sf.basedb.core.data.JobData#getContextType()}. 
			<li>{@link net.sf.basedb.core.data.JobData#getSubContext()}. 
			</ul>
			No special database update is needed. Only increase the schema version.
		</td>
	</tr>
	
	<tr>
		<td>63</td>
		<td>
			Added:
			<ul>
			<li>{@link net.sf.basedb.core.data.JobData#isDryRun()}. 
			</ul>
			The update will set the value for all existing jobs to false.
		</td>
	</tr>
	
	<tr>
		<td>64</td>
		<td>
			Added:
			<ul>
			<li>{@link net.sf.basedb.core.data.FileData#getCharacterSet()}. 
			</ul>
			No special database update is needed. Only increase the schema version.
		</td>
	</tr>
	<tr>
		<td>65</td>
		<td>
			Added:
			<ul>
			<li>{@link net.sf.basedb.core.data.QuantityData}. 
			<li>{@link net.sf.basedb.core.data.UnitData}. 
			<li>{@link net.sf.basedb.core.data.UnitSymbolData}. 
			<li>{@link net.sf.basedb.core.data.AnnotationData#getUnit()}. 
			<li>{@link net.sf.basedb.core.data.AnnotationTypeData#getQuantity()}. 
			</ul>
			No special database update is needed. Only increase the schema version.
		</td>
	</tr>
	<tr>
		<td>66</td>
		<td>
			Added:
			<ul>
			<li>{@link net.sf.basedb.core.data.BioPlateData}.
			<li>{@link net.sf.basedb.core.data.BioWellData}.
			</ul>
			No special database update is needed. Only increase the schema version.
		</td>
	</tr>
	<tr>
		<td>67</td>
		<td>
			Added:
			<ul>
			<li>{@link net.sf.basedb.core.data.BioMaterialListData}.
			</ul>
			No special database update is needed. Only increase the schema version.
		</td>
	</tr>
	<tr>
		<td>68</td>
		<td>
			Added:
			<ul>
			<li>{@link net.sf.basedb.core.data.ProjectData#getAutoPermission()}.
			</ul>
			The update sets the value for all existing project to 31 (=read, use, write, delete).
		</td>
	</tr>
	<tr>
		<td>69</td>
		<td>
			Added:
			<ul>
			<li>{@link net.sf.basedb.core.data.RegisteredData}.
			</ul>
			No special database update is needed. Only increase the schema version.
		</td>
	</tr>
	<tr>
		<td>70</td>
		<td>
			Added:
			<ul>
			<li>{@link net.sf.basedb.core.data.BioAssaySetData#getIntensityTransform()}.
			<li>{@link net.sf.basedb.core.data.FormulaData#getSourceIntensityTransform()}
			<li>{@link net.sf.basedb.core.data.FormulaData#getResultIntensityTransform()}
			</ul>
			The update sets the values for all existing bioassay sets to 0 (=no transform). 
		</td>
	</tr>
	<tr>
		<td>71</td>
		<td>
			Deprecated:
			<ul>
			<li>{@link net.sf.basedb.plugins.BioAssaySetExporter}.
			</ul>
			No special database update is needed. Only increase the schema version.
		</td>
	</tr>
	<tr>
		<td>72</td>
		<td>
			Made {@link Protocol} an {@link Annotatable} item.
			No special database update is needed. Only increase the schema version.
		</td>
	</tr>
	</table>

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2009-05-19 11:24:55 +0200 (Tue, 19 May 2009) $
*/
public final class Update
{
	/**
		Log core events.
	*/
	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.core.Update");

	/**
		Update the database by modifying existing items to follow new requirements.
		
		@param progress An object implementing the {@link ProgressReporter}
			interface
		@param rootLogin The root user login
		@param rootPassword The root user password
		@throws BaseException
	 */
	public static synchronized void updateDatabase(ProgressReporter progress, String rootLogin, String rootPassword)
		throws BaseException
	{
		final int total_progress_steps = Install.NEW_SCHEMA_VERSION;
		final float progress_factor = 100 / total_progress_steps;
		if (progress != null) progress.display(0, "Updating database...");
		org.hibernate.Session session = null;
		try
		{
			Application.start(false, false, false);

			// Test root user account
			SessionControl sc = Application.newSessionControl(null, null, null);
			sc.login(rootLogin, rootPassword, null, false);
			if (sc.getLoggedInUserId() != SystemItems.getId(User.ROOT))
			{
				throw new PermissionDeniedException("User '" + rootLogin + "' is not the root account.");
			}
			
			session = HibernateUtil.newSession();
			int schemaVersion = getSchemaVersion(session);
			
			if (schemaVersion < 2)
			{
				if (progress != null) progress.display((int)(1*progress_factor), "--Updating schema version: " + schemaVersion + " -> 2...");
				schemaVersion = updateToSchemaVersion2(session);
			}

			if (schemaVersion < 3)
			{
				if (progress != null) progress.display((int)(2*progress_factor), "--Updating schema version: " + schemaVersion + " -> 3...");
				schemaVersion = setSchemaVersionInTransaction(session, 3);
			}
			
			if (schemaVersion < 4)
			{
				if (progress != null) progress.display((int)(3*progress_factor), "--Updating schema version: " + schemaVersion + " -> 4...");
				schemaVersion = updateToSchemaVersion4(session);
			}
			
			if (schemaVersion < 5)
			{
				if (progress != null) 
				{
					progress.display((int)(4*progress_factor), "--Updating schema version: " + schemaVersion + " -> 5...");
					progress.display((int)(4*progress_factor), "--NOTE! The AffyFeatures table is no longer used by BASE. " +
						"It may be removed from the database.");
				}
				schemaVersion = setSchemaVersionInTransaction(session, 5);
			}
			
			if (schemaVersion < 6)
			{
				if (progress != null) progress.display((int)(5*progress_factor), "--Updating schema version: " + schemaVersion + " -> 6...");
				schemaVersion = setSchemaVersionInTransaction(session, 6);
			}
			
			if (schemaVersion < 7)
			{
				if (progress != null) progress.display((int)(6*progress_factor), "--Updating schema version: " + schemaVersion + " -> 7...");
				schemaVersion = updateToSchemaVersion7(session);
			}

			// schemaVersion 8, 9 only require an update to the version number
			if (schemaVersion < 9)
			{
				if (progress != null) progress.display((int)(8*progress_factor), "--Updating schema version: " + schemaVersion + " -> 9...");
				schemaVersion = setSchemaVersionInTransaction(session, 9);
			}

			if (schemaVersion < 10)
			{
				if (progress != null) progress.display((int)(9*progress_factor), "--Updating schema version: " + schemaVersion + " -> 10...");
				schemaVersion = updateToSchemaVersion10(sc);
			}
			
			if (schemaVersion < 18)
			{
				if (progress != null) progress.display((int)(10*progress_factor), "--Updating schema version: " + schemaVersion + " -> 18...");
				schemaVersion = setSchemaVersionInTransaction(session, 18);
			}
			
			if (schemaVersion < 19)
			{
				if (progress != null) progress.display((int)(18*progress_factor), "--Updating schema version: " + schemaVersion + " -> 19...");
				schemaVersion = updateToSchemaVersion19(session);
				if (progress != null) progress.display((int)(18*progress_factor), 
					"--NOTE! The \"PluginConfigurationValues\" table is no longer used by BASE. It may be removed from the database.");
			}

			// Schemaversion 20-23 only updates the version number 
			if (schemaVersion < 23)
			{
				if (progress != null) progress.display((int)(19*progress_factor), "--Updating schema version: " + schemaVersion + " -> 23...");
				schemaVersion = setSchemaVersionInTransaction(session, 23);
			}

			if (schemaVersion < 24)
			{
				if (progress != null) progress.display((int)(23*progress_factor), "--Updating schema version: " + schemaVersion + " -> 24...");
				schemaVersion = updateToSchemaVersion24(session);
				if (progress != null) progress.display((int)(23*progress_factor), 
				"--NOTE! The \"FileValues\" table is no longer used by BASE. It may be removed from the database.");
			}
			
			if (schemaVersion < 25)
			{
				if (progress != null) progress.display((int)(24*progress_factor), "--Updating schema version: " + schemaVersion + " -> 25...");
				schemaVersion = updateToSchemaVersion25(session);
			}

			if (schemaVersion < 26)
			{
				if (progress != null) progress.display((int)(25*progress_factor), "--Updating schema version: " + schemaVersion + " -> 27...");
				schemaVersion = setSchemaVersionInTransaction(session, 27);
			}
			
			if (schemaVersion < 28)
			{
				if (progress != null) progress.display((int)(27*progress_factor), "--Updating schema version: " + schemaVersion + " -> 28...");
				schemaVersion = updateToSchemaVersion28(sc);
			}
			
			if (schemaVersion < 29)
			{
				if (progress != null) progress.display((int)(28*progress_factor), "--Updating schema version: " + schemaVersion + " -> 29...");
				schemaVersion = setSchemaVersionInTransaction(session, 29);
			}
			
			if (schemaVersion < 30)
			{
				if (progress != null) progress.display((int)(29*progress_factor), "--Updating schema version: " + schemaVersion + " -> 30...");			 
				schemaVersion = updateToSchemaVersion30(session);
			}

			if (schemaVersion < 31)
			{
				if (progress != null) progress.display((int)(30*progress_factor), "--Updating schema version: " + schemaVersion + " -> 31...");
				schemaVersion = updateToSchemaVersion31(session);
			}		
			
			//	Schemaversion 32-34 only updates the version number 
			if (schemaVersion < 34)
			{
				if (progress != null) progress.display((int)(33*progress_factor), "--Updating schema version: " + schemaVersion + " -> 34...");				 
				schemaVersion = setSchemaVersionInTransaction(session, 34);
			}
			
			if (schemaVersion < 35)
			{
				if (progress != null) progress.display((int)(34*progress_factor), "--Updating schema version: " + schemaVersion + " -> 35...");				 
				schemaVersion = updateToSchemaVersion35(session);
			}
			
			//	Schemaversion 36-37 only updates the version number 
			if (schemaVersion < 37)
			{
				if (progress != null) progress.display((int)(36*progress_factor), "--Updating schema version: " + schemaVersion + " -> 37...");
				schemaVersion = setSchemaVersionInTransaction(session, 37);
			}
			
			if (schemaVersion < 38)
			{
				if (progress != null) progress.display((int)(37*progress_factor), "--Updating schema version: " + schemaVersion + " -> 38...");				 
				schemaVersion = updateToSchemaVersion38(session);
			}
			
			if (schemaVersion < 39)
			{
				if (progress != null) progress.display((int)(38*progress_factor), "--Updating schema version: " + schemaVersion + " -> 39...");				 
				schemaVersion = updateToSchemaVersion39(session);
			}
					
			if (schemaVersion < 41)
			{
				if (progress != null) progress.display((int)(40*progress_factor), "--Updating schema version: " + schemaVersion + " -> 41...");
				schemaVersion = setSchemaVersionInTransaction(session, 41);
			}

			if (schemaVersion < 42)
			{
				if (progress != null) progress.display((int)(41*progress_factor), "--Updating schema version: " + schemaVersion + " -> 42...");
				schemaVersion = updateToSchemaVersion42(session);
			}
			
			if (schemaVersion < 43)
			{
				if (progress != null) progress.display((int)(42*progress_factor), "--Updating schema version: " + schemaVersion + " -> 43...");				 
				schemaVersion = updateToSchemaVersion43(session);
			}

			if (schemaVersion < 44)
			{
				if (progress != null) progress.display((int)(43*progress_factor), "--Updating schema version: " + schemaVersion + " -> 44...");
				schemaVersion = updateToSchemaVersion44(sc);
			}
			
			if (schemaVersion < 45)
			{
				if (progress != null) progress.display((int)(44*progress_factor), "--Updating schema version: " + schemaVersion + " -> 45...");
				schemaVersion = updateToSchemaVersion45(session);
			}
			
			// Schemaversion 46-48 only updates the version number
			if (schemaVersion < 48)
			{
				if (progress != null) progress.display((int)(47*progress_factor), "--Updating schema version: " + schemaVersion + " -> 48...");
				schemaVersion = setSchemaVersionInTransaction(session, 48);
			}

			if (schemaVersion < 49)
			{
				if (progress != null) progress.display((int)(48*progress_factor), "--Updating schema version: " + schemaVersion + " -> 49...");
				schemaVersion = updateToSchemaVersion49(session);
			}
			
			if (schemaVersion < 50)
			{
				if (progress != null) progress.display((int)(49*progress_factor), "--Updating schema version: " + schemaVersion + " -> 50...");
				schemaVersion = setSchemaVersionInTransaction(session, 50);
			}
			
			if (schemaVersion < 51)
			{
				if (progress != null) progress.display((int)(50*progress_factor), "--Updating schema version: " + schemaVersion + " -> 51...");
				schemaVersion = updateToSchemaVersion51(session);
			}
			
			// Schemaversion 52-55 only updates the version number
			if (schemaVersion < 55)
			{
				if (progress != null) progress.display((int)(54*progress_factor), "--Updating schema version: " + schemaVersion + " -> 55...");
				schemaVersion = setSchemaVersionInTransaction(session, 55);
			}
			
			if (schemaVersion < 56)
			{
				if (progress != null) progress.display((int)(55*progress_factor), "--Updating schema version: " + schemaVersion + " -> 56...");
				schemaVersion = updateToSchemaVersion56(session);
			}
			
			if (schemaVersion < 57)
			{
				if (progress != null) progress.display((int)(56*progress_factor), "--Updating schema version: " + schemaVersion + " -> 57...");
				schemaVersion = updateToSchemaVersion57(session);
			}
			
			if (schemaVersion < 59)
			{
				if (progress != null) progress.display((int)(58*progress_factor), "--Updating schema version: " + schemaVersion + " -> 59...");
				schemaVersion = setSchemaVersionInTransaction(session, 59);
			}
			
			if (schemaVersion < 60)
			{
				if (progress != null) progress.display((int)(59*progress_factor), "--Updating schema version: " + schemaVersion + " -> 60...");
				schemaVersion = updateToSchemaVersion60(session);
			}
			
			if (schemaVersion < 63)
			{
				if (progress != null) progress.display((int)(62*progress_factor), "--Updating schema version: " + schemaVersion + " -> 63...");
				schemaVersion = setSchemaVersionInTransaction(session, 63);
			}
			// Schemaversion 64-72 only updates the version number
			if (schemaVersion < 72)
			{
				if (progress != null) progress.display((int)(71*progress_factor), "--Updating schema version: " + schemaVersion + " -> 72...");
				schemaVersion = setSchemaVersionInTransaction(session, 72);
			}
			
			sc.logout();
			if (progress != null) progress.display(100, "Database updated successfully.");
		}
		catch (BaseException ex)
		{
			if (progress != null) progress.display(100, "Database update failed: " + ex.getMessage()+"\n");
			throw ex;
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
			Application.stop();
		}
	}
	
	/**
		Get the current schema version number as it is stored in the database.
		@return The schema version number, or 0 if no schem version is found
	*/
	private static int getSchemaVersion(org.hibernate.Session session)
		throws BaseException
	{
		org.hibernate.Transaction tx = null;
		int schemaVersion = 0;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_SCHEMA_VERSION");
			
			SchemaVersionData data = HibernateUtil.loadData(SchemaVersionData.class, query);
			schemaVersion = data == null ? 0 : data.getSchemaVersion();
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("getSchemaVersion: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}

	/**
		Update the database with a new schema version number. This method should
		be called from an open transaction.
		
		@param schemaVersion The new schema version number
	 */
	private static void setSchemaVersion(org.hibernate.Session session, int schemaVersion)
		throws BaseException
	{
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "SET_SCHEMA_VERSION");
		query.setInteger("schemaVersion", schemaVersion);
		query.setInteger("build", Application.getBuild());
		HibernateUtil.executeUpdate(query);
	}
	
	/**
		Update the schema version version number using a new transaction.
		@param schemaVersion The new schema version number
	 */
	private static int setSchemaVersionInTransaction(org.hibernate.Session session, int schemaVersion)
	{
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			setSchemaVersion(session, schemaVersion);
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("setSchemaVersion: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}
	
	/**
	 	Set Jobs.project_id to 0 for all old jobs with a null value in this column.
	 	
		@return The new schema version (=2)
	*/
	private static int updateToSchemaVersion2(org.hibernate.Session session)
		throws BaseException
	{
		final int schemaVersion = 2;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);

			// Set project_id to 0 for all jobs with a null value
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "SET_PROJECT_ID_FOR_JOBS");
			/*
				UPDATE JobData job
				SET job.activeProjectId = 0
				WHERE job.activeProjectId IS NULL
			*/
			HibernateUtil.executeUpdate(query);

			// Update the shcema version number
			setSchemaVersion(session, schemaVersion);

			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion2: OK");
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion2: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}

	/**
	 	Find the unique index on HelpTextData.externalId and remove it.
	 	
		@return The new schema version (=4)
	*/
	private static int updateToSchemaVersion4(org.hibernate.Session session)
		throws BaseException
	{
		final int schemaVersion = 4;
		org.hibernate.Transaction tx = null;
		try
		{
			TableInfo helpInfo = null;
			try
			{
				helpInfo = HibernateUtil.doJdbcWork(session, 
					new TableInfoWork(new Table("HelpTexts")));
			}
			catch (SQLException ex)
			{
				throw new BaseException(ex);
			}

			tx = HibernateUtil.newTransaction(session);
			DbEngine dbEngine = HibernateUtil.getDbEngine();
			String indexName = helpInfo.findIndexName(null, Collections.singleton("external_id"));
			if (indexName != null)
			{
				String sql = dbEngine.getDropIndexSql(null, null, "HelpTexts", indexName, true);
				try
				{
					log.info("Dropping index: " + sql);
					HibernateUtil.doJdbcWork(session, new ExecuteUpdateWork(null, sql));
				}
				catch (SQLException ex)
				{
					throw new BaseException(ex);
				}
			}
	
			// Update the schema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion4: OK");
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion4: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}
	
	/**
	 	Set RawBioAssays.has_data to true or false depending on the number of spots.
	 	Delete all Formulas with affymetrix raw data type
		@return The new schema version (=7)
	*/
	private static int updateToSchemaVersion7(org.hibernate.Session session)
		throws BaseException
	{
		final int schemaVersion = 7;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
	
			// Set has_data for all raw bioassays with a null value
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "SET_HAS_DATA_FOR_RAWBIOASSAYS");
			/*
				UPDATE RawBioAssayData rba
				SET rba.hasData = CASE 
					WHEN rba.spots > 0 THEN true
					ELSE false
					END CASE
				WHERE rba.hasData IS NULL
			*/
			HibernateUtil.executeUpdate(query);
	
			// Delete all affymetrix formulas
			query = HibernateUtil.createQuery(session, "SELECT frm FROM FormulaData frm WHERE frm.rawDataType = 'affymetrix'");
			List<FormulaData> affyFormulas = HibernateUtil.loadList(FormulaData.class, query, null);
			for (FormulaData affyFormula : affyFormulas)
			{
				HibernateUtil.deleteData(session, affyFormula);
			}
			
			// Update the shcema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion7: OK");
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion7: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}	
	
	/**
	 	Load all plugins and call {@link PluginDefinition#loadPluginInformation(String, String, boolean)}
	 	for each one so that the {@link PluginDefinition#supportsConfigurations()} and
	 	{@link PluginDefinition#requiresConfiguration()} return correct values.
		@return The new schema version (=10)
	*/
	private static int updateToSchemaVersion10(SessionControl sc)
		throws BaseException
	{
		final int schemaVersion = 10;
		DbControl dc = null;
		try
		{
			dc = sc.newDbControl();
	
			ItemQuery<PluginDefinition> pluginQuery = PluginDefinition.getQuery();
			pluginQuery.include(Include.MINE, Include.OTHERS);
			
			for (PluginDefinition plugin : pluginQuery.list(dc))
			{
				try
				{
					plugin.loadPluginInformation(plugin.getJarPath(), plugin.getClassName(), false);
				}
				catch (Throwable t)
				{}
			}

			// Update the schema version number
			setSchemaVersion(dc.getHibernateSession(), schemaVersion);
	
			// Commit the changes
			dc.commit();
			log.info("updateToSchemaVersion10: OK");
		}
		catch (BaseException ex)
		{
			log.error("updateToSchemaVersion10: FAILED", ex);
			throw ex;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return schemaVersion;
	}	


	/**
	 	Set JobData.getParameterVersion and PluginConfigurationData.getParameterVersion
	 	to 1 if they have null values. Copy existing info from PluginConfigurationValues
	 	to VersionedPluginConfigurationValues.
		@return The new schema version (=19)
	*/
	private static int updateToSchemaVersion19(org.hibernate.Session session)
		throws BaseException
	{
		final int schemaVersion = 19;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
	
			// Set parameterVersion = 1 for all plugin configurations with a null value
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "SET_PARAMETER_VERSION_FOR_PLUGINCONFIGURATIONS");
			/*
				UPDATE PluginConfigurationData pc
				SET pc.parameterVersion = 1 
				WHERE pc.parameterVersion IS NULL
			*/
			HibernateUtil.executeUpdate(query);
	
			// Set parameterVersion = 1 for all plugin configurations with a null value
			query = HibernateUtil.getPredefinedQuery(session, "SET_PARAMETER_VERSION_FOR_JOBS");
			/*
				UPDATE JobData job
				SET job.parameterVersion = 1 
				WHERE job.parameterVersion IS NULL
			*/
			HibernateUtil.executeUpdate(query);
			
			// Copy existing parameters
			try
			{
				String sql = PredefinedQuery.getQueryString("COPY_PLUGINCONFIGURATION_PARAMETERS");
				/*
					INSERT INTO `VersionedPluginConfigurationValues` 
					(`pluginconfiguration_id`, `name`, `parameter_version`, `value_id`)
					SELECT `pluginconfiguration_id`, `name`, 1, `value_id`
					FROM `PluginConfigurationValues`
				*/
				HibernateUtil.doJdbcWork(session, new ExecuteUpdateWork(null, sql));
			
				sql = PredefinedQuery.getQueryString("DELETE_UNVERSIONED_PLUGINCONFIGURATION_PARAMETERS");
				/*
					DELETE FROM `PluginConfigurationValues`
				*/
				HibernateUtil.doJdbcWork(session, new ExecuteUpdateWork(null, sql));
			}
			catch (SQLException ex)
			{
				throw new BaseException(ex);
			}
			
			// Update the shcema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion19: OK");
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion19: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}	
	
	/**
	 	Move FileValues to ItemValues. Requires setting the ParameterValues.discrimitator
	 	to 10 for all parameter with a value of 9 and copying the contents of the
	 	FileValues table.
	 	
		@return The new schema version (=24)
	*/
	private static int updateToSchemaVersion24(org.hibernate.Session session)
		throws BaseException
	{
		final int schemaVersion = 24;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			try
			{
				// Change discriminator for file values
				String sql = PredefinedQuery.getQueryString("CHANGE_FILE_DISCRIMINATOR");
				/*
					UPDATE `ParameterValues` 
					SET `discriminator` = 10 
					WHERE `discriminator` = 9
				*/
				HibernateUtil.doJdbcWork(session, new ExecuteUpdateWork(null, sql));
			
				sql = PredefinedQuery.getQueryString("COPY_FILE_PARAMETERS");
				/*
					INSERT INTO `ItemValues` (`id`, `data_class`, `data_class_id`)
					SELECT `id`, 'net.sf.basedb.core.data.FileData', `value`
					FROM `FileValues`
				*/
				HibernateUtil.doJdbcWork(session, new ExecuteUpdateWork(null, sql));

				// Delete old file values
				sql = PredefinedQuery.getQueryString("DELETE_FILE_PARAMETERS");
				/*
					DELETE FROM `FileValues`
				*/
				HibernateUtil.doJdbcWork(session, new ExecuteUpdateWork(null, sql));
			}
			catch (SQLException ex)
			{
				throw new BaseException(ex);
			}
			
			// Update the shcema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion24: OK");
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion24: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}

	/**
	 	Set the itemType property on all DiskUsageData items.
		@return The new schema version (=25)
	*/
	private static int updateToSchemaVersion25(org.hibernate.Session session)
		throws BaseException
	{
		final int schemaVersion = 25;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);

			// Query for setting the item type
			org.hibernate.Query duQuery = HibernateUtil.getPredefinedQuery(session, 
				"SET_ITEMTYPE_ON_DISKUSAGE");
			/*
				UPDATE DiskUsageData du
				SET du.itemType = :itemType
				WHERE du.id IN (:itemList)
			*/
			
			// Files
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
				"LOAD_DISKUSAGEID_FOR_FILES");
			/**
				SELECT f.diskUsage.id 
				FROM FileData f
			*/
			List<Integer> ids = HibernateUtil.loadList(Integer.class, query, null);
			if (ids.size() > 0)
			{
				duQuery.setParameter("itemType", Item.FILE.getValue(), org.hibernate.Hibernate.INTEGER);
				duQuery.setParameterList("itemList", ids, org.hibernate.Hibernate.INTEGER);
				HibernateUtil.executeUpdate(duQuery);
			}
			
			// Experiments
			query = HibernateUtil.getPredefinedQuery(session, 
				"LOAD_DISKUSAGEID_FOR_EXPERIMENTS");
			/**
				SELECT xp.diskUsage.id 
				FROM ExperimentData xp
			*/
			ids = HibernateUtil.loadList(Integer.class, query, null);
			if (ids.size() > 0)
			{
				duQuery.setParameter("itemType", Item.EXPERIMENT.getValue(), org.hibernate.Hibernate.INTEGER);
				duQuery.setParameterList("itemList", ids, org.hibernate.Hibernate.INTEGER);
				HibernateUtil.executeUpdate(duQuery);
			}

			// Raw bioassays
			query = HibernateUtil.getPredefinedQuery(session, 
				"LOAD_DISKUSAGEID_FOR_RAWBIOASSAYS");
			/**
				SELECT rba.diskUsage.id 
				FROM RawBioAssayData rba
			*/
			ids = HibernateUtil.loadList(Integer.class, query, null);
			if (ids.size() > 0)
			{
				duQuery.setParameter("itemType", Item.RAWBIOASSAY.getValue(), org.hibernate.Hibernate.INTEGER);
				duQuery.setParameterList("itemList", ids, org.hibernate.Hibernate.INTEGER);
				HibernateUtil.executeUpdate(duQuery);
			}
			
			// Update the shcema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion25: OK");
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion25: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}

	/**
	 	Find the number of raw mappings for the spot with the maximum number
	 	for each data cube
		@return The new schema version (=28)
	*/
	private static int updateToSchemaVersion28(SessionControl sc)
		throws BaseException
	{
		final int schemaVersion = 28;
		DbControl dc = null;
		try
		{
			dc = sc.newDbControl();
			org.hibernate.Session session = dc.getHibernateSession();
			
			// Query for loading all data cubes
			org.hibernate.Query query = HibernateUtil.createQuery(session, 
				"FROM DataCubeData dcd WHERE dcd.maxRawMappingsForSpot = -1");
			org.hibernate.Query updateQuery = HibernateUtil.getPredefinedQuery(session, 
				"SET_MAX_MAPPING_ON_DATACUBE");
			/*
				UPDATE DataCubeData dcd
				SET dcd.maxRawMappingsForSpot = :mapCount
				WHERE dcd.id = :dataCube
			*/
			for (DataCubeData cubeData : HibernateUtil.loadList(DataCubeData.class, query, null))
			{
				DataCube cube = dc.getItem(DataCube.class, cubeData);
				int mapCount = cube.countSpotMappings();
				updateQuery.setInteger("mapCount", mapCount);
				updateQuery.setInteger("dataCube", cube.getId());
				HibernateUtil.executeUpdate(updateQuery);
			}

			// Update the shcema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			dc.commit();
			log.info("updateToSchemaVersion28: OK");
		}
		catch (BaseException ex)
		{
			log.error("updateToSchemaVersion28: FAILED", ex);
			throw ex;
		}
		finally
		{
			if (dc != null) dc.close(); 
		}
		return schemaVersion;
	}

	/**
	 	Vefify and update the remaining quantity of all biomaterials.
		@return The new schema version (=30)
	*/
	private static int updateToSchemaVersion30(org.hibernate.Session session)
		throws BaseException
	{
		final int schemaVersion = 30;
		
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			
			// For temporary storing the new remaining quantities
			Map<MeasuredBioMaterialData, Float> remaining = new HashMap<MeasuredBioMaterialData, Float>(); 
			
			// Load used quantity for all events
			org.hibernate.Query eventQuery = HibernateUtil.getPredefinedQuery(session, 
				"GET_USED_QUANTITY_EVENTS");
				/*			 
					SELECT evt.bioMaterial, evt.usedQuantity
					FROM BioMaterialEventData evt
				*/
			
			// Load used quantity for sources to all events
			org.hibernate.Query sourcesQuery = HibernateUtil.getPredefinedQuery(session, 
				"GET_USED_QUANTITY_SOURCES");
				/*
					SELECT index(src), src.usedQuantity
					FROM BioMaterialEventData evt
					JOIN evt.sources src					
				 */
			List<Object[]> events = HibernateUtil.loadList(Object[].class, eventQuery, null);
			events.addAll(HibernateUtil.loadList(Object[].class, sourcesQuery, null));
			
			// Calculate the remaining quantity for each biomaterial
			// o[0] = biomaterial, o[1] = used quantity			
			for (Object[] o : events)
			{
				MeasuredBioMaterialData bioMaterial = (MeasuredBioMaterialData)o[0];				
				Float usedQuantity = (Float)o[1];				
				if (usedQuantity != null)
				{
					Float currentRemaining = remaining.get(bioMaterial);
					if (currentRemaining == null) currentRemaining = 0.0f;
					currentRemaining = currentRemaining - usedQuantity;
					remaining.put(bioMaterial, currentRemaining);
				}
			}
			
			// Update the remaining quantity
			org.hibernate.Query updateQuery = HibernateUtil.getPredefinedQuery(session, 
				"SET_REMAINING_QUANTITY");
			/*
				UPDATE MeasuredBioMaterialData mbm 
				SET mbm.remainingQuantity = :remain
				WHERE mbm = :bioMaterial
			*/
			for (Map.Entry<MeasuredBioMaterialData, Float> entry : remaining.entrySet())
			{
				MeasuredBioMaterialData bioMaterial = entry.getKey();
				Float remainingQuantity = entry.getValue();
				updateQuery.setEntity("bioMaterial", bioMaterial);
				updateQuery.setFloat("remain", remainingQuantity);
				updateQuery.executeUpdate();
			}
			
			// Update the shcema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion30: OK");
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion30: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}
	
	/**
	 	Set the writeProtected flag on all files with a null value to false.
	 	Set the lastUpdated property on all files with a null value to the 
	 	last modified time of the internal file.
		@return The new schema version (=31)
	*/
	private static int updateToSchemaVersion31(org.hibernate.Session session)
	{
		final int schemaVersion = 31;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
	
			// Set writeProtected to false for all files with a null value
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
				"SET_WRITE_PROTECTED_ON_FILES");
			/*
				UPDATE FileData f
				SET f.writeProtected = false
				WHERE f.writeProtected IS NULL
			*/
			HibernateUtil.executeUpdate(query);
		
			query = HibernateUtil.getPredefinedQuery(session, "GET_INTERNAL_FILES_WITH_NO_LAST_UPDATE");
			/*
				SELECT f
				FROM FileData f
				WHERE f.lastUpdate IS NULL AND NOT f.internalName IS NULL
			*/
			List<FileData> files = HibernateUtil.loadList(FileData.class, query, null);
			for (FileData fData : files)
			{
				java.io.File fio = File.getAbsolutePath(fData.getInternalName());
				if (fio.exists())
				{
					fData.setLastUpdate(new Date(fio.lastModified()));
				}
			}
			
			// Update the shcema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion31: OK");
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion31: FAILED", ex);
			throw ex;
		}
		return schemaVersion;		
	}
	
	/**
	 	Check and update the file size and MD5 on existing file items since
	 	they may be incorrect
		@return The new schema version (=35)
	*/
	private static int updateToSchemaVersion35(org.hibernate.Session session)
	{
		final int schemaVersion = 35;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
	
			// Load all files
			org.hibernate.Query query = HibernateUtil.createQuery(session, 
				"SELECT f FROM FileData f");
			
			List<FileData> files = HibernateUtil.loadList(FileData.class, query, null);
			for (FileData fData : files)
			{
				if (fData.getInternalName() != null)
				{
					java.io.File fio = File.getAbsolutePath(fData.getInternalName());
					if (fio.exists())
					{
						long realSize = fio.length();
						long dbSize = fData.getSize();
						if (realSize != dbSize)
						{
							fData.setSize(realSize);
							fData.setMd5(null);
							fData.getDiskUsage().setBytes(realSize);
						}
					}
				}
			}
			
			// Update the shcema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion35: OK");
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion35: FAILED", ex);
			throw ex;
		}
		return schemaVersion;		
	}
	
	/**
	 	Set the experiment for jobs that belong to a single transformation.
		@return The new schema version (=38)
	*/
	private static int updateToSchemaVersion38(org.hibernate.Session session)
	{
		final int schemaVersion = 38;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
	
			// Load all transformations
			org.hibernate.Query query = HibernateUtil.createQuery(session, 
				"SELECT t FROM TransformationData t");
			
			List<TransformationData> transformations = HibernateUtil.loadList(TransformationData.class, query, null);
			Map<JobData, Set<ExperimentData>> experiments = new HashMap<JobData, Set<ExperimentData>>(); 
			for (TransformationData transformation : transformations)
			{
				JobData job = transformation.getJob();
				if (job != null)
				{
					ExperimentData experiment = transformation.getExperiment();
					if (!experiments.containsKey(job))
					{
						experiments.put(job, new HashSet<ExperimentData>());
					}
					experiments.get(job).add(experiment);
				}
			}
			
			for (Map.Entry<JobData, Set<ExperimentData>> entry : experiments.entrySet())
			{
				JobData job = entry.getKey();
				Set<ExperimentData> jobExperiments = entry.getValue();
				if (jobExperiments.size() == 1)
				{
					job.setExperiment(jobExperiments.iterator().next());
				}
			}
			
			// Update the shcema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion38: OK");
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion38: FAILED", ex);
			throw ex;
		}
		return schemaVersion;		
	}
	/**
		Remove AnyToAny links created by older versions of Base1PluginExecuter
		@return The new schema version (=39)
	*/
	private static int updateToSchemaVersion39(org.hibernate.Session session)
	{
		final int schemaVersion = 39;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
	
			// Load Base1PluginExecuter
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
					"GET_PLUGINDEFINITION_FOR_CLASSNAME");
			/*
				SELECT pd
				FROM PluginDefinitionData pd
				WHERE pd.className = :className
			*/
			query.setString("className", "net.sf.basedb.plugins.Base1PluginExecuter");
			PluginDefinitionData base1Executer = HibernateUtil.loadData(PluginDefinitionData.class, query);
			
			// Load all plugin configurations for net.sf.basedb.plugins.Base1PluginExecuter
			query = HibernateUtil.getPredefinedQuery(session, 
				"GET_PLUGINCONFIGURATIONS_FOR_PLUGINDEFINITION", "pc");
			/*
				SELECT {1}
				FROM PluginConfigurationData pc
				WHERE pc.pluginDefinition = :pluginDefinition
			*/
			query.setEntity("pluginDefinition", base1Executer);
			List<PluginConfigurationData> configurations = 
				HibernateUtil.loadList(PluginConfigurationData.class, query, null);

			// Delete AnyToAny for each configuration
			query = HibernateUtil.getPredefinedQuery(session, "DELETE_ALL_ANYTOANY_FROM");
			/*
				DELETE FROM AnyToAnyData ata 
				WHERE ata.fromId = :fromId AND ata.fromType = :fromType
			*/
			query.setInteger("fromType", Item.PLUGINCONFIGURATION.getValue());
			for (PluginConfigurationData pc : configurations)
			{
				query.setInteger("fromId", pc.getId());
				HibernateUtil.executeUpdate(query);
			}
			
			// Update the schema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion39: OK");
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion39: FAILED", ex);
			throw ex;
		}
		return schemaVersion;		
	}

	/**
		Has been replaced with {@link #updateToSchemaVersion44(SessionControl)}
		@return The new schema version (=42)
	*/
	private static int updateToSchemaVersion42(org.hibernate.Session session)
	{
		final int schemaVersion = 42;
		return schemaVersion;		
	}
	
	/**
		Update property filter on 'guiContexts.itemType' to '$ctx.itemType'
		@return The new schema version (=43)
	*/
	private static int updateToSchemaVersion43(org.hibernate.Session session)
	{
		final int schemaVersion = 43;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
	
			// Change property values -- query must ignore filters
			// that may cause duplicate filters on the same property
			org.hibernate.Query query = HibernateUtil.getPredefinedSQLQuery(session, 
					"UPDATE_PROPERTY_FILTER");
			/*
				UPDATE [PropertyFilters] 
				SET [property] = :newProperty
				WHERE [property] = :oldProperty
				AND [context_id] NOT IN 
				(
				   SELECT [context_id] 
				   FROM [PropertyFilters]
				   WHERE [property] = :newProperty
				)
			*/
			query.setString("oldProperty", "guiContexts.itemType");
			query.setString("newProperty", "$ctx.itemType");
			HibernateUtil.executeUpdate(query);
	
			//Delete any remaining filter with the old property
			query = HibernateUtil.getPredefinedSQLQuery(session, 
				"DELETE_PROPERTY_FILTER");
			/*
				DELETE FROM [PropertyFilters] 
				WHERE [property] = :property
			*/
			query.setString("property", "guiContexts.itemType");
			HibernateUtil.executeUpdate(query);
			
			// Update the schema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion43: OK");
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion43: FAILED", ex);
			throw ex;
		}
		return schemaVersion;		
	}

	/**
		<ul>
		<li>Set a platform for array designs and raw bioassays
		<li>Change experiment raw data type if it is "affymetrix" to "platform.affymetrix"
		<li>Change project default raw data type if it is "affymetrix" to "platform.affymetrix"
		<li>Move CEL and CDF files into FileSet:s
		<li>Remove filters on ArrayDesign.isAffyChip property
		<li>Add READ permission to PLATFORM and DATAFILETYPE for all plug-ins
			working with ARRAYDESIGN or RAWBIOASSAY
		</ul>
	*/
	@SuppressWarnings("deprecation")
	private static int updateToSchemaVersion44(SessionControl sc)
	{
		final int schemaVersion = 44;
		DbControl dc = null;
		try
		{
			dc = sc.newDbControl();
			org.hibernate.Session session = dc.getHibernateSession();
	
			// Items that we need
			Platform affymetrix = Platform.getByExternalId(dc, Platform.AFFYMETRIX);
			Platform generic = Platform.getByExternalId(dc, Platform.GENERIC);
			DataFileType cdfType = DataFileType.getByExternalId(dc, DataFileType.AFFYMETRIX_CDF);
			DataFileType celType = DataFileType.getByExternalId(dc, DataFileType.AFFYMETRIX_CEL);
			RawDataType affyRaw = affymetrix.getRawDataType();
			
			// Query to count number of features in db
			org.hibernate.Query countQuery = HibernateUtil.getPredefinedQuery(session, 
				"COUNT_FEATURES_FOR_ARRAYDESIGN");

			// Load array designs
			org.hibernate.Query designQuery = HibernateUtil.createQuery(session, 
				"SELECT ad FROM ArrayDesignData ad WHERE ad.platform is null");
			List<ArrayDesignData> arrayDesigns = 
				HibernateUtil.loadList(ArrayDesignData.class, designQuery, sc);
			for (ArrayDesignData dd : arrayDesigns)
			{
				ArrayDesign design = dc.getItem(ArrayDesign.class, dd);
				if (design.getData().isAffyChip())
				{
					design.setPlatform(affymetrix);
					AnyToAny anyCdf = null;
					try
					{
						anyCdf = AnyToAny.getByName(dc, design, Affymetrix.CDF_LINK_NAME);
					}
					catch (Throwable t)
					{}
					AnyToAny.unlinkFrom(dc, design, Affymetrix.CDF_LINK_NAME);
					if (anyCdf != null)
					{
						File cdfFile = (File)anyCdf.getTo();
						FileSet fileSet = design.getFileSet();
						fileSet.setMember(cdfFile, cdfType);
						fileSet.validate(dc, true);
					}
				}
				else
				{
					design.setPlatform(generic);
					countQuery.setInteger("arrayDesign", design.getId());
					int dbFeatures = HibernateUtil.loadData(Long.class, countQuery).intValue();
					design.getData().setNumDbFeatures(dbFeatures);
				}
			}
			
			// Load rawbioassays
			org.hibernate.Query rbaQuery = HibernateUtil.createQuery(session, 
				"SELECT rba FROM RawBioAssayData rba WHERE rba.platform is null");
			List<RawBioAssayData> rawBioAssays = 
				HibernateUtil.loadList(RawBioAssayData.class, rbaQuery, sc);
			for (RawBioAssayData dd : rawBioAssays)
			{
				RawBioAssay rba = dc.getItem(RawBioAssay.class, dd);
				if ("affymetrix".equals(rba.getData().getRawDataType()))
				{
					rba.getData().setPlatform(affymetrix.getData());
					rba.getData().setRawDataType(affyRaw.getId());
					rba.getData().setSpots(0); // No db spots in affy files
					AnyToAny anyCel = null;
					try
					{
						anyCel = AnyToAny.getByName(dc, rba, Affymetrix.CEL_LINK_NAME);
					}
					catch (Throwable t)
					{}
					AnyToAny.unlinkFrom(dc, rba, Affymetrix.CEL_LINK_NAME);
					if (anyCel != null)
					{
						File celFile = (File)anyCel.getTo();
						FileSet fileSet = rba.getFileSet();
						fileSet.setMember(celFile, celType);
						fileSet.validate(dc, true);
					}
				}
				else
				{
					rba.getData().setPlatform(generic.getData());
				}
			}
	
			// Change raw data type for 'affymetrix' experiments to 'platform.affymetrix'
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
				"CHANGE_EXPERIMENT_RAWDATATYPE");
			/*
				UPDATE ExperimentData
				SET rawDataType = :newRawDataType
				WHERE rawDataType = :oldRawDataType
			*/
			query.setString("oldRawDataType", "affymetrix");
			query.setString("newRawDataType", affyRaw.getId());
			HibernateUtil.executeUpdate(query);
		
			// Change raw data type for 'affymetrix' projects to 'platform.affymetrix'
			query = HibernateUtil.getPredefinedQuery(session, 
				"GET_PROJECTS_WITH_DEFAULT");
			/**
				SELECT prj
				FROM ProjectData prj
				WHERE prj.projectDefaults[:name] = :value
			*/
			query.setString("name", "raw_data_type");
			query.setString("value", "affymetrix");
			List<ProjectData> projects = HibernateUtil.loadList(ProjectData.class, query, sc);
			for (ProjectData project : projects)
			{
				Project p = dc.getItem(Project.class, project);
				p.setDefaultRawDataType(affyRaw);
				p.setDefaultItem(dc, affymetrix, Project.Default.PLATFORM);
			}
		
			// Update property filters on 'rawDataType=affymetrix' to 'platform.affymetrix'
			query = HibernateUtil.getPredefinedSQLQuery(session, 
				"UPDATE_PROPERTY_FILTER_VALUE");
			/*
				UPDATE [PropertyFilters]
				SET [value] = :newValue
				WHERE [property] = :property AND [value] = :oldValue
			 */
			query.setString("property", "rawDataType");
			query.setString("oldValue", "affymetrix");
			query.setString("newValue", affyRaw.getId());
			HibernateUtil.executeUpdate(query);
			
			// Remove filters on the 'affyChip' property
			query = HibernateUtil.getPredefinedSQLQuery(session, 
				"DELETE_PROPERTY_FILTER");
			/*
				DELETE FROM PropertyFilters pf
				WHERE pf.property = :property
			 */
			query.setString("property", "affyChip");
			HibernateUtil.executeUpdate(query);

			// Add READ permission to PLATFORM and DATAFILETYPE for all plug-ins
			// working with ARRAYDESIGN or RAWBIOASSAY
			int adKeyId = SystemItems.getRoleKeyId(Item.ARRAYDESIGN);
			int rbaKeyId = SystemItems.getRoleKeyId(Item.RAWBIOASSAY);
			
			org.hibernate.Query pQuery = HibernateUtil.getPredefinedQuery(session, 
				"GET_PLUGINS_WITH_KEYS");
			/*
				SELECT DISTINCT pk.pluginDefinitionId
				FROM PluginKeys pk WHERE pk.keyId IN (:keys)
			*/
			pQuery.setParameterList("keys", 
				Arrays.asList( new Integer[] { adKeyId, rbaKeyId }), Type.INT.getHibernateType());
			List<Integer> pluginIds = HibernateUtil.loadList(Integer.class, pQuery, sc);
			for (int pluginId : pluginIds)
			{
				addPluginPermission(session, pluginId, Item.PLATFORM, 0, 2046);
				addPluginPermission(session, pluginId, Item.DATAFILETYPE, 0, 2046);
			}
			
			// Update the schema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			dc.commit();
			log.info("updateToSchemaVersion44: OK");
		}
		catch (BaseException ex)
		{
			log.error("updateToSchemaVersion44: FAILED", ex);
			throw ex;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return schemaVersion;		
	}

	private static void addPluginPermission(org.hibernate.Session session, int pluginId, Item itemType, int granted, int denied)
	{
		PluginKeys pk = new PluginKeys();
		pk.setPluginDefinitionId(pluginId);
		pk.setKeyId(SystemItems.getRoleKeyId(itemType));
		if (session.get(PluginKeys.class, pk) == null)
		{
			pk.setGranted(granted);
			pk.setDenied(denied);
			session.save(pk);
		}
	}
	
	/**
		Update useInternalJobQueue for plug-ins.
		@return The new schema version (=45)
	*/
	private static int updateToSchemaVersion45(org.hibernate.Session session)
	{
		final int schemaVersion = 45;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
	
			// Change property values
			org.hibernate.Query query = HibernateUtil.createQuery(session, 
				"SELECT DISTINCT jas.pluginDefinition FROM JobAgentSettingsData jas");
			List<PluginDefinitionData> plugins = 
				HibernateUtil.loadList(PluginDefinitionData.class, query, null);
			
			for (PluginDefinitionData pd : plugins)
			{
				pd.setUseInternalJobQueue(false);
			}
	
			// Update the schema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion45: OK");
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion45: FAILED", ex);
			throw ex;
		}
		return schemaVersion;		
	}
	
	/**
	 	Find the unique index on Feature (row, column and block) and remove it.
	 	
		@return The new schema version (=49)
	*/
	private static int updateToSchemaVersion49(org.hibernate.Session session)
		throws BaseException
	{
		final int schemaVersion = 49;
		org.hibernate.Transaction tx = null;
		try
		{
			TableInfo info = null;
			try
			{
				info = HibernateUtil.doJdbcWork(session, 
					new TableInfoWork(new Table("Features")));
			}
			catch (SQLException ex)
			{
				throw new BaseException(ex);
			}
	
			tx = HibernateUtil.newTransaction(session);
			DbEngine dbEngine = HibernateUtil.getDbEngine();
			String indexName = info.findIndexName("uniquecoordinate", 
				new HashSet<String>(Arrays.asList(new String[] {"arraydesignblock_id", "row", "column"})));
			if (indexName != null)
			{
				String sql = dbEngine.getDropIndexSql(null, null, "Features", indexName, true);
				try
				{
					log.info("Dropping index: " + sql);
					HibernateUtil.doJdbcWork(session, new ExecuteUpdateWork(null, sql));
				}
				catch (SQLException ex)
				{
					throw new BaseException(ex);
				}
			}
	
			// Update the schema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion49: OK");
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion49: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}	
	
	/**
	 	Change all plug-ins with an absolute JAR path to relative 
	 	paths.
	 	
		@return The new schema version (=51)
	*/
	private static int updateToSchemaVersion51(org.hibernate.Session session)
		throws BaseException
	{
		final int schemaVersion = 51;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			
			// Change property values
			org.hibernate.Query query = HibernateUtil.createQuery(session, 
				"SELECT pd FROM PluginDefinitionData pd WHERE NOT pd.jarPath IS NULL");
			List<PluginDefinitionData> plugins = 
				HibernateUtil.loadList(PluginDefinitionData.class, query, null);
			
			for (PluginDefinitionData pd : plugins)
			{
				pd.setJarPath(PluginDefinition.convertToRelative(pd.getJarPath()));
			}
			
			// Update the schema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion51: OK");
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion51: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}
	
	/**
	 	Count the number of reporters in existing reporter lists.
	 	
		@return The new schema version (=56)
	*/
	private static int updateToSchemaVersion56(org.hibernate.Session session)
		throws BaseException
	{
		final int schemaVersion = 56;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			
			// Change property values
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
				"SET_REPORTERLIST_SIZE");
			/*
				UPDATE ReporterListData rl
				SET rl.size = size(rl.reporterListScores)
				WHERE rl.size IS NULL
			*/
			HibernateUtil.executeUpdate(query);
			
			// Update the schema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion56: OK");
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion56: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}
	
	/**
	 	Grant {@link Permission#SELECT_JOBAGENT} permission to
	 	administrators, power users and job agent roles.
	 	
		@return The new schema version (=57)
	*/
	private static int updateToSchemaVersion57(org.hibernate.Session session)
		throws BaseException
	{
		final int schemaVersion = 57;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			
			int selectPermission = Permission.SELECT_JOBAGENT.grantValue();
			
			RoleData admin = HibernateUtil.loadData(session, RoleData.class, SystemItems.getId(Role.ADMINISTRATOR));
			RoleData power = HibernateUtil.loadData(session, RoleData.class, SystemItems.getId(Role.POWER_USER));
			RoleData agent = HibernateUtil.loadData(session, RoleData.class, SystemItems.getId(Role.JOBAGENT));
			RoleKeyData systemKey = HibernateUtil.loadData(session, RoleKeyData.class, 
					SystemItems.getRoleKeyId(Item.SYSTEM));
			
			Map<RoleData, Integer> rolePermissions = systemKey.getRoles();
			rolePermissions.put(admin, Permission.merge(selectPermission, rolePermissions.get(admin)));
			rolePermissions.put(power, Permission.merge(selectPermission, rolePermissions.get(power)));
			rolePermissions.put(agent, Permission.merge(selectPermission, rolePermissions.get(agent)));

			// Update the schema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion57: OK");
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion57: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}
	
	/**
	 	Update incorrect quota values for experiments.
	 	
		@return The new schema version (=60)
	*/
	private static int updateToSchemaVersion60(org.hibernate.Session session)
		throws BaseException
	{
		final int schemaVersion = 60;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			
			org.hibernate.Query query = HibernateUtil.createQuery(session, 
				"SELECT exp FROM ExperimentData exp");
			List<ExperimentData> experiments = 
				HibernateUtil.loadList(ExperimentData.class, query, null);
			for (ExperimentData exp : experiments)
			{
				exp.getDiskUsage().setBytes(exp.getBytes());
			}

			// Update the schema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion60: OK");
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion60: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}
	
	/**
	 	Adjust the existing items in the database to be compatible with the latest mappings.
	 	No update to the schema version should be done here. The code must also  
	 	consider the case that the update failes at a later stage and that the same
	 	update is executed again.
	 	
	  	@param update FALSE if it is an installation. TRUE if it is an update.
	  	@param progress An object implementing the {@link ProgressReporter}
	 		interface
	  	@param rootLogin The root user login
	  	@param rootPassword The root user password
	  	@throws BaseException
	*/
	public static synchronized void adjustExistingItems(boolean update, ProgressReporter progress, String rootLogin, String rootPassword)
		throws BaseException
	{
		if (update)
		{
			org.hibernate.Transaction tx = null;
			org.hibernate.Session session = null;
			try
			{
				Application.start(false, false, false);
	
				// Test root user account
				SessionControl sc = Application.newSessionControl(null, null, null);
				sc.login(rootLogin, rootPassword, null, false);
				if (sc.getLoggedInUserId() != SystemItems.getId(User.ROOT))
				{
					throw new PermissionDeniedException("User '" + rootLogin + "' is not the root account.");
				}
				
				session = HibernateUtil.newSession();
				int schemaVersion = getSchemaVersion(session);
				tx = HibernateUtil.newTransaction(session);
				
				//Update the items in Hardware types.
				if (schemaVersion < 9)
				{
					// Set removed to FALSE for all Hardware types with a null value
					org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "SET_REMOVED_FOR_HARDWARETYPES");
					/*
						UPDATE HardwareTypeData hwd
						SET hwd.removed = false
						WHERE hwd.removed IS NULL
					*/
					HibernateUtil.executeUpdate(query);
					
					//Correct a misspelt system_id for hardware type 'Hybridization station' in the latest schema_version
					query = HibernateUtil.createQuery(session, "UPDATE HardwareTypeData hwd SET hwd.systemId='net.sf.basedb.core.HardwareType.HYBRIDIZATION_STATION' WHERE systemId = 'net.sf.basedb.core.HarwareType.HYBRIDIZATION_STATION'");
					HibernateUtil.executeUpdate(query);
				}
				
				if (schemaVersion < 10)
				{
					// Set supportsConfiguration to true and requiresConfiguration to false for all plugins with null values
					org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "SET_CONFIG_FOR_PLUGINS");
					/*
						UPDATE PluginDefinitionData pd
						SET pd.supportsConfiguration = true, pd.requiresConfiguration = false
						WHERE pd.supportsConfiguration IS NULL
					*/
					HibernateUtil.executeUpdate(query);
				}
				
				if (schemaVersion < 12)
				{
					// Flag incorrect plate mappings as removed and change the name
					org.hibernate.Query pmQuery = HibernateUtil.getPredefinedQuery(session, "GET_INCORRECT_PLATEMAPPINGS");
					/*
						SELECT pm
						FROM PlateMappingData pm
						WHERE pm.sourceCount = 4 AND pm.destinationCount = 1
						AND 
						(
							(pm.sourceGeometry.rows = 8 AND pm.sourceGeometry.columns = 12)
							OR
							(pm.sourceGeometry.rows = 16 AND pm.sourceGeometry.columns = 24)
						)
					*/
					List<PlateMappingData> mappings = HibernateUtil.loadList(PlateMappingData.class, pmQuery, null);
					if (mappings != null && mappings.size() > 0)
					{
						progress.append("\n");
						org.hibernate.Query pQuery = HibernateUtil.getPredefinedQuery(session, "GET_PLATES_WITH_MAPPING");
						/*
							SELECT p
							FROM PlateData p
							WHERE p.plateMapping = :plateMapping
						*/
						String message = "The update has found mappings installed by an earlier release \n"+
							"that are incorrect. Plates that have used these mappings doesn't have the correct \n"+
							"reporter information in the wells. You will have to recreate you plates with the \n"+
							"correct mappings. Any data, child plates, array designs, raw data, etc. that \n"+
							"depends on the incorrect plates are also incorrect and have to be recreated.";
						
						for (PlateMappingData mapping : mappings)
						{
							pQuery.setEntity("plateMapping", mapping);
							List<PlateData> plates = HibernateUtil.loadList(PlateData.class, pQuery, null);
							if (plates != null && plates.size() > 0)
							{
								// Plates with the mapping was found
								progress.append("Incorrect plate mapping found: " + mapping.getName() + "[id="+mapping.getId()+"]; renaming and marking for deletion\n");
								mapping.setRemoved(true);
								mapping.setName("INCORRECT: "+mapping.getName());
								mapping.setDescription(message);
								
								for (PlateData plate : plates)
								{
									progress.append("Plate using incorrect mapping found: " + plate.getName() + "[id="+plate.getId()+"]; renaming and marking for deletion\n");
									plate.setRemoved(true);
									plate.setName("CREATED FROM INCORRECT MAPPING: " + plate.getName());
									plate.setDescription(message);
								}
							}
							else
							{
								progress.append("Incorrect plate mapping found: " + mapping.getName() + "[id="+mapping.getId()+"]; it was not used so it was deleted\n");
								// Delete the mapping
								session.delete(mapping);
							}
						}
						progress.append(message+"\n");
					}
				}
				
				if (schemaVersion < 13)
				{
					// Set the PluginDefinitionData.isTrusted property to TRUE for all plugins with a null value
					org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "SET_TRUSTED_FOR_PLUGINDEFINITIONS");
					/*
						UPDATE PluginDefinitionData pd
						SET pd.trusted = true
						WHERE pd.trusted IS NULL
					*/
					HibernateUtil.executeUpdate(query);
				}
				
				if (schemaVersion < 14)
				{
					// Set the PluginDefinitionData.usePermissions property to FALSE for all plugins with a null value.
					org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "SET_USE_PERMISSION_FOR_PLUGINDEFINITIONS");
					/*
					  	UPDATE PluginDefinitionData pd
					  	SET pd.usePermissions = false
					  	WHERE pd.usePermissions IS NULL
					*/
					HibernateUtil.executeUpdate(query);
				}
				
				if (schemaVersion < 15)
				{
					// Set the JobData.sendMessage property to TRUE for all jobs with a null value.
					org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "SET_SENDMESSAGE_FOR_JOBS");
					/*
					  	UPDATE JobData job
					  	SET job.sendMessage = true
					  	WHERE job.sendMessage IS NULL
					*/
					HibernateUtil.executeUpdate(query);
				}
			
				if (schemaVersion < 20)
				{
					// Set coloring info to FALSE for all formulas
					org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "SET_COLORING_FOR_FORMULAS");
					/*
					  	UPDATE FormulaData frm
					  	SET 
					  		frm.coloring.usingColors = false,
					  		frm.coloring.logarithmic = false
					  	WHERE frm.coloring.usingColors IS NULL
					*/
					HibernateUtil.executeUpdate(query);

					// Set coloring info to FALSE for all formulas
					query = HibernateUtil.getPredefinedQuery(session, "SET_COLORING_FOR_EXTRAVALUETYPES");
					/*
					  	UPDATE ExtraValueTypeData evt
					  	SET 
					  		evt.coloring.usingColors = false,
					  		evt.coloring.logarithmic = false
					  	WHERE evt.coloring.usingColors IS NULL
					*/
					HibernateUtil.executeUpdate(query);
				}
				
				if (schemaVersion < 21)
				{
					// Set immediate execution flag for all plugins with a null value
					org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
							"SET_IMMEDIATE_EXECUTION_FOR_PLUGINS");
					/*
						UPDATE PluginDefinitionData plg
						SET plg.allowImmediateExecution = false
						WHERE plg.allowImmediateExecution IS NULL
					*/
					HibernateUtil.executeUpdate(query);
				}
				
				if (schemaVersion < 22)
				{
					// Set isProtocolParameter to false for all existing annotation types
					org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
							"SET_PROTOCOL_PARAMETER_FOR_ANNOTATION_TYPES");
					/*
						UPDATE AnnotationTypeData at
						SET at.protocolParameter = false
						WHERE at.protocolParameter IS NULL
					*/
					HibernateUtil.executeUpdate(query);
				}
			
				if (schemaVersion < 28)
				{
					// Set maxRawMappingsForSpot to -1 for all data cubes with a null value
					org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
							"SET_MAX_MAPPING_ON_DATACUBES");
					/*
						UPDATE DataCubeData dcd
						SET dcd.maxRawMappingsForSpot = -1
						WHERE dcd.maxRawMappingsForSpot IS NULL
					*/
					HibernateUtil.executeUpdate(query);
				}
				
				if (schemaVersion < 32)
				{
					// Set isDefault to false for groups with a null value
					org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
						"SET_DEFAULT_ON_GROUPS");
					/*
						UPDATE GroupData grp
						SET grp.default = false
						WHERE grp.default IS NULL
					*/
					HibernateUtil.executeUpdate(query);
					
					// Set isDefault to false for roles with a null value
					query = HibernateUtil.getPredefinedQuery(session, 
						"SET_DEFAULT_ON_ROLES");
					/*
						UPDATE RoleData rle
						SET rle.default = false
						WHERE rle.default IS NULL
					*/
					HibernateUtil.executeUpdate(query);
				}
				
				if (schemaVersion < 33)
				{
					// Set delete_job to false for jobs with null value
					org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
							"SET_REMOVE_JOB_ON_JOBS");
					/*
						UPDATE JobData j
						SET j.removeJobWhenFinished = false
						WHERE j.removeJobWhenFinished IS NULL
					*/
					HibernateUtil.executeUpdate(query);
				}
			
				if (schemaVersion < 34)
				{
					// Set item_id to 0 for contexts with null value
					org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
						"SET_ITEM_ID_ON_CONTEXTS");
					/*
						UPDATE ContextData ctx
						SET ctx.itemId = 0
						WHERE ctx.itemId IS NULL
					*/
					HibernateUtil.executeUpdate(query);
				}
				
				if (schemaVersion < 36)
				{
					// Set averageMethod to arithmetic mean (=2) for all formulas with a null value
					org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
						"SET_AVERAGE_METHOD_ON_FORMULAS");
					/*
						UPDATE FormulaData f
						SET f.averageMethod = 2
						WHERE f.averageMethod IS NULL
					*/
					HibernateUtil.executeUpdate(query);
					
					// Set averageMethod to arithmetic mean (=2) for all extravaluetypes with a null value
					query = HibernateUtil.getPredefinedQuery(session, 
						"SET_AVERAGE_METHOD_ON_EXTRAVALUETYPES");
					/*
						UPDATE ExtraValueTypeData evt
						SET evt.averageMethod = 2
						WHERE evt.averageMethod IS NULL
					*/
					HibernateUtil.executeUpdate(query);
				}
				
				if (schemaVersion < 41)
				{
					// Set autoCompress to false for all MIME types with a null value
					org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
						"SET_AUTOCOMPRESS_ON_MIMETYPES");
					/*
						UPDATE MimeTypeData mt
						SET mt.autoCompress = false
						WHERE mt.autoCompress IS NULL
					*/
					HibernateUtil.executeUpdate(query);
					
					// Set autoCompress to false for all directories types with a null value
					query = HibernateUtil.getPredefinedQuery(session, 
						"SET_AUTOCOMPRESS_ON_DIRECTORIES");
					/*
						UPDATE DirectoryData dir
						SET dir.autoCompress = false
						WHERE dir.autoCompress IS NULL
					*/
					HibernateUtil.executeUpdate(query);
					
					// Set compressed to false and compressedSize to size 
					// for all files with a null value
					query = HibernateUtil.getPredefinedQuery(session, 
						"SET_COMPRESSED_ON_FILES");
					/*
						UPDATE FileData f
						SET f.compressed = false,
						SET f.compressedSize = f.size
						WHERE f.compressed IS NULL
					*/
					HibernateUtil.executeUpdate(query);
				}
				
				if (schemaVersion < 44)
				{
					// Set numDbFeatures and numFileFeatures to 0 for all ArrayDesign
					// Later we will count or check the CDF file.
					org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
						"SET_NUMFEATURES_ON_ARRAYDESIGNS");
					/*
						UPDATE ArrayDesignData ad
						SET ad.numDbFeatures = 0, ad.numFileFeatures = 0
						WHERE ad.numDbFeatures IS NULL OR ad.numFileFeatures IS NULL
					*/
					HibernateUtil.executeUpdate(query);

					// Set numFileSpots to 0 for all RawBioassays
					// Later we will count or check the CEL file.
					query = HibernateUtil.getPredefinedQuery(session, 
						"SET_NUMFILESPOTS_ON_RAWBIOASSAY");
					/*
						UPDATE RawBioAssayData rbd
						SET rbd.numFileSpots = 0
						WHERE rbd.numFileSpots IS NULL
					*/
					HibernateUtil.executeUpdate(query);
				}
			
				if (schemaVersion < 45)
				{
					// Set useInternalJobQueue to true for all plug-ins
					org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
						"SET_USEJOBQUEUE_ON_PLUGINS");
					/*
						UPDATE PluginDefinitionData pld
						SET pld.useInternalJobQueue = true
						WHERE pld.useInternalJobQueue IS NULL
					*/
					HibernateUtil.executeUpdate(query);
				}
				
				if (schemaVersion < 49)
				{
					// Set featureIdentificationMethod to COORDINATE for all array designs
					org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
						"SET_FEATUREIDENTIFICATIONMETHOD_ON_ARRAYDESIGNS");
					/*
						UPDATE ArrayDesignData ad
						SET ad.featureIdentificationMethod = :method
						WHERE ad.featureIdentificationMethod IS NULL
					*/
					query.setInteger("method", FeatureIdentificationMethod.COORDINATES.getValue());
					HibernateUtil.executeUpdate(query);
				}
				
				if (schemaVersion < 52)
				{
					// Set numArrays = 1 on all array designs
					org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
						"SET_NUMARRAYS_ON_ARRAYDESIGNS");
					/*
						UPDATE ArrayDesignData ad
						SET ad.numArrays = :numArrays
						WHERE ad.numArrays IS NULL
					*/
					query.setInteger("numArrays", 1);
					HibernateUtil.executeUpdate(query);

					// Set numArrays = 1 on all hybridizations
					query = HibernateUtil.getPredefinedQuery(session, 
						"SET_NUMARRAYS_ON_HYBRIDIZATIONS");
					/*
						UPDATE HybridizationData hyb
						SET hyb.numArrays = :numArrays
						WHERE hyb.numArrays IS NULL
					*/
					query.setInteger("numArrays", 1);
					HibernateUtil.executeUpdate(query);
					
					// Set arrayNum = 1 on all raw bioassays
					query = HibernateUtil.getPredefinedQuery(session, 
						"SET_ARRAYNUM_ON_RAWBIOASSAYS");
					/*
						UPDATE RawBioAssayData rba
						SET rba.arrayNum = :arrayNum
						WHERE rba.arrayNum IS NULL
					*/
					query.setInteger("arrayNum", 1);
					HibernateUtil.executeUpdate(query);
				}
				
				if (schemaVersion < 53)
				{
					// Set featureDiameter = 0 on all array design blocks
					org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
						"SET_FEATUREDIAMETER_ON_ARRAYDESIGNBLOCKS");
					/*
						UPDATE ArrayDesignBlockData adb
						SET adb.featureDiameter = :featureDiameter
						WHERE ad.featureDiameter IS NULL
					*/
					query.setInteger("featureDiameter", 0);
					HibernateUtil.executeUpdate(query);
				}
				
				if (schemaVersion < 58)
				{
					// Set autoShare = false on all directories
					org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
						"SET_AUTOSHARE_ON_DIRECTORIES");
					/*
						UPDATE DirectoryData dir
						SET dir.autoShare = :autoShare
						WHERE dir.autoShare IS NULL
					*/
					query.setBoolean("autoShare", false);
					HibernateUtil.executeUpdate(query);
				}
				
				if (schemaVersion < 59)
				{
					// Set numFileSpots and numFileReporters to 0 on all bioassay sets
					org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
						"SET_NUMFILEVALUES_ON_BIOASSAYSETS");
					/*
						UPDATE BioAssaySetData bas
						SET bas.numFileReporters = 0, bas.numFileSpots = 0
						WHERE bas.numFileReporters IS NULL OR bas.numFileSpots IS NULL
					*/
					HibernateUtil.executeUpdate(query);
					
					// Set numFileSpots to 0 on all bioassays
					query = HibernateUtil.getPredefinedQuery(session, 
						"SET_NUMFILEVALUES_ON_BIOASSAYS");
					/*
						UPDATE BioAssayData ba
						SET ba.numFileSpots = 0
						WHERE ba.numFileSpots IS NULL
					*/
					HibernateUtil.executeUpdate(query);
					
					// Set numFileValues to 0 on all extra values
					query = HibernateUtil.getPredefinedQuery(session, 
						"SET_NUMFILEVALUES_ON_EXTRAVALUES");
					/*
						UPDATE ExtraValueData ev
						SET ev.numFileValues = 0
						WHERE ev.numFileValues IS NULL
					*/
					HibernateUtil.executeUpdate(query);
				}

				if (schemaVersion < 63)
				{
					// Set dryRun = false on all jobs
					org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
						"SET_DRYRUN_ON_JOBS");
					/*
						UPDATE JobData job
						SET job.dryRun = false
						WHERE job.dryRun IS NULL
					*/
					HibernateUtil.executeUpdate(query);
				}
				
				if (schemaVersion < 68)
				{
					// Set autoPermission=31 on all projects
					org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
						"SET_AUTOPERMISSION_ON_PROJECTS");
					/*
						UPDATE ProjectData prj
						SET prj.autoPermission = :permission
						WHERE prj.autoPermission IS NULL
					*/
					query.setInteger("permission", Permission.grant(Permission.READ, 
							Permission.USE, Permission.WRITE, Permission.DELETE));
					HibernateUtil.executeUpdate(query);
				}
				
				if (schemaVersion < 70)
				{
					// Set intensityTransform=0 on all bioassay sets
					org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
						"SET_INTENSITYTRANSFORM_ON_BIOASSAYSETS");
					/*
						UPDATE BioAssaySetData bas
						SET bas.intensityTransform = :transform
						WHERE bas.intensityTransform IS NULL
					*/
					query.setInteger("transform", 0);
					HibernateUtil.executeUpdate(query);
				}
				
				//	Commit the changes
				HibernateUtil.commit(tx);
				log.info("adjustExistingItems: OK");					
			}
			catch (BaseException ex)
			{
				if (tx != null) HibernateUtil.rollback(tx);
				if (progress != null) progress.display(100, "The adjustment of the existing items failed: " + ex.getMessage()+"\n");
				log.info("adjustExistingItems: FAILED");
				throw ex;
			}
			finally
			{
				if (session != null) HibernateUtil.close(session);
				Application.stop();
			}
		}
	}
}
