/*
	$Id: BioMaterialEventData.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.data;

import java.util.Date;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;

/**
	This class is used to register events for measured biomaterials.

	@author Nicklas
	@version 2.0
	@see net.sf.basedb.core.BioMaterialEvent
	@see <a href="../../../../../../html/developerdoc/api/data_api.biomaterials.html">Biomaterials overview</a>
	@base.modified $Date: 2009-04-06 14:52:39 +0200 (Mon, 06 Apr 2009) $
	@hibernate.class table="`BioMaterialEvents`" lazy="false"
*/
public class BioMaterialEventData
	extends BasicData
	implements RegisteredData
{

	public BioMaterialEventData()
	{}
	
	/*
		From the RegisteredData interface
		-------------------------------------------
	*/
	private Date entryDate;
	public Date getEntryDate()
	{
		return entryDate;
	}
	public void setEntryDate(Date entryDate)
	{
		this.entryDate = entryDate;
	}
	// -------------------------------------------
	
	private int eventType;
	/**
		Get the type of event, 1=creation event, 2=hybridization event, 3=other event
		@hibernate.property column="`event_type`" type="int" not-null="true" update="false"
	*/
	public int getEventType()
	{
		return eventType;
	}
	public void setEventType(int eventType)
	{
		this.eventType = eventType;
	}

	private MeasuredBioMaterialData bioMaterial;
	/**
		Get the biomaterial produced or affected by this event.
		@hibernate.many-to-one column="`biomaterial_id`" not-null="false" outer-join="false" update="false"
	*/
	public MeasuredBioMaterialData getBioMaterial()
	{
		return bioMaterial;
	}
	public void setBioMaterial(MeasuredBioMaterialData bioMaterial)
	{
		this.bioMaterial = bioMaterial;
	}
	
	private Float usedQuantity;
	/**
		Get the used quantity of the biomaterial affected by this event.
		This can be both positive and negative.
		@hibernate.property column="`used_quantity`" type="float" not-null="false" 
	*/
	public Float getUsedQuantity()
	{
		return usedQuantity;
	}
	public void setUsedQuantity(Float usedQuantity)
	{
		this.usedQuantity = usedQuantity;
	}
	
	
	private HybridizationData hybridization;
	/**
		Get the hybridization that was created by this event.
		@hibernate.many-to-one column="`hybridization_id`" not-null="false" outer-join="false" update="false"
	*/
	public HybridizationData getHybridization()
	{
		return hybridization;
	}
	public void setHybridization(HybridizationData hybridization)
	{
		this.hybridization = hybridization;
	}

	private UserData user;
	/**
		Get the user that is responsible for this event, typically this is the user that
		entered the information in the database.
		@hibernate.many-to-one column="`user_id`" not-null="false" outer-join="false" update="false"
	*/
	public UserData getUser()
	{
		return user;
	}
	public void setUser(UserData user)
	{
		this.user = user;
	}

	private ProtocolData protocol;
	/**
		Get the protocol used in this event.
		@hibernate.many-to-one column="`protocol_id`" not-null="false" outer-join="false"
	*/
	public ProtocolData getProtocol()
	{
		return protocol;
	}
	public void setProtocol(ProtocolData protocol)
	{
		this.protocol = protocol;
	}

	private HardwareData hardware;
	/**
	 	Get the hardware used in this event
	 	@hibernate.many-to-one column="`hardware_id`" not-null="false" outer-join="false"
	*/
	public HardwareData getHardware()
	{
		return this.hardware;
	}
	public void setHardware(HardwareData hardware)
	{
		this.hardware = hardware;
	}
	
	private Date eventDate;
	/**
		Get the date this event happened in the lab.
		@hibernate.property column="`event_date`" type="date" not-null="false"
	*/
	public Date getEventDate()
	{
		return eventDate;
	}
	public void setEventDate(Date eventDate)
	{
		this.eventDate = eventDate;
	}

	/**
		The maximum length of the comment about this event.
	*/
	public static final int MAX_COMMENT_LENGTH = 65535;
	private String comment;
	/**
		Get the comment about this event.
		@hibernate.property column="`comment`" type="text" not-null="false"
	*/
	public String getComment()
	{
		return comment;
	}
	public void setComment(String comment)
	{
		this.comment = comment;
	}
	
	private Map<MeasuredBioMaterialData, UsedQuantity> sources;
	/**
		Get a map containing the source biomaterials and the used quantity for
		this event.
		@hibernate.map table="`BioMaterialEventSources`" lazy="true" cascade="delete"
		@hibernate.collection-key column="`event_id`"
		@hibernate.index-many-to-many column="`biomaterial_id`" class="net.sf.basedb.core.data.MeasuredBioMaterialData"
		@hibernate.collection-composite-element class="net.sf.basedb.core.data.UsedQuantity"
	*/
	public Map<MeasuredBioMaterialData, UsedQuantity> getSources()
	{
		if (sources == null) sources = new HashMap<MeasuredBioMaterialData, UsedQuantity>();
		return sources;
	}
	void setSources(Map<MeasuredBioMaterialData, UsedQuantity> sources)
	{
		this.sources = sources;
	}
	
	private Set<MeasuredBioMaterialData> sourceBioMaterials;
	/**
		This is the inverse end. The children is the non-pooled biomaterials
		that has this item as a parent.
		@see MeasuredBioMaterialData#getParent()
		@since 2.9
		@hibernate.set table="`BioMaterialEventSources`" inverse="true" lazy="true"
		@hibernate.collection-key column="`event_id`"
		@hibernate.collection-many-to-many column="`biomaterial_id`" class="net.sf.basedb.core.data.MeasuredBioMaterialData"
	*/
	Set<MeasuredBioMaterialData> getSourceBioMaterials()
	{
		return sourceBioMaterials;
	}
	void setSourceBioMaterials(Set<MeasuredBioMaterialData> sourceBioMaterials)
	{
		this.sourceBioMaterials = sourceBioMaterials;
	}

}
