/*
	$Id: ParameterValueData.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.data;

import java.util.Arrays;
import java.util.List;

/**
	The base class for the different types of parameter values.
	@author Nicklas, Samuel
	@version 2.0
	@see <a href="../../../../../../html/developerdoc/api/data_api.parameters.html">Parameters overview</a>
	@hibernate.class table="`ParameterValues`" discriminator-value="-1" lazy="true" batch-size="10"
	@hibernate.discriminator column="`discriminator`" type="int"
	@base.modified $Date: 2009-04-06 14:52:39 +0200 (Mon, 06 Apr 2009) $
*/
public abstract class ParameterValueData<T>
	extends BasicData
{
	ParameterValueData()
	{}
	
	ParameterValueData(T... values)
	{
		replaceValues(Arrays.asList(values));
	}

	/**
		The maximum length of the label of the parameter that can be
		stored in the database.
		@see #setLabel(String)
	*/
	public static final int MAX_LABEL_LENGTH = 255;
	private String label;
	/**
		Get the label of the parameter.
		@hibernate.property column="`label`" type="string" size="255" not-null="false"
	*/
	public String getLabel()
	{
		return label;
	}
	public void setLabel(String label)
	{
		this.label = label;
	}
	
	/**
		The maximum length of the description of the parameter that can be
		stored in the database.
		@see #setDescription(String)
	*/
	public static final int MAX_DESCRIPTION_LENGTH = 65535;
	private String description;
	/**
		Get the description of the parameter.
		@hibernate.property column="`description`" type="text" not-null="false"
	*/
	public String getDescription()
	{
		return description;
	}
	public void setDescription(String description)
	{
		this.description = description;
	}
	
	/**
		Get values of this parameter. Hibernate mapped in each subclass 
		to a bag containing all values for the parameter.
		@return Collection of values.
	*/
	public abstract List<T> getValues();
	
	/**
		Set values of this parameter.
		@param values Collection of values.
	*/
	abstract void setValues(List<T> values);
	
	/**
		Replace the values in the database with the new
		values in the specified list. It is expected that
		the list contains only objects of the correct type.
	*/
	@SuppressWarnings("unchecked")
	public void replaceValues(List<?> values)
	{
		List<T> current = getValues();
		current.clear();
		current.addAll((List<T>) values);
	}
	
	/**
		Replace the current list with a single new value.
		It is expected that the value is of the correct type.
	*/
	@SuppressWarnings("unchecked")
	public void setSingleValue(Object value)
	{
		List<T> current = getValues();
		current.clear();
		current.add((T) value);
	}
}
