/**
	$Id: ParameterValuesWrapper.java 4516 2008-09-11 20:09:17Z jari $

	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.plugin;

import java.util.List;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.ParameterType;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.PluginConfiguration;
import net.sf.basedb.core.PluginDefinition;

/**
	A helper class that will wrap the parameter values from one job, one configuration
	and one request object. The main use is to simplify reading parameter values that
	may be located at different places. The {@link #getValue(String)} and {@link #getValues(String)}
	methods will first check the request, if no parameter is present the job will be checked
	and at last the configuration.
	<p>
	The {@link #setValue(String, ParameterType, Object)} and 
	{@link #setValues(String, ParameterType, List)}
	can be configured to write through to either the job or the configuration.

	@author nicklas
	@version 2.4
	@base.modified $Date: 2008-09-11 22:09:17 +0200 (Thu, 11 Sep 2008) $
*/
public class ParameterValuesWrapper
	implements ParameterValues
{
	
	private final Request request;
	private final ParameterValues job;
	private final ParameterValues config;
	private final boolean preferJob;
	
	public ParameterValuesWrapper(Request request, ParameterValues job, 
		ParameterValues config, boolean preferJob)
	{
		this.request = request;
		this.job = job;
		this.config = config;
		this.preferJob = preferJob;
	}
	
	/**
		Get the ID of either the job or the configuration, depending in the
		<code>preferJob</code> parameter that was set when creating this object.
		@return The ID of the job or configuration or 0 if there is none
	*/
	public int getId()
	{
		if (preferJob && job != null)
		{
			return job.getId();
		}
		else if (config != null)
		{
			return config.getId();
		}
		return 0;
	}

	public Job getJob()
	{
		return job == null ? null : job.getJob();
	}

	public PluginConfiguration getPluginConfiguration()
	{
		return config == null ? null : config.getPluginConfiguration();
	}

	/**
		Get the PluginDefinition of either the job or the configuration, depending in the
		<code>preferJob</code> parameter that was set when creating this object.
		@return The PluginDefinition of the job or configuration or null if there is none
	*/
	public PluginDefinition getPluginDefinition()
	{
		if (preferJob && job != null)
		{
			return job.getPluginDefinition();
		}
		else if (config != null)
		{
			return config.getPluginDefinition();
		}
		return null;
	}

	/**
		Get the value of the parameter with the specified name. This method first looks
		for a value in the Request object. If not found, it looks in the job object,
		and at last in the configuration object.
		@return The value of the specified parameter
	*/
	public Object getValue(String name)
		throws PermissionDeniedException, BaseException
	{
		Object value = null;
		if (request != null) value = request.getParameterValue(name);
		if (value == null && job != null) value = job.getValue(name);
		if (value == null && config != null) value = config.getValue(name);
		return value;
	}

	/**
		Get the values of the parameter with the specified name. This method first looks
		for a value in the Request object. If not found, it looks in the job object,
		and at last in the configuration object.
		@return The value of the specified parameter
	*/
	public List<?> getValues(String name)
		throws PermissionDeniedException, BaseException
	{
		List<?> values = null;
		if (request != null) values = request.getParameterValues(name);
		if (values == null && job != null) values = job.getValues(name);
		if (values == null && config != null) values = config.getValues(name);
		return values;
	}

	/**
		Set the value on either the job or configuration, depending in the
		<code>preferJob</code> parameter that was set when creating this object.
	*/
	public <T> void setValue(String name, ParameterType<T> type, T value)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		if (preferJob && job != null)
		{
			job.setValue(name, type, value);
		}
		else if (config != null)
		{
			config.setValue(name, type, value);
		}
	}

	/**
		Set the value on either the job or configuration, depending in the
		<code>preferJob</code> parameter that was set when creating this object.
	*/
	public <T> void setValues(String name, ParameterType<T> type, List<T> values)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		if (preferJob && job != null)
		{
			job.setValues(name, type, values);
		}
		else if (config != null)
		{
			config.setValues(name, type, values);
		}
	}

}
