/*
	$Id: Annotations.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.query;

import net.sf.basedb.core.Operator;
import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.AnnotationSimpleRestriction;
import net.sf.basedb.core.AnnotationBetweenRestriction;
import net.sf.basedb.core.AnnotationInRestriction;
import net.sf.basedb.core.InvalidDataException;

/**
	A factory class to create restrictions based on annotation values.

	@author Nicklas
	@version 2.0
*/
public class Annotations
{

	/**
		Create a restriction that searches for annotations
		equal to a given value on the root entity of a query.
		@param annotationType The annotation type to search
		@param value The value to search for
		@param includeInheriting If items inherting the annotation should be returned
			by the query or not
		@deprecated Use {@link #eq(String, AnnotationType, Object, boolean)} instead with alias=null
	*/
	public static Restriction eq(AnnotationType annotationType, Object value, boolean includeInheriting)
		throws InvalidDataException
	{
		return eq(null, annotationType, value, includeInheriting);
	}

	/**
		Create a restriction that searches for annotations
		equal to a given value on the root or a joined entity of a 
		query.
		
		@param alias The alias of the joined entity or null to use the root element
		@param annotationType The annotation type to search. Can not be null.
		@param value The value to search for. Can not be null.
		@param includeInheriting If items inheriting the annotation should be returned
			by the query or not
	  	@return A restriction on annotations.
	 	@throws InvalidDataException If annotationType or value are null.
		@since 2.4
	*/
	public static Restriction eq(String alias, AnnotationType annotationType, Object value, boolean includeInheriting)
		throws InvalidDataException
	{
		return new AnnotationSimpleRestriction(null, annotationType, Operator.EQ, value, includeInheriting);
	}
	
	/**
		Create a restriction that searches for annotations
		not equal to a given value on the root entity of a query.
		@param annotationType The annotation type to search
		@param value The value to search for
		@param includeInheriting If items inherting the annotation should be returned
			by the query or not
		@deprecated Use {@link #neq(String, AnnotationType, Object, boolean)} instead with alias=null
	*/
	public static Restriction neq(AnnotationType annotationType, Object value, boolean includeInheriting)
		throws InvalidDataException
	{
		return neq(null, annotationType, value, includeInheriting);
	}
	
	/**
		Create a restriction that searches for annotations
		not equal to a given value on the root or a joined entity of a 
		query.
	
		@param alias The alias of the joined entity or null to use the root element
		@param annotationType The annotation type to search. Can not be null.
		@param value The value to search for. Can not be null.
		@param includeInheriting If items inheriting the annotation should be returned
			by the query or not
	 	@return A restrictions to use for annotations. 
	 	@throws InvalidDataException If annotationType or value are null.
		@since 2.4
	*/
	public static Restriction neq(String alias, AnnotationType annotationType, Object value, boolean includeInheriting)
		throws InvalidDataException
	{
		return new AnnotationSimpleRestriction(alias, annotationType, Operator.NEQ, value, includeInheriting);
	}

	/**
		Create a restriction that searches for annotations
		less than a given value on the root entity of a query.
		@param annotationType The annotation type to search
		@param value The value to search for
		@param includeInheriting If items inherting the annotation should be returned
			by the query or not
		@deprecated Use {@link #lt(String, AnnotationType, Object, boolean)} instead with alias=null
	*/
	public static Restriction lt(AnnotationType annotationType, Object value, boolean includeInheriting)
		throws InvalidDataException
	{
		return lt(null, annotationType, value, includeInheriting);
	}

	/**
		Create a restriction that searches for annotations
		less than a given value on the root or a joined entity of a 
		query.
	
		@param alias The alias of the joined entity or null to use the root element
		@param annotationType The annotation type to search. Can not be null.
		@param value The value to search for. Can not be null.
		@param includeInheriting If items inheriting the annotation should be returned
			by the query or not
	 	@return A restriction on annotations. 
	 	@throws InvalidDataException If annotationType parameter or value parameter are null.
		@since 2.4
	*/
	public static Restriction lt(String alias, AnnotationType annotationType, Object value, boolean includeInheriting)
		throws InvalidDataException
	{
		return new AnnotationSimpleRestriction(alias, annotationType, Operator.LT, value, includeInheriting);
	}

	/**
		Create a restriction that searches for annotations
		less than or equal to a given value on the root entity of a query.
		@param annotationType The annotation type to search
		@param value The value to search for
		@param includeInheriting If items inherting the annotation should be returned
			by the query or not
		@deprecated Use {@link #lteq(String, AnnotationType, Object, boolean)} instead with alias=null
	*/
	public static Restriction lteq(AnnotationType annotationType, Object value, boolean includeInheriting)
		throws InvalidDataException
	{
		return lteq(null, annotationType, value, includeInheriting);
	}

	/**
		Create a restriction that searches for annotations
		less than or equal to a given value on the root or a joined entity of a 
		query.
	
		@param alias The alias of the joined entity or null to use the root element
		@param annotationType The annotation type to search. Can not be null.
		@param value The value to search for. Can not be null.
		@param includeInheriting If items inheriting the annotation should be returned
			by the query or not
	 	@return A restriction on annotations. 
	 	@throws InvalidDataException If annotationType parameter or value parameter are null.
		@since 2.4
	*/
	public static Restriction lteq(String alias, AnnotationType annotationType, Object value, boolean includeInheriting)
		throws InvalidDataException
	{
		return new AnnotationSimpleRestriction(alias, annotationType, Operator.LTEQ, value, includeInheriting);
	}

	/**
		Create a restriction that searches for annotations
		greater than a given value on the root entity of a query.
		@param annotationType The annotation type to search
		@param value The value to search for
		@param includeInheriting If items inherting the annotation should be returned
			by the query or not
		@deprecated Use {@link #gt(String, AnnotationType, Object, boolean)} instead with alias=null
	*/
	public static Restriction gt(AnnotationType annotationType, Object value, boolean includeInheriting)
		throws InvalidDataException
	{
		return gt(null, annotationType, value, includeInheriting);
	}

	/**
		Create a restriction that searches for annotations
		greater than a given value on the root or a joined entity of a 
		query.
	
		@param alias The alias of the joined entity or null to use the root element
		@param annotationType The annotation type to search. Can not be null.
		@param value The value to search for. Can not be null.
		@param includeInheriting If items inheriting the annotation should be returned
			by the query or not
	 	@return A restriction on annotations.
	 	@throws InvalidDataException If annotationType parameter or value parameter are null.
		@since 2.4
	*/
	public static Restriction gt(String alias, AnnotationType annotationType, Object value, boolean includeInheriting)
		throws InvalidDataException
	{
		return new AnnotationSimpleRestriction(alias, annotationType, Operator.GTEQ, value, includeInheriting);
	}
	
	
	/**
		Create a restriction that searches for annotations
		greater than or equal to a given value on the root entity of a query.
		@param annotationType The annotation type to search
		@param value The value to search for
		@param includeInheriting If items inherting the annotation should be returned
			by the query or not
		@deprecated Use {@link #gteq(String, AnnotationType, Object, boolean)} instead with alias=null
	*/
	public static Restriction gteq(AnnotationType annotationType, Object value, boolean includeInheriting)
		throws InvalidDataException
	{
		return gteq(null, annotationType, value, includeInheriting);
	}

	/**
		Create a restriction that searches for annotations
		greater than or equal to a given value on the root or a joined entity of a 
		query.
	
		@param alias The alias of the joined entity or null to use the root element
		@param annotationType The annotation type to search. Can not be null.
		@param value The value to search for. Can not be null.
		@param includeInheriting If items inheriting the annotation should be returned
			by the query or not
	 	@return A restriction on on annotations. 
	 	@throws InvalidDataException If annotationType parameter or value parameter are null. 
		@since 2.4
	*/
	public static Restriction gteq(String alias, AnnotationType annotationType, Object value, boolean includeInheriting)
		throws InvalidDataException
	{
		return new AnnotationSimpleRestriction(alias, annotationType, Operator.GTEQ, value, includeInheriting);
	}

	/**
		Create a restriction that searches for annotations
		matching a given value on the root entity of a query.
		@param annotationType The annotation type to search
		@param value The value to search for
		@param includeInheriting If items inherting the annotation should be returned
			by the query or not
		@deprecated Use {@link #like(String, AnnotationType, Object, boolean)} instead with alias=null
	*/
	public static Restriction like(AnnotationType annotationType, Object value, boolean includeInheriting)
		throws InvalidDataException
	{
		return like(null, annotationType, value, includeInheriting);
	}

	/**
		Create a restriction that searches for annotations
		matching a given value on the root or a joined entity of a 
		query.
		
		@param alias The alias of the joined entity or null to use the root element
		@param annotationType The annotation type to search. Can not be null.
		@param value The value to search for. Can not be null.
		@param includeInheriting If items inheriting the annotation should be returned
			by the query or not
	 	@return A restriction on annotations. 
	 	@throws InvalidDataException If annotationType parameter or value parameter are null.
		@since 2.4
	*/
	public static Restriction like(String alias, AnnotationType annotationType, Object value, boolean includeInheriting)
		throws InvalidDataException
	{
		return new AnnotationSimpleRestriction(alias, annotationType, Operator.LIKE, value, includeInheriting);
	}

	
	/**
		Create a restriction that searches for annotations
		between two given values on the root entity of a query. 
		If the end values are included or excluded
		depends on the database.
		
		@param annotationType The annotation type to search
		@param lowValue The lower bound
		@param highValue The higher bound
		@param includeInheriting If items inherting the annotation should be returned
			by the query or not
		@deprecated Use {@link #between(String, AnnotationType, Object, Object, boolean)}
			instead with alias=null
	*/
	public static Restriction between(AnnotationType annotationType, Object lowValue, Object highValue, boolean includeInheriting)
		throws InvalidDataException
	{
		return between(null, annotationType, lowValue, highValue, includeInheriting);
	}

	/**
		Create a restriction that searches for annotations
		between two given values on the root or a joined entity of a 
		query. If the end values are included or excluded
		depends on the database.
		
		@param alias The alias of the joined entity or null to use the root element
		@param annotationType The annotation type to search. Can not be null.
		@param lowValue The lower bound. Can not be null.
		@param highValue The higher bound. Can not be null.
		@param includeInheriting If items inheriting the annotation should be returned
			by the query or not
	 	@return A restriction on annotations.
	 	@throws InvalidDataException If any of the required parameters are null.
		@since 2.4
	*/
	public static Restriction between(String alias, AnnotationType annotationType, Object lowValue, Object highValue, boolean includeInheriting)
		throws InvalidDataException
	{
		return new AnnotationBetweenRestriction(alias, annotationType, lowValue, highValue, includeInheriting);
	}
	
	/**
		Create a restriction that searches for annotations on the root entity of a query
		equal to any value in a list of given values.
		@param annotationType The annotation type to search
		@param includeInheriting If items inherting the annotation should be returned
			by the query or not
		@param values An array of values
		@deprecated Use {@link #in(String, AnnotationType, boolean, Object[])}
			instead with alias=null
	*/
	public static Restriction in(AnnotationType annotationType, boolean includeInheriting, Object... values)
		throws InvalidDataException
	{
		return in(null, annotationType, includeInheriting, values);
	}

	/**
		Create a restriction that searches for annotations
		on the root or a joined entity
		equal to any value in a list of given values.
		
		@param alias The alias of the joined entity or null to use the root element
		@param annotationType The annotation type to search. Can not be null.
		@param includeInheriting If items inheriting the annotation should be returned
			by the query or not
		@param values An array of values. Can not be null.
	 	@return A restriction on annotations. 
	 	@throws InvalidDataException If annotationType parameter or values parameter are null 
		@since 2.4
	*/
	public static Restriction in(String alias, AnnotationType annotationType, boolean includeInheriting, Object... values)
		throws InvalidDataException
	{
		return new AnnotationInRestriction(alias, annotationType, includeInheriting, values);
	}

}
