/*
	$Id: InRestriction.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.query;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.BaseException;

/**
	Compare if one expression is contained in a set of other expressions:
	e1 IN (e2[0], e2[1], ...)

	@author Samuel, Nicklas
	@version 2.0
	@see Restrictions#in(Expression, Expression[])
	@base.modified $Date: 2009-04-06 14:52:39 +0200 (Mon, 06 Apr 2009) $
*/
public class InRestriction
	implements Restriction
{
	private final Expression e1;
	private final Expression[] e2;

	InRestriction(Expression e1, Expression... e2)
	{
		assert e1 != null : "e1 == null";
		assert e2 != null : "e2 == null";
		this.e1 = e1;
		this.e2 = e2;
	}

	/*
		From the QueryElement interface
		-------------------------------------------
	*/
	public String toQl(Query query, DbControl dc)
		throws BaseException
	{
		StringBuilder sb = new StringBuilder();
		int i = 0;
		String exp1 = e1.toQl(query, dc);
		sb.append("(").append(exp1).append(" IN (");
		boolean checkNull = false;
		for (Expression e : e2)
		{
			if (e == null)
			{
				checkNull = true;
				continue;
			}
			if (i > 0) sb.append(",");
			sb.append(e.toQl(query, dc));
			i++;
		}
		sb.append(")");
		if (i == 0) sb.setLength(0);
		if (checkNull)
		{
			if (sb.length() > 0)sb.append (" OR ");
			sb.append(exp1).append(" IS NULL");
		}
		sb.append(")");
		return sb.toString();
	}
	/**
		@return The expressions
	*/
	public Collection<? extends QueryElement> getChildren()
	{
		List<Expression> all = new ArrayList<Expression>(Arrays.asList(e2));
		all.add(e1);
		return Collections.unmodifiableList(all);
	}
	// -------------------------------------------
	/*
		From the Object class
		-------------------------------------------
	*/
	public String toString()
	{
		StringBuilder sb = new StringBuilder();
		int i = 0;
		sb.append("(").append(e1.toString()).append(" IN (");
		for (Expression e : e2)
		{
			if (i > 0) sb.append(",");
			sb.append(e == null ? "NULL" : e.toString());
			i++;
		}
		sb.append(")");
		return sb.toString();
	}
	// -------------------------------------------

}

