/**
	$Id: SignalReceiver.java 4516 2008-09-11 20:09:17Z jari $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.signal;

/**
	A signal receiver is a class that knows how to receive a {@link Signal} from a
	{@link SignalTransporter} and how to distribute this to a {@link SignalHandler}.
	Signal receivers are needed to decouple the actual signal handling from
	the transportation of signals. A signal receiver implementation usually
	comes in pair with a singal transporter implementation. See, for example,
	the {@link SocketSignalReceiver} and {@link SocketSignalTransporter} implementations
	which uses network sockets to transport signals.
	<p>
	Signal receivers must be thread safe since a single receiver may be used to 
	handle signals for multiple jobs at the same time.
	
	@author nicklas
	@version 2.6
	@see SignalTransporter
	@see SignalHandler
	@base.modified $Date: 2008-09-11 22:09:17 +0200 (Thu, 11 Sep 2008) $
*/
public interface SignalReceiver
{

	/**
		Initialise the signal receiver. The syntax and meaning of the 
		parameter sent to it is implementation-dependant. When this method 
		return the signal receiver is ready to register signal handlers and 
		receive signals.
		@param params Parameters needed to initialise the receiver
	*/
	public void init(String params);
	
	/**
		Close the receiver. The receiver should close and cleanup any opened
		resources, unregister itself if neccesary and stop listening for signals.

		@param wait If the value is zero or positive, the {@link Signal#ABORT}
			should be sent to all registered signal handlers. The signal receiver 
			should then wait at most the specified number of milliseconds for
			all signal handlers to get unregistered. If the timeout expires before
			all signal handlers has processed the signal, the signal receiver should
			continue it's shutdown.
	*/
	public void close(int wait);
	
	/**
		Get the signal transporter class that should be used to
		send signals to this receiver.
		@return A class objects that must be a subclass of {@link SignalTransporter}
	*/
	public Class<? extends SignalTransporter> getSignalTransporterClass();
	
	/**
		Register a signal handler with this receiver. The receiver should return
		a string, that, when used with the {@link SignalTransporter#init(String)}
		method lets the transporter know how to send a signal to the specific 
		handler. The {@link SocketSignalReceiver} will for example, include the
		IP number and port the transporter should connect to and a unique ID that
		allows the receiver to locate the signal hander the signal is intended for.
		The format of the ID is implementation specific and need only to be understandable
		by the receiver and transporter implementations.
		@param handler The signal handler that should receive signals through this
			receiver
		@return A string that allows a transporter instance locate and send a signal to
			the given signal handler
		@see AbstractSignalReceiver
	*/
	public String registerSignalHandler(SignalHandler handler);
	
	/**
		Unregister a signal handler from this receiever. After this method returns, the
		receiver must not send any more signals to the specified handler.
		
		@param handler The handler to unregister
	*/
	public void unregisterSignalHandler(SignalHandler handler);
	
	/**
		Send a signal to all registered signal handlers that supports
		it. The main purpose of this method is to be able to send
		the {@link Signal#ABORT} when the system is shutting down.
		@param signal The signal to send
	*/
	public void sendToAll(Signal signal);
	
}
