/*
	$Id: BioAssaySetFilterUtil.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DynamicSpotQuery;
import net.sf.basedb.core.FilterBatcher;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.Transformation;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.VirtualColumn;
import net.sf.basedb.core.DynamicResultIterator;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.SqlResult;
import net.sf.basedb.core.signal.ThreadSignalHandler;

import java.sql.SQLException;
import java.util.List;


/**
	This is a utility class for plugins that filters bioassay set data.
	
	@base.modified $Date: 2009-04-06 14:52:39 +0200 (Mon, 06 Apr 2009) $
	@author Nicklas
	@version 2.0
*/
public class BioAssaySetFilterUtil
{

	/**
		Create a filtered bioassay set from a parent bioassay set using
		a simple restriction. The filtering is done by the 
		{@link FilterBatcher#insert(net.sf.basedb.core.AbstractSqlQuery)} which is 
		very fast since most of the work is done by the database.
		<p>
		The calling code mustn't forget to commit the transaction.
		
		@param dc The <code>DbControl</code> object to use for database acecss
		@param parent The parent bioassay set to start from
		@param bioAssays Only filter on the specified bio assays, if the list
			is null or empty all bioassays will be filtered
		@param job The job that the transformation should be linked to, or
			null if this method isn't called by a job
		@param filter A filter expression
		@param progress A <code>ProgressReporter</code> object (optional, current not used)
	 	@return A filtered bioassay set. 
		@throws InvalidDataException If any of the required parameters is null
		@throws BaseException If creating a filtered bioassay set fails in some way.
	*/
	public static BioAssaySet createFilteredBioAssaySet(DbControl dc, BioAssaySet parent, 
		List<BioAssay> bioAssays, Job job, Restriction filter, ProgressReporter progress)
		throws InvalidDataException, BaseException
	{
		if (parent == null) throw new InvalidUseOfNullException("parent");
		if (filter == null) throw new InvalidUseOfNullException("filter");
	
		if (progress != null) progress.display(5, "Initialising spot filter...");
		ThreadSignalHandler.checkInterrupted();

		// Create Transformation
		Transformation t = parent.newTransformation(job);
		dc.saveItem(t);

		// Create the filtered BioAssaySet in the same datacube and layer
		BioAssaySet child = t.newProduct(null, null, false);
		child.setName(parent.getName());
		dc.saveItem(child);

		// Create the batchers we need
		FilterBatcher filterBatcher = child.getFilterBatcher();

		// Create the query
		DynamicSpotQuery query = parent.getSpotData();
		query.select(Dynamic.select(VirtualColumn.COLUMN));
		query.select(Dynamic.select(VirtualColumn.POSITION));
		query.restrict(filter);
		query.restrict(
			Restrictions.eq(
				Dynamic.column(VirtualColumn.COLUMN),
				Expressions.parameter("column")
			)
		);

		//	 Prepare progress reporting
		int spotsTotal = parent.getNumSpots();
		if (bioAssays != null && bioAssays.size() > 0)
		{
			spotsTotal = 0;
			for (BioAssay ba : bioAssays)
			{
				spotsTotal += ba.getNumSpots();
			}
		}
		
		long spotsDone = 0;
		int spotsAfter = 0;
		if (progress != null) progress.display(10, "Filtering spots (0 done; "+spotsTotal+" total)...");

		// Create filter
		if (bioAssays == null || bioAssays.size() == 0) bioAssays =  parent.getBioAssays().list(dc);
		for (BioAssay parentBa : bioAssays)
		{
			ThreadSignalHandler.checkInterrupted();
			BioAssay childBa = child.newBioAssay(parentBa);
			dc.saveItemIf(child, childBa, false);
			spotsDone += parentBa.getNumSpots();
			query.setParameter("column", (int)parentBa.getDataCubeColumnNo(), Type.INT);
			spotsAfter += filterBatcher.insert(query);

			if (progress != null) 
			{
				int percent = 10 + (int)(90L * spotsDone / spotsTotal);
				progress.display(percent, 
					"Filtering spot intensities ("+spotsDone+" done; "+spotsTotal+" total)...");
			}
		}
		
		// Close and clean up, calling code must commit
		filterBatcher.close();

		if (progress != null)
		{
			progress.display(100, "Filtering spot intensities ("+spotsTotal+" done; "+spotsTotal+" total)...\n");
		}

		// Return  
		return child;
	}
	
	public static BioAssaySet createFilteredBioAssaySet(DbControl dc, BioAssaySet parent, 
		List<BioAssay> bioAssays, Job job, DynamicFilter filter, ProgressReporter progress)
		throws InvalidDataException, BaseException
	{
		if (parent == null) throw new InvalidUseOfNullException("parent");
		if (filter == null) throw new InvalidUseOfNullException("filter");
	
		if (progress != null) progress.display(5, "Initialising spot filter...");
		ThreadSignalHandler.checkInterrupted();

		// Create Transformation
		Transformation t = parent.newTransformation(job);
		dc.saveItem(t);

		// Create the filtered BioAssaySet in the same datacube and layer
		BioAssaySet child = t.newProduct(null, null, bioAssays == null || bioAssays.size() == 0);
		child.setName(parent.getName());
		dc.saveItem(child);
		
		// Create the batchers we need
		FilterBatcher filterBatcher = child.getFilterBatcher();

		// Create the query
		DynamicSpotQuery query = parent.getSpotData();
		query.select(Dynamic.select(VirtualColumn.COLUMN));
		query.select(Dynamic.select(VirtualColumn.POSITION));
		filter.configureQuery(query);
		
		int spotsTotal = parent.getNumSpots(); // Total number of spots

		// Create child bioassays and filter
		if (bioAssays != null && bioAssays.size() > 0)
		{
			spotsTotal = 0;
			Expression[] selectedColumns = new Expression[bioAssays.size()];
			for (int i = 0; i < selectedColumns.length; ++i)
			{
				BioAssay ba = bioAssays.get(i);
				BioAssay childBa = child.newBioAssay(ba);
				dc.saveItemIf(child, childBa, false);
				selectedColumns[i] = Expressions.integer(ba.getDataCubeColumnNo());
				spotsTotal += ba.getNumSpots();
			}
			Expression column = Dynamic.column(VirtualColumn.COLUMN);
			query.restrict(Restrictions.in(column, selectedColumns));
		}
		
		// Prepare progress reporting
		int interval = 10;  // Update progress after this many spots >= 10
		long spotsDone = 0;
		int spotsAfter = 0;
		if (progress != null)
		{
			interval = spotsTotal / 20;
			if (interval < 10) interval = 10;
			progress.display(10, "Filtering spots (0 done; "+spotsTotal+" total)...");
		}

		// Do the filtering
		if (filter.useIncludeSpot())
		{
			DynamicResultIterator result = query.iterate(dc);
			try
			{
				while (result.hasNext())
				{
					ThreadSignalHandler.checkInterrupted();
					SqlResult data = result.next();
					if (filter.includeSpot(data))
					{
						short column = data.getShort(1);
						int position = data.getInt(2);
						filterBatcher.insert(column, position);
						spotsAfter++;
					}
					
					if (progress != null)
					{
						spotsDone++;
						if (spotsDone % interval == 0)
						{
							int percent = 10 + (int)(90L * spotsDone / spotsTotal);
							progress.display(percent, 
								"Filtering spot intensities ("+spotsDone+" done; "+spotsTotal+" total)...");
						}
					}
				}
			}
			catch (SQLException ex)
			{
				throw new BaseException(ex);
			}
		}
		else
		{
			filterBatcher.insert(query);
		}

		// Close and clean up, calling code must commit
		filterBatcher.close();
		if (progress != null)
		{
			progress.display(100, "Filtering spot intensities ("+spotsTotal+" done; "+spotsTotal+" total)...\n");
		}

		// Return  
		return child;
	}


}
