/*
	$Id: RemovableUtil.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util;

import net.sf.basedb.core.AnyToAny;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Directory;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.File;
import net.sf.basedb.core.FileSetMember;
import net.sf.basedb.core.FileStoreEnabled;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultList;
import net.sf.basedb.core.Removable;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.Transformation;

import java.util.Set;

/**

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2009-04-06 14:52:39 +0200 (Mon, 06 Apr 2009) $
*/
public class RemovableUtil
{

	public static int setRemoved(DbControl dc, Item itemType, Set<Integer> items, boolean removed)
		throws BaseException
	{
		int numRemoved = 0;
		for (Integer id : items)
		{
			if (id != null)
			{
				Removable item = (Removable)itemType.getById(dc, id);
				if (item.isRemoved() != removed) 
				{
					try
					{
						item.setRemoved(removed);
						numRemoved++;
					}
					catch (PermissionDeniedException ex)
					{}
				}
				else
				{
					numRemoved++;
				}
			}
		}
		return numRemoved;
	}
	
	/**
	 	Set the removed flag for selected items, including the subitems.
	 	@param dc The database control to use when accessing the database
	 	@param itemType The itemType to delete
	 	@param items A set of ids from the items that shall be deleted.
	 	@param removed True if the items should be deleted. False if the items should be restored.
	 	@return An int[2] where [0] is the number of removed/restored items, including the parent
	 		and [1] is the total number of items that were included in the selection.
	 	@throws BaseException If there is an error
	*/
	public static int[] removeRecursively(DbControl dc, Item itemType, Set<Integer> items, boolean removed)
		throws BaseException
	{
		int[] numTotalCount = {0,0};
		for (Integer id : items)
		{			
			int[] numHandled = {0,0};
			if (id != null)
			{				 
				Removable item = (Removable)itemType.getById(dc, id);
				numHandled = removeDependingItems(dc, item, removed);
				if ((item instanceof Directory || item instanceof File) && !removed)
				{
					int[] numRestored = {0,0};
					Directory dir = null;
					
					if (item instanceof Directory)
					{
						dir = ((Directory)item).getParent();
					}
					else 
					{
						dir = ((File)item).getDirectory();
					}					
					numRestored = restorePath(dir);
					numHandled[0] += numRestored[0];
					numHandled[1] += numRestored[1];
				}
				numHandled[0] += setRemovedFlag(item, removed) ? 1 : 0;				
			}
			numTotalCount[0] += numHandled[0];
			numTotalCount[1] += numHandled[1];
		}
		return numTotalCount;
	}	
	
	public static int delete(DbControl dc, Item itemType, Set<Integer> items)
		throws BaseException
	{
		int numDeleted = 0;
		for (Integer id : items)
		{
			if (id != null)
			{
				BasicItem item = itemType.getById(dc, id);
				if (item instanceof FileStoreEnabled)
				{
					try
					{
						removeFileSetMembers(dc, (FileStoreEnabled)item, true);
					}
					catch(BaseException ex)
					{}
				}
				try
				{
					dc.deleteItem(item);
					numDeleted++;
				}
				catch (PermissionDeniedException ex)
				{}
			}
		}
		return numDeleted;
	}
	
	/**
	 	Restores the path to a directory
	  	@param parentDir The Directory that is being restored
	  	@return An int[2] were [0] is the number of restored directories and
	  		[1] is the total number of directories.
	*/
	private static int[] restorePath(Directory parentDir)
	{
		int[] countRestored = {0,0};
		if (!parentDir.isRootDirectory())
		{
			countRestored = restorePath(parentDir.getParent());
		}
		countRestored[1]++;
		countRestored[0] += setRemovedFlag(parentDir, false) ? 1 : 0;
		
		return countRestored;
	}
	
	/**
	 	Sets the removed-flag for all subitems to a parent
	 	@param dc The <code>DbControl</code> wich will be used for database access
	 	@param item Parent of the subitems that shall be removed
	 	@param removed True if the subitems should be flagged removed. 
	 			False if the subitems should be restored
	 	@return An int[2] where [0] is the number of removed items and
	 				[1] is the total number of subitems.
		@throws BaseException If there is an error
	*/
	private static int[] removeDependingItems(DbControl dc, Removable item, boolean removed)
		throws BaseException
	{
		int numItemsRemoved = 0;
		int[] numSubItemsRemoved = {0,0};
		
		/**
		 	item is a <code>Directory</code>
		*/
		if (item instanceof Directory)
		{
			ItemQuery<Directory> subDirQuery = ((Directory)item).getSubDirectories();
			ItemQuery<File> fileQuery = ((Directory)item).getFiles();
			
			ItemResultList<Directory> subDirectories = null;
			ItemResultList<File> files = null;
			
			subDirQuery.include(Include.MINE, Include.OTHERS, Include.REMOVED, Include.SHARED);
			fileQuery.include(Include.REMOVED, Include.MINE, Include.OTHERS, Include.SHARED);
									
			int totalFiles = 0;
			int totalDirectories = 0;
			
			try
			{
				subDirectories = subDirQuery.list(dc);
				files = fileQuery.list(dc);
			}
			catch (BaseException ex)
			{}
			
			if (files != null)
			{
				totalFiles = files.size();
				for (File file : files)
				{
					numItemsRemoved += setRemovedFlag(file, removed) ? 1 : 0;
				}				
			}
			
			if (subDirectories != null)
			{
				totalDirectories = subDirectories.size();
				for (Directory dir : subDirectories)
				{
					int[]numTemp = removeDependingItems(dc, dir, removed);
					numSubItemsRemoved[0] += numTemp[0];
					numSubItemsRemoved[1] += numTemp[1];
					numItemsRemoved += setRemovedFlag(dir, removed) ? 1 : 0;
				}				
				numSubItemsRemoved[1] += totalFiles + totalDirectories;
			}
		}
			
		/**
		 	item is a <code>Transformation</code>
		*/
		else if (item instanceof Transformation)
		{
			// Remove AnyToAny items this item links to.
			ItemQuery<AnyToAny> linkedItemQuery = AnyToAny.getLinksFrom((Transformation)item);			
			numSubItemsRemoved = sumArrays(numSubItemsRemoved, removeAnyToAnyLinkedTo(dc, linkedItemQuery, removed));
			if (removed)
			{
				ItemQuery<BioAssaySet> bioAssayQuery = ((Transformation)item).getProducts(); 			
				ItemResultList<BioAssaySet> bioAssaySets = null;
				
				bioAssayQuery.include(Include.REMOVED,Include.MINE, Include.OTHERS, Include.SHARED);
				
				try
				{
					bioAssaySets = bioAssayQuery.list(dc);
				}
				catch (BaseException ex)
				{}
				
				if (bioAssaySets != null)
				{
					for (BioAssaySet bsa : bioAssaySets)
					{
						numSubItemsRemoved = sumArrays(numSubItemsRemoved, removeDependingItems(dc, bsa, removed));						
						numItemsRemoved  += setRemovedFlag(bsa, removed) ? 1 : 0;
					}
					numSubItemsRemoved[1] += bioAssaySets.size();
				}
			}
			else
			{
				Transformation tf = (Transformation)item;
				BioAssaySet bas = tf.getSource();
				// The transformation is not a root transformation
				if (bas != null)
				{
					int[]numTemp = removeDependingItems(dc, bas, removed);
					numSubItemsRemoved[0] += numTemp[0];
					numSubItemsRemoved[1] += numTemp[1];
					
					numItemsRemoved += setRemovedFlag(bas, removed) ? 1 : 0;
					numSubItemsRemoved[1]++;
				}
			}
		}
		
		
		/**
		 	item is a <code>BioAssaySet</code>
		*/
		else if (item instanceof BioAssaySet)
		{
			numSubItemsRemoved = sumArrays(numSubItemsRemoved, removeFileSetMembers(dc, (BioAssaySet)item, removed));			
			if (removed)
			{
				ItemQuery<Transformation> transformationQuery = ((BioAssaySet)item).getTransformations();
				ItemResultList<Transformation> transformations = null;
				
				transformationQuery.include(Include.REMOVED,Include.MINE, Include.OTHERS, Include.SHARED);
				
				try
				{
					transformations = transformationQuery.list(dc);
				}
				catch (BaseException ex)
				{}
				
				if (transformations != null)
				{
					for (Transformation tf : transformations)
					{
						numSubItemsRemoved = sumArrays(numSubItemsRemoved, removeDependingItems(dc, tf, removed));
						numItemsRemoved += setRemovedFlag(tf, removed) ? 1 : 0;
					}
					numSubItemsRemoved[1] += transformations.size();
				}						
			}
			else
			{
				BioAssaySet bas = (BioAssaySet)item;
				Transformation tf = bas.getTransformation();
				if (tf != null)
				{
					int[]numTemp = removeDependingItems(dc, tf, removed);
					numSubItemsRemoved[0] += numTemp[0];
					numSubItemsRemoved[1] += numTemp[1];
					numItemsRemoved += setRemovedFlag(tf, removed) ? 1 : 0;
					numSubItemsRemoved[1]++;
				}
			}
		}
		
		/**
		 	item is an <code>Experiment</code>
		*/
		else if (item instanceof Experiment)
		{
			ItemQuery<Transformation> transformationQuery = ((Experiment)item).getRootTransformations();
			ItemResultList<Transformation> transformations = null;
			
			transformationQuery.include(Include.REMOVED,Include.MINE, Include.OTHERS, Include.SHARED);
			
			try
			{
				transformations = transformationQuery.list(dc);
			}
			catch (BaseException ex)
			{}
			
			if (transformations != null)
			{
				for (Transformation tf : transformations)
				{
					int[]numTemp = removeDependingItems(dc, tf, removed);
					numSubItemsRemoved[0] += numTemp[0];
					numSubItemsRemoved[1] += numTemp[1];
					numItemsRemoved += setRemovedFlag(tf, removed) ? 1 : 0;
				}
				numSubItemsRemoved[1] += transformations.size();
			}				
		}
		
		numSubItemsRemoved[0] += numItemsRemoved;
		return numSubItemsRemoved;
	}
	
	private static boolean setRemovedFlag(Removable item, boolean removed)
	{
		boolean isRemoved = false;
		if (item.isRemoved() != removed)
		{
			try
			{
				item.setRemoved(removed);
				isRemoved = true;
			}
			catch (PermissionDeniedException ex)
			{}
		}
		else
		{
			isRemoved = true;
		}
		return isRemoved;
	}	
	
	// Adds numbers with same index and returns the summarized array.
	private static int[] sumArrays(int[] baseArray, int[] newResult)
	{
		for (int i=0; i<baseArray.length; i++)
		{
			baseArray[i] += newResult[i];
		}	
		return baseArray;
	}
	

	
	/*
	 	Removes all removable items that are linked to items in a list
	 */
	private static int[] removeAnyToAnyLinkedTo(DbControl dc, ItemQuery<AnyToAny> any2AnyQuery, boolean removed)
	{
		int[] numA2AItemsRemoved = {0,0};
		try
		{
			for (AnyToAny a2a : any2AnyQuery.list(dc))
			{
				if (a2a.getTo() instanceof Removable) 
				{
					Removable removableItem = (Removable)a2a.getTo();
					if (removableItem instanceof Directory)
					{
						numA2AItemsRemoved = removeDependingItems(dc, removableItem, removed);
					}
					numA2AItemsRemoved[0] += setRemovedFlag(removableItem, removed) ? 1 : 0;
					numA2AItemsRemoved[1]++;
				}
			}
		}
		catch (BaseException ex)
		{} 
		return numA2AItemsRemoved;
	}
	
	/*
	 	Sets all fileset members to be removed
	 */
	private static int[] removeFileSetMembers(DbControl dc, FileStoreEnabled fileStore, boolean removed)
	{
		int[] counter = {0,0};
		try
		{
			ItemQuery<FileSetMember> memberQuery = fileStore.getFileSet().getMembers();
			for (FileSetMember member : memberQuery.list(dc))
			{
				counter[0] += setRemovedFlag(member.getFile(), removed) ? 1 : 0;
				counter[1]++;
			}
		}
		catch(BaseException ex)
		{}
		return counter;
	}

}
