/**
	$Id: StaticCache.java 4829 2009-03-23 14:28:03Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util;

import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FilterInputStream;
import java.io.FilterOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.Serializable;
import java.lang.ref.WeakReference;
import java.util.List;
import java.util.Map;
import java.util.TimerTask;
import java.util.WeakHashMap;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.regex.Pattern;

import net.sf.basedb.core.InvalidDataException;

/**
	A cache for storing data to files on the file system. The typical
	use case is to store data this is expensive to get from the database
	in a file for later retrieval. The cache can be used in streaming mode
	with the any of the {@link #read(String, int)} or {@link #write(String, InputStream, int)}
	methods and their variants. It can also be used to store any {@link Serializable}
	object with {@link #store(String, Object, int)} and {@link #load(String, int)}.
	<p>
	
	In all cases the cached entry is identified by a key which is more or
	less directly translated to directories on the file system.
	<p>
	
	This class is thread-safe and can be used by multiple threads at the same
	time. Write requests to the same entry in the cache are allowed to one 
	thread at a time. Any number of threads may read from the same cache entry
	as long as no thread is writing to that entry.
	
	@author Nicklas
	@version 2.11
	@base.modified $Date: 2009-03-23 15:28:03 +0100 (Mon, 23 Mar 2009) $
*/
public class StaticCache
{

	public static final Pattern validKey = Pattern.compile("[\\w\\/\\.\\-]+");

	/**
		Log static cache events.
	*/
	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.util.StaticCache");

	/**
		Checks if the given key is a vilid cache entry key.
		Allowed characters are: any alphanumerical character + ./-
		@param key The key to check
		@return TRUE if the key is valid
	*/
	public static boolean isValidKey(String key)
	{
		return validKey.matcher(key).matches();
	}
	
	private final File root;
	private final Map<String, LockEntry> locks;
	private boolean disabled;
	
	
	/**
		Creates a new static cache.
		@param root A directory were the cache stores it's files
	*/
	public StaticCache(File root)
	{
		this.root = root;
		this.locks = new WeakHashMap<String, LockEntry>();
		log.info("Creating static cache in directory " + root);
	}

	/**
		Check if the cache has been disabled. A disabled cache ignores
		all calls to read or write data.
		@return A boolean that is TRUE if the cache is disabled
	*/
	public boolean isDisabled()
	{
		return disabled;
	}
	
	/**
		Disable or enable the cache.
		@param disabled TRUE if the cache should be disabled
	*/
	public void setDisabled(boolean disabled)
	{
		if (disabled)
		{
			log.info("Disabling static cache in directory " + root);
		}
		else
		{
			log.info("Enabling static cache in directory " + root);
		}
		this.disabled = disabled;
	}
	
	/**
		Remove all files that matches the specified filter from
		the cache. This method is synchronized and can only
		be executed by one thread at a time.
		
		@param filter A file filter that matches the files to
			remove
	*/
	public synchronized void cleanUp(FileFilter filter)
	{
		log.info("Cleaning up static cache: " + root);
		List<File> oldFiles = FileUtil.findFiles(root, filter);
		if (oldFiles == null) return;
		
		log.debug("Found " + oldFiles.size() + " files that should be deleted");
		int numDeleted = 0;
		for (File f : oldFiles)
		{
			if (f.delete())
			{
				log.debug("Removed cached file: " + f);
				numDeleted++;
			}
			else
			{
				log.warn("Failed to remove cached file: " + f);
			}
		}
		log.info("Removed " + numDeleted + " files from the static cache: " + root);
	}
	
	/**
		Creates a file filter that matches all files that
		are older than the specified age.
		@param age The age in milliseconds
		@return A file filter
		@see OlderThanFileFilter
	*/
	public FileFilter olderThan(long age)
	{
		return new OlderThanFileFilter(age, true);
	}
	
	/**
		Creates a task that cleans up this cache when it is 
		executed.
		@param filter The file filter that determines
			which file that should be deleted
		@return A TimerTask object
		@see #cleanUp(FileFilter)
	*/
	public TimerTask cleanUpTask(FileFilter filter)
	{
		return new CleanupTask(this, filter);
	}
	
	/**
		Store binary information in the cache. If the entry already exists,
		it is overwritten. If the entry is locked by other threads,
		this thread waits the specified number of milliseconds before
		returning.
		
		@param key The cache key
		@param in An input stream to read from
		@param timeout A timeout in milliseconds to wait for a write lock
			on the requested cache entry
		@return The number of bytes written
	*/
	public long write(String key, InputStream in, int timeout)
		throws IOException
	{
		if (disabled) return 0;
		long numBytes = 0;
		OutputStream out = null;
		try
		{
			out = getOutputStream(key, timeout);
			if (out != null)
			{
				numBytes = FileUtil.copy(in, out);
				out.flush();
			}
		}
		finally
		{
			if (out != null) out.close();
		}
		return numBytes;
	}
	
	/**
		Get an output stream that can be used to write binary information
		to the cache. If the entry already exists, it is overwritten. 
		If the entry is locked by other threads, this thread waits the 
		specified number of milliseconds before giving up.
		<p>
		NOTE! It is very important that the caller closes the 
		output stream as soon as all data has been written to it.
		Failure to do so may result in locking the cache entry from
		reading by other threads.
		
		@param key The cache key
		@param timeout A timeout in milliseconds to wait for a write lock
			on the requested cache entry
		@return An output stream, or null if a write lock could not
			be aquired
	*/
	public OutputStream write(String key, int timeout)
		throws IOException
	{
		if (disabled) return null;
		return getOutputStream(key, timeout);
	}
	
	/**
		Read binary information from the cache. The contents of the 
		specified cache entry will be copied to the specified output 
		stream.

		@param key The cache key
		@param out An output stream to write the cache contents to
		@param timeout A timeout in milliseconds to wait for a 
			read lock on the requested cache entry
		@return The number of bytes copied
		@throws IOException
	*/
	public long read(String key, OutputStream out, int timeout)
		throws IOException
	{
		if (disabled) return 0;
		long numBytes = 0;
		InputStream in = null;
		try
		{
			in = getInputStream(key, timeout);
			if (in != null) 
			{
				numBytes = FileUtil.copy(in, out);
			}
		}
		finally
		{
			if (in != null) in.close();
		}
		return numBytes;
	}
	
	/**
		Get an input stream for reading binary information from the cache.
	
		@param key The cache key
		@param timeout A timeout in milliseconds to wait for a 
			read lock on the requested cache entry
		@return An input stream or null if the cache entry doesn't exists
			or if a read lock could not be aquired
		@throws IOException
	*/
	public InputStream read(String key, int timeout)
		throws IOException
	{
		if (disabled) return null;
		return getInputStream(key, timeout);
	}
	
	/**
		Store a serializable object in the cache. If the entry already exists,
		it is overwritten. If the entry is locked by other threads,
		this thread waits the specified number of milliseconds before
		returning.
		
		@param key The cache key
		@param object The object to store in the cache, it must be a
			{@link Serializable} object
		@param timeout A timeout in milliseconds to wait for a write lock
			on the requested cache entry
		@return TRUE if the object could be stored, FALSE otherwise
	*/
	public boolean store(String key, Object object, int timeout)
	{
		if (disabled) return false;
		OutputStream out = null;
		boolean stored = false;
		try
		{
			out = getOutputStream(key, timeout);
			if (out != null)
			{
				ObjectOutputStream oos = new ObjectOutputStream(out);
				oos.writeObject(object);
				oos.flush();
				stored = true;
			}
		}
		catch (IOException ex)
		{
			log.warn("Could not store cached entry: " + key, ex);
		}
		finally
		{
			if (out != null) 
			{
				try
				{
					out.close();
				}
				catch (IOException ex)
				{}
			}
		}
		return stored;
	}
	
	/**
		Read a serializable object from the cache.
		
		@param key The cache key
		@param timeout A timeout in milliseconds to wait for a read lock
			on the requested cache entry
		@return The materialized object, or null if the cache entry doesn't
			exists or if a read lock couldn't be aquired
	*/
	public Object load(String key, int timeout)
	{
		if (disabled) return null;
		Serializable object = null;
		InputStream in = null;
		try
		{
			in = getInputStream(key, timeout);
			if (in != null)
			{
				ObjectInputStream oin = new ObjectInputStream(in);
				try
				{
					object = (Serializable)oin.readObject();
				}
				catch (ClassNotFoundException ex)
				{
					throw new IOException(ex);
				}
			}
		}
		catch (IOException ex)
		{
			log.warn("Could not load cached entry: " + key, ex);
		}
		finally
		{
			if (in != null) 
			{
				try
				{
					in.close();
				}
				catch (IOException ex)
				{}
			}
		}
		return object;		
	}
	
	private void validateKey(String key)
	{
		if (!isValidKey(key))
		{
			throw new InvalidDataException("Invalid cache key: " + key);
		}
	}
	
	/**
		Get a lock-safe input stream.
	*/
	private InputStream getInputStream(String key, int timeout)
		throws IOException
	{
		validateKey(key);
		log.debug("Read request for static cache: " + key);
		File f = new File(root, key);
		if (!f.exists()) 
		{
			log.debug("Cache entry doesn't exist: " + key);
			return null;
		}
		LockEntry lock = aquireLock(key, false, timeout);
		if (lock == null) return null;
		InputStream in = null;
		try
		{
			f.setLastModified(System.currentTimeMillis());
			in = new LockSafeInputStream(new FileInputStream(f), lock);
		}
		finally
		{
			if (in == null) lock.readLock().unlock();
		}
		return in;
	}

	/**
		Get a lock-safe output stream.
	*/
	private OutputStream getOutputStream(String key, int timeout)
		throws IOException
	{
		validateKey(key);
		log.debug("Write request for static cache: " + key);
		LockEntry lock = aquireLock(key, true, timeout);
		if (lock == null) return null;
		OutputStream out = null;
		try
		{
			File f = new File(root, key);
			f.getParentFile().mkdirs();
			f.createNewFile();
			out = new LockSafeOutputStream(new FileOutputStream(f), lock);
		}
		finally
		{
			if (out == null) lock.writeLock().unlock();
		}
		return out;
	}

	/**
		Aquire a read or write lock on a given cache entry.
		@return A lock manager or null if the lock could not be aquired
	*/
	private LockEntry aquireLock(String key, boolean writeLock, int timeout)
	{
		String lockType = writeLock ? "write" : "read";
		log.debug("Trying to get " + lockType + " lock on: " + key);
		LockEntry rwLock = null;
		synchronized (locks)
		{
			log.debug("Number of known lock entries: " + locks.size());
			rwLock = locks.get(key);
			if (rwLock == null)
			{
				log.debug("Create new lock holder for: " + key);
				rwLock = new LockEntry(new ReentrantReadWriteLock(), key);
				locks.put(key, rwLock);
			}
		}
		
		Lock lock = writeLock ? rwLock.writeLock() : rwLock.readLock();
		try
		{
			if (!lock.tryLock(timeout, TimeUnit.MILLISECONDS)) 
			{
				log.debug("Timeout while waiting for " + lockType + " lock on: " + key);
				rwLock = null;
			}
		}
		catch (InterruptedException ex)
		{
			log.debug("Interrupted while waiting for " + lockType + " lock on: " + key);
			rwLock = null;
		}
		if (rwLock != null) log.debug("Got " + lockType + " lock on: " + key);
		return rwLock;
	}
	
	/**
		A lock-safe output stream that releases the associated 
		write lock when the stream is closed.
		<p>
		NOTE! It is important that we keep a strong reference
		to the 'key' in this class until the stream is closed, 
		since otherwise the locked entry may be reclaimed by the 
		garbage collector.
	*/
	static class LockSafeOutputStream
		extends FilterOutputStream
	{
		private final Throwable calledFrom;
		private ReadWriteLock lock;
		private String key;
		private boolean closed;

		LockSafeOutputStream(OutputStream out, LockEntry lock)
		{
			super(out);
			this.lock = lock;
			this.key = lock.getKey();
			this.calledFrom = new Throwable();
		}

		@Override
		public void close() 
			throws IOException
		{
			if (closed) return;
			log.debug("Releasing write lock on: " + key);
			lock.writeLock().unlock();
			closed = true;
			lock = null;
			key = null;
			super.close();
		}
		
		@Override
		protected void finalize()
			throws Throwable
		{
			if (!closed) 
			{
				log.warn("Found unreleased write lock on: " + key, calledFrom);
				close();
			}
			super.finalize();
		}
	
		@Override
		public String toString()
		{
			return "LockSafeOutputStream[" + key + "]";
		}

	}
	
	
	/**
		A lock-safe input stream that releases the associated 
		read lock when the stream is closed.
		<p>
		NOTE! It is important that we keep a strong reference
		to the 'key' in this class until the stream is closed, 
		since otherwise the locked entry may be reclaimed by the 
		garbage collector.
	*/
	static class LockSafeInputStream
		extends FilterInputStream
	{
	
		private final Throwable calledFrom;
		private ReadWriteLock lock;
		private String key;
		private boolean closed;
	
		LockSafeInputStream(InputStream in, LockEntry lock)
		{
			super(in);
			this.lock = lock;
			this.key = lock.getKey();
			this.calledFrom = new Throwable();
		}
	
		@Override
		public void close() 
			throws IOException
		{
			if (closed) return;
			log.debug("Releasing read lock on: " + key);
			lock.readLock().unlock();
			closed = true;
			lock = null;
			key = null;
			super.close();
		}
		
		@Override
		protected void finalize()
			throws Throwable
		{
			if (!closed) 
			{
				log.warn("Found unreleased read lock on: " + key, calledFrom);
				close();
			}
			super.finalize();
		}
		
		@Override
		public String toString()
		{
			return "LockSafeInputStream[" + key + "]";
		}
	}

	/**
		Keeps track of a locked cached entry. The 'key' is the same key
		instance that was used to create the entry in the first place.
		It is important that all active maintainers of a lock also
		keeps a strong reference to the key since the lock entry may
		otherwise be reclaimed by the garbage collector.
	*/
	static class LockEntry
		implements ReadWriteLock
	{
		private ReadWriteLock rwLock;
		private WeakReference<String> keyRef;
		
		LockEntry(ReadWriteLock rwLock, String key)
		{
			this.rwLock = rwLock;
			this.keyRef = new WeakReference<String>(key);
		}
		
		@Override
		public Lock writeLock()
		{
			return rwLock.writeLock();
		}
		
		@Override
		public Lock readLock()
		{
			return rwLock.readLock();
		}
		
		public String getKey()
		{
			return keyRef.get();
		}
		
		@Override
		public String toString()
		{
			return "LockEntry[" + getKey() + "]";
		}
	}
	
	/**
		A timer task that clean up the cache when it is executed.
	*/
	public class CleanupTask
		extends TimerTask
	{

		private final StaticCache cache;
		private final FileFilter filter;
		
		public CleanupTask(StaticCache cache, FileFilter filter)
		{
			this.cache = cache;
			this.filter = filter;
		}
		
		@Override
		public void run()
		{
			cache.cleanUp(filter);
		}
	}
}
