/**
	$Id: AbstractBioAssaySetExporter.java 4926 2009-05-11 11:01:45Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.export.spotdata;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DatabaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DynamicPositionQuery;
import net.sf.basedb.core.DynamicResultIterator;
import net.sf.basedb.core.DynamicSpotQuery;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.VirtualColumn;
import net.sf.basedb.core.VirtualTable;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.JoinType;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Selects;
import net.sf.basedb.core.query.SqlResult;
import net.sf.basedb.core.signal.SignalException;
import net.sf.basedb.util.formatter.Formatter;
import net.sf.basedb.util.formatter.ToStringFormatter;

/**
	An abstract superclass for all exporters that exports spot data from
	a bioassay set. The purpose of this class is to collect some minimal
	configurations options (such as the source bioassay set and a DbControl)
	and to outline a generic export procedure. This class also has a lot of 
	support functionality for subclass that wants to specify exactly which 
	information that should be exported (spot data, reporter annotations, 
	assay annotations, etc.), query generation, progress reporting, thread
	interruption controls, etc.
	<p>
	
	The exporter has been designed to export three types of information
	associated with a bioassay set.
	<ul>
	<li>Spot data: intensity values, extra values, raw data values
	<li>Reporter data: reporter annotations, position extra values
	<li>Bioassay data: annotations
	</ul>
	
	<p>
	Subclass implementors can override {@link #doExport()} 
	if they wish. A better option is to override the other methods 
	that are called from this method in this order:
	<ol>
		<li>{@link #beginExport()}: One time only
		<li>{@link #exportGlobalHeader()}: One time only. If false is
			returned, the export skips to the last step
		<li>{@link #exportSectionHeader()}: One or more times
		<li>{@link #exportSectionData()}: One or more times
		<li>{@link #exportSectionFooter()}: One or more times. If true is 
			returned, the steps 3-5 are repeated.
		<li>{@link #endExport(RuntimeException)}: Always called last. If
			the parameter is not null an error has occurred
	</ol>


	@author Nicklas
	@version 2.12
	@base.modified $Date: 2009-05-11 13:01:45 +0200 (Mon, 11 May 2009) $
*/
public abstract class AbstractBioAssaySetExporter
{
	
	private BioAssaySet source;
	private DbControl dc;
	private ProgressReporter progress;
	
	private List<AssayField> assayFields;
	private List<DynamicField> reporterFields;
	private List<DynamicField> spotFields;
	private boolean averageOnReporter;

	private List<BioAssay> bioAssays;
	private Map<Integer, Object[]> reporterCache;
	private Map<Short, Integer> assayIndexMap;
	
	protected AbstractBioAssaySetExporter()
	{
		this.spotFields = new ArrayList<DynamicField>();
		this.reporterFields = new ArrayList<DynamicField>();
		this.assayFields = new ArrayList<AssayField>();
	}
	
	/*
		public Configuration properties
		-------------------------------
	*/
	/**
		Set's the DbControl that should be used to get data from the database.
	*/
	public void setDbControl(DbControl dc)
	{
		this.dc = dc;
	}
	
	/**
		Get the current DbControl.
	*/
	public DbControl getDbControl()
	{
		return dc;
	}
	
	/**
		Set the progress reporter that is used to report progress. Subclasses
		can call {@link #setProgress(int, String)} to update the current status.
	*/
	public void setProgressReporter(ProgressReporter progress)
	{
		this.progress = progress;
	}
	
	/**
		Get the progress reporter
	*/
	public ProgressReporter getProgressReporter()
	{
		return progress;
	}
	
	/**
		Set the bioassay set that is used as a source for the spot data.
	*/
	public void setSource(BioAssaySet source)
	{
		this.source = source;
	}
	
	/**
		Get the source bioassay set.
	*/
	public BioAssaySet getSource()
	{
		return source;
	}
	// -----------------------------------
	
	/*
		Optional configuration properties
		----------------------------------
	*/
	/**
		Add information about a spot field that should be exported.
		@param field The spot field (null is ignored)
	*/
	protected void addSpotField(DynamicField field)
	{
		if (field != null) spotFields.add(field);
	}
	
	/**
		Get the list of registered spot fields, in the order they were 
		registered.
	*/
	protected List<DynamicField> getSpotFields()
	{
		return spotFields;
	}
	
	/**
		Add information about a reporter field that should be exported.
		@param field The reporter field (null is ignored)
	*/
	protected void addReporterField(DynamicField field)
	{
		if (field != null) reporterFields.add(field);
	}
	
	/**
		Get the list of registered reporter fields, in the order they were 
		registered.
	*/
	protected List<DynamicField> getReporterFields()
	{
		return reporterFields;
	}
	
	/**
		Add information about an assay field that should be exported.
		@param field The assay field (null is ignored)
	*/
	protected void addAssayField(AssayField field)
	{
		if (field != null) assayFields.add(field);
	}
	
	/**
		Get the list of registered assay fields, in the order they were 
		registered.
	*/
	protected List<AssayField> getAssayFields()
	{
		return assayFields;
	}

	/**
		Specify if the exported data should be averaged on reporter
		or not.
	*/
	protected void setAverageOnReporter(boolean averageOnReporter)
	{
		this.averageOnReporter = averageOnReporter;
	}

	protected boolean getAverageOnReporter()
	{
		return averageOnReporter;
	}
	// ---------------------------------------------
	
	/**
		Run the export. 
	*/
	public void doExport()
	{
		RuntimeException e = null;
		try
		{
			beginExport();
			boolean cntinue = exportGlobalHeader();
			while (cntinue)
			{
				exportSectionHeader();
				exportSectionData();
				cntinue = exportSectionFooter();
			}
			exportGlobalFooter();
		}
		catch (RuntimeException ex)
		{
			e = ex;
			throw e;
		}
		finally
		{
			endExport(e);
		}
	}

	/*
		Export methods to be implemented by subclasses
		----------------------------------------------
	*/
	/**
		Make preparations for the export. Subclasses that need to initialise
		things should override this method. This is the first method that
		is called from {@link #doExport()} and is only called once. The
		default implementation does nothing.
	*/
	protected void beginExport()
	{}
	
	/**
		Export global headers or other data that only appear once in the
		beginning of the exported file. This method is called after {@link 
		#beginExport()} and is only called once. The return value can be used
		to indicate if the export should continue with sections or with global 
		footers. The default implementation does nothing and returns TRUE.
		@return TRUE to continue with a section {@link #exportSectionHeader()}, 
			FALSE to continue with global footers {@link #exportGlobalFooter()}
	*/
	protected boolean exportGlobalHeader()
	{
		return true;
	}
	
	/**
		Start a new section in the export by exporting section headers.
		This method is called after {@link #exportGlobalHeader()} and may
		be called multiple times as determined by the return value of
		{@link #exportSectionFooter()}. The default implementation does
		nothing.
	*/
	protected void exportSectionHeader()
	{}
	
	/**
		Export section data. This method is called after {@link #exportSectionHeader()}
		and may be called multiple times as determined by the return value of
		{@link #exportSectionFooter()}. The default implementation does
		nothing.
	*/
	protected void exportSectionData()
	{}
	
	/**
		End the section by exporting a footer. This method is called after {@link 
		#exportSectionData()}. The return value is used to determine if the export
		should be repeated with a new section or if it should move on to global footers.
		The default implementation returns false.
		@return TRUE to continue with another section {@link #exportSectionHeader()}, 
			FALSE to continue with global footers {@link #exportGlobalFooter()}
	 */
	protected boolean exportSectionFooter()
	{
		return false;
	}

	/**
		Export global footer or other data that only appears once in the
		end of the exported file. This method is called after FALSE has been
		returned from either {@link #exportGlobalHeader()} or {@link 
		#exportSectionFooter()}. The default implementation does nothing.
	*/
	protected void exportGlobalFooter()
	{}
	
	/**
		End the export and clean up/close all aquired resources. This method
		is always called last. In the case of an error the exception parameter
		has a non-null value. The default implementation clears up values that 
		has been cached so it is recommended that subclasses always call this 
		method as part of their own cleanup. If the subclass implementation 
		wants to throw a different exception it is recommended that the given 
		exception is chained to allow developers to debug problems.
	*/
	protected void endExport(RuntimeException e)
	{
		if (bioAssays != null) bioAssays.clear();
		bioAssays = null;
		if (reporterCache != null) reporterCache.clear();
		reporterCache = null;
	}
	// -------------------------------------------
	
	/*
		Helper methods
		--------------
	*/
	/**
		Get a list with all bioassays in the source bioassay set. The list
		is sorted by assay name and id. The list is cached internally and
		if this method is called several times, the same list is returned.
	*/
	protected List<BioAssay> getBioAssays()
	{
		if (bioAssays == null)
		{
			ItemQuery<BioAssay> query = getSource().getBioAssays();
			query.order(Orders.asc(Hql.property("name")));
			query.order(Orders.asc(Hql.property("id")));
			bioAssays = new ArrayList<BioAssay>(query.list(getDbControl()));
		}
		return bioAssays;
	}
	
	/**
		Adds all experimental factors from the experiment that the
		source bioassay set belongs to as assay fields. The experimental
		factors are ordered by name.
	*/
	protected void addExperimentalFactorsAsAssayFields()
	{
		// Get configuration options
		DbControl dc = getDbControl();
		BioAssaySet source = getSource();
		Experiment exp = BioAssaySet.getById(dc, source.getId()).getExperiment();

		// Load the experimental factors and register them
		ItemQuery<AnnotationType> query = exp.getExperimentalFactors();
		query.include(Include.MINE, Include.SHARED, Include.IN_PROJECT, Include.OTHERS);
		query.order(Orders.asc(Hql.property("name")));
		for (AnnotationType at : query.list(dc))
		{
			AnnotationAssayField af = new AnnotationAssayField();
			af.setAnnotationType(at);
			addAssayField(af);
		}
	}
	
	/**
		Get a configured query for spot data that returns all
		fields that has been registered with {@link #addSpotField(DynamicField)}.
		The fields are selected in the order they were registered. 
		This method also adds the position and column of the spot as the
		last two fields. Note! If the {@link #setAverageOnReporter(boolean)}
		option has been enabled there is no unique position so the query
		returns the reporter id instead of the position in this case.
		
		@param bioAssayRestriction If TRUE a restriction is added to the 
			query to make it return data for a single bioassay only.
			Use <code>query.setParameter("bioAssayColumn", (int)bioAssay.getDataCubeColumnNo(), Type.INT)</code> 
			to set the value for the restriction
		@return A prepared query
	*/
	protected DynamicSpotQuery getSpotQuery(boolean bioAssayRestriction)
	{
		// Get configuration options
		DbControl dc = getDbControl();
		BioAssaySet source = getSource();
		boolean forAverage = getAverageOnReporter();
		
		// Create and prepare query
		DynamicSpotQuery query = source.getSpotData();
		if (forAverage) query.joinReporters(JoinType.INNER);
		
		// Add spot fields
		int index = 0;
		Expression emptyExpression = Expressions.string("");
		for (DynamicField f : getSpotFields())
		{
			Expression e = f.getExpression(dc, source, forAverage);
			if (e == null)	e = emptyExpression;
			query.select(Selects.expression(e, "f" + index));
			index++;
		}
		
		// Add position and column fields
		Expression posExpression = null;
		Expression colExpression = Dynamic.column(VirtualColumn.COLUMN);
		if (forAverage)
		{
			posExpression = Dynamic.column(VirtualTable.POSITION, VirtualColumn.REPORTER_ID);
			query.group(colExpression);
			query.group(posExpression);
		}
		else
		{
			posExpression = Dynamic.column(VirtualColumn.POSITION);
		}
		query.select(Selects.expression(posExpression, "pos"));
		query.select(Selects.expression(colExpression, "col"));
		query.order(Orders.asc(posExpression));
		query.order(Orders.asc(colExpression));
		
		// Optional restriction
		if (bioAssayRestriction)
		{
			query.restrict(Restrictions.eq(colExpression, Expressions.parameter("bioAssayColumn")));
		}
		return query;
	}
	
	/**
		Get a configured query for position data that returns all
		fields that has been registered with {@link #addReporterField(DynamicField)}.
		The fields are selected in the order they were registered. This method also 
		adds the position of the spot as the last field. 
		<p>
		Note 1! If the {@link #setAverageOnReporter(boolean)}
		option has been enabled there is no unique position so the query
		returns the reporter id instead of the position in this case.
		<p>
		Note 2!  If the {@link #setAverageOnReporter(boolean)}
		option has been enabled the query may return the same reporter
		information multiple times.
		<p>
		Note 3! A subclass may override this method, but it should at least still
		select the same number of fields as is expected from the default 
		implementation.
		@return A prepared query
	*/
	protected DynamicPositionQuery getReporterQuery()
	{
		// Get configuration options
		DbControl dc = getDbControl();
		BioAssaySet source = getSource();
		boolean forAverage = getAverageOnReporter();

		// Create and prepare query
		DynamicPositionQuery query = source.getPositionData();
		query.joinReporters(JoinType.LEFT);
		
		// Add reporter fields
		int index = 0;
		Expression emptyExpression = Expressions.string("");
		for (DynamicField f : getReporterFields())
		{
			Expression e = f.getExpression(dc, source, false);
			if (e == null)	e = emptyExpression;
			query.select(Selects.expression(e, "f" + index));
			index++;
		}
		
		// Add position field
		Expression posExpression = null;
		if (forAverage)
		{
			posExpression = Dynamic.column(VirtualTable.POSITION, VirtualColumn.REPORTER_ID);
		}
		else
		{
			posExpression = Dynamic.column(VirtualColumn.POSITION);
		}
		query.select(Selects.expression(posExpression, "pos"));

		return query;
	}

	/**
		Loads and caches reporter data in an internal structure
		for quick access. This method will execute the query returned
		by {@link #getReporterQuery()} and read all data into a memory
		structure. The data is indexed by position number or by reporter
		id (if average on reporters is enabled).
	*/
	@SuppressWarnings("unchecked")
	protected int cacheReporterData()
	{
		// Get configuration options
		DbControl dc = getDbControl();
		BioAssaySet source = getSource();
		List<DynamicField> reporterFields = getReporterFields();
		int numReporterFields = reporterFields.size();
		int posIndex = numReporterFields + 1;
		Formatter[] formatters = new Formatter[numReporterFields];
		for (int i = 0; i < numReporterFields; ++i)
		{
			Formatter f = reporterFields.get(i).getFormatter();
			if (f == null) f = new ToStringFormatter();
			formatters[i] = f;
		}
		
		// Get reporter query
		DynamicPositionQuery query = getReporterQuery();
		
		reporterCache = new HashMap<Integer, Object[]>(source.getNumSpots());
		DynamicResultIterator it = query.iterate(dc);
		try
		{
			while (it.hasNext())
			{
				checkInterrupted();
				SqlResult result = it.next();
				int position = result.getInt(posIndex);
				if (!reporterCache.containsKey(position))
				{
					Object[] data = new Object[numReporterFields];
					int index = 0;
					for (int i = 1; i <= numReporterFields; ++i)
					{
						data[index] = formatters[index].format(result.getObject(i));
						++index;
					}
					reporterCache.put(position, data);
				}
			}
		}
		catch (SQLException e)
		{
			throw new DatabaseException(e);
		}
		finally
		{
			if (it != null) it.close();
		}
		return reporterCache.size();
	}
	
	/**
		Prepare a cache that maps the bioassay column ({@link BioAssay#getDataCubeColumnNo()})
		value to an index in the 'data' array this is generated. The map is generated as
		follows:
		<ul>
		<li>first assay -&gt; firstIndex
		<li>second assay -&gt; firstIndex + columnsPerAssay
		<li>third assay -&gt; firtsIndex + 2 * columnsPerAssay
		<li>... and so on
		</ul>
		<p>
		Use {@link #getAssayIndex(short)} to get the indexed values.
		
		@param assays The assays that should be mapped
		@param firstIndex The index of the first assay
		@param columnsPerAssay Number of columns for each assay
	*/
	protected void prepareAssayIndexMap(List<BioAssay> assays, int firstIndex, int columnsPerAssay)
	{
		if (assayIndexMap == null) assayIndexMap = new HashMap<Short, Integer>(assays.size());
		int index = firstIndex;
		for (BioAssay ba : assays)
		{
			assayIndexMap.put(ba.getDataCubeColumnNo(), index);
			index += columnsPerAssay;
		}
	}
	
	/**
		Get the cached index value for an assay when you know the assay's column
		number.
		@param column The column number ({@link BioAssay#getDataCubeColumnNo()})
		@return The index of the first position in the 'data' array that
			data for the assay should be written to.
		@see #prepareAssayIndexMap(List, int, int)
	*/
	protected int getAssayIndex(short column)
	{
		return assayIndexMap.get(column);
	}
	
	/**
		Copy cached reporter data into a data array. 
		@param key The cache key (=position or reporter id)
		@param data The data array to copy to
		@param startIndex The start index in the data array for the
			first reporter field
	*/
	protected void copyReporterFields(int key, Object[] data, int startIndex)
	{
		Object[] reporterData = reporterCache.get(key);
		if (reporterData != null)
		{
			System.arraycopy(reporterData, 0, data, startIndex, reporterData.length);
		}
	}
	
	/**
		Checks if the currently executing thread has been interrupted and
		throws a {@link SignalException} if it has.
	*/
	protected void checkInterrupted()
	{
		if (Thread.interrupted()) throw new SignalException("Aborted by user");
	}
	
	/**
		Update the progress of the export.
		@see ProgressReporter
	*/
	protected void setProgress(int percent, String message)
	{
		if (progress != null)
		{
			progress.display(percent, message);
		}
	}
	// ----------------------------
}
