/**
	$Id: LookupFormatter.java 4857 2009-03-27 11:47:31Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.formatter;

import java.util.HashMap;
import java.util.Map;

import net.sf.basedb.util.Enumeration;

/**
	A formatter implementation that uses a map to lookup values from
	a set of keys. A default value may be specified that is returned
	in case no lookup has been defined for a given key.
	
	@author nicklas
	@version 2.12
	@base.modified $Date: 2009-03-27 12:47:31 +0100 (Fri, 27 Mar 2009) $
*/
public class LookupFormatter<T>
	implements Formatter<T>
{
	private final Map<T, String> map;
	private String defaultValue;
	
	/**
		Create a new lookup formatter
	*/
	public LookupFormatter()
	{
		this.map = new HashMap<T, String>();
	}
	
	/**
		Create a new lookup formatter and initialise it with
		the key-value pairs in the given map. The key value pairs
		are copied to a new map internal to the lookup formatter.
		@param in A map object or null
	*/
	public LookupFormatter(Map<? extends T, ? extends String> in)
	{
		this();
		if (in != null) putAll(in);
	}
	
	/**
		Create a new lookup formatter and initialise it with
		the key-value pairs in the given enumeration. The key value pairs
		are copied to a new map internal to the lookup formatter.
		@param in An enumeration object or null
	*/
	public LookupFormatter(Enumeration<? extends T, ? extends String> in)
	{
		this();
		if (in != null) putAll(in);
	}

	/*
		From the Formatter interface
		-------------------------------------------
	*/
	@Override
	public String format(T key)
	{
		String value = defaultValue;
		if (map.containsKey(key))
		{
			value = map.get(key);
		}
		return value;
	}
	public T parseString(String value)
	{
		throw new UnsupportedOperationException("parseString: " + value);
	}
	// -------------------------------------------

	/**
		Set a default value that is returned by the {@link #format(Object)}
		method in case no lookup has been specified for a key.
	*/
	public void setDefaultValue(String defaultValue)
	{
		this.defaultValue = defaultValue;
	}
	
	/**
		Get the default value.
	*/
	public String getDefaultValue()
	{
		return defaultValue;
	}
	
	/**
		Add a lookup key/value pair.
		@param key The key
		@param value The value that is returned as the lookup value for the given key
	*/
	public void put(T key, String value)
	{
		map.put(key, value);
	}
	
	/**
		Copy multiple lookup key/value pairs from a map
		@param in The map to copy key/value pair from
	*/
	public void putAll(Map<? extends T, ? extends String> in)
	{
		map.putAll(in);
	}
	
	/**
		Copy multiple lookup key/value pairs from an enumeration.
		@param in The enumeration to copy key/value pair from
	*/
	public void putAll(Enumeration<? extends T, ? extends String> in)
	{
		for (int i = 0; i < in.size(); ++i)
		{
			map.put(in.getKey(i), in.getValue(i));
		}
	}
	
	/**
		Checks if a given key has been registered.
		@param key The key
		@return TRUE if an entry exists for the key, or FALSE otherwise
	*/
	public boolean containsKey(Object key)
	{
		return map.containsKey(key);
	}
	
	/**
		Get registered lookup value for a given key
		@param key The key
		@return The registered lookup value, or null if no value has
			been registered
	*/
	public String get(Object key)
	{
		return map.get(key);
	}

	/**
		Get the number of registered lookup entries.
	*/
	public int size()
	{
		return map.size();
	}
	
	/**
		Checks if the lookup map is empty or not.
	*/
	public boolean isEmpty()
	{
		return map.isEmpty();
	}
	
	/**
		Remove all registered lookup entries.
	*/
	public void clear()
	{
		map.clear();
	}
	
	/**
		Remove an entry from the lookup map.
		@param key The key
		@return The value that was associated with the key
	*/
	public String remove(Object key)
	{
		return map.remove(key);
	}
	
}
