/**
	$Id: GenericOverview.java 4789 2009-02-24 10:07:07Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Project;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Order;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.util.overview.cache.NodeCache;
import net.sf.basedb.util.overview.loader.NodeLoader;
import net.sf.basedb.util.overview.loader.BasicItemNodeLoaderFactory;
import net.sf.basedb.util.overview.loader.NodeLoaderFactory;
import net.sf.basedb.util.overview.validator.BasicItemNodeValidatorFactory;
import net.sf.basedb.util.overview.validator.NodeValidatorFactory;

/**
	Implementation for holding a generic overview of items. The
	overview can be rooted on all main items that have a {@link NodeLoader}
	implementation that supports it.
	<p>
	This class is currently hardcoded to use a {@link BasicItemNodeLoaderFactory}
	and {@link BasicItemNodeValidatorFactory} as factories for the node
	loaders and validators. This may change in future releases. Use {@link 
	#getNodeLoaderFactory()} and {@link #getNodeValidatorFactory()} to get
	access to the factories.
	<p>
	The root item of the overview is specified in the constructor, as is
	a project that will be used for validating against project defaults. 
	We recommend that the currently active project is used.
	<p>
	The overview has been designed to support lazy loading. To begin with 
	the overview will only contain the root item. Use {@link #expand(DbControl, Node, boolean)}
	to expand a node. If the <code>recursive</code> parameter is true the loading
	will continue until the "end" is reached, otherwise only the immediate children
	will be loaded. Note! The lazy loading functionality has not yet been tested
	(as of BASE 2.10) and may not work in all cases due to bugs in the code.	
	To create an overview and load the entire tree in one go, use:
	<pre class="code">
DbControl dc = ...
BasicItem item = ...
Project project = ...
GenericOverview go = new GenericOverview(dc, item, project);
go.expand(go.getRootNode().getId(), dc, true);
go.updateFailureCountOnNodes();
</pre>

	@author Nicklas
	@version 2.10
	@base.internal This class is still being developed. It may change 
		without notice in future versions of BASE.
	@base.modified $Date: 2009-02-24 11:07:07 +0100 (Tue, 24 Feb 2009) $
*/
public class GenericOverview
	implements OverviewContext
{
	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.util.overview.GenericOverview");
	private static final boolean debug = log.isDebugEnabled();


	private final ValidationOptions validationOptions;
	private final BasicItemNodeLoaderFactory nodeLoaderFactory;
	private final BasicItemNodeValidatorFactory nodeValidatorFactory;
	private final Project project;
	private final BasicItem rootItem;
	private Node rootNode;
	private Map<String, Node> allNodes;
	private NodeCache<Object> nodeCache;
	private Map<Object, Object> objectCache;
	private List<Failure> failures;

	/**
		Create a new generic overview. 
		
		@param dc A DbControl to use for database access
		@param rootItem The root item of the overview
		@param project A project to use for validation of project default 
			(we recommend that the currently active project is used)
	*/
	public GenericOverview(DbControl dc, BasicItem rootItem, Project project)
	{
		this.nodeLoaderFactory = new BasicItemNodeLoaderFactory();
		this.nodeValidatorFactory = new BasicItemNodeValidatorFactory();
		this.validationOptions = new ValidationOptions(
			OverviewUtil.getValidationPresets(dc.getSessionControl()).getDefault());
		this.rootItem = rootItem;
		this.project = project;
		reset(dc);
	}

	/**
		Get the current node loader factory. NOTE! This is currently (BASE 2.10)
		hardcoded to be a {@link BasicItemNodeLoaderFactory} but may change in 
		(=be configurable) in future releases.
	*/
	@Override
	public NodeLoaderFactory<BasicItem, Object> getNodeLoaderFactory()
	{
		return nodeLoaderFactory;
	}
	
	/**
		Get the current node validator factory. NOTE! This is currently (BASE 2.10)
		hardcoded to be a {@link BasicItemNodeValidatorFactory} but may change in 
		(=be configurable) in future releases.
	*/
	@Override
	public NodeValidatorFactory getNodeValidatorFactory()
	{
		return nodeValidatorFactory;
	}

	/**
		Get the project of this overview is using for validating against
		project defaults (may be null).
	*/
	@Override
	public Project getProject()
	{
		return project;
	}

	/**
		Retreive an object stored in the overview cache.
		@param cacheKey The key to use for cache lookup
		@return The object, or null if no object has been cached with the
			given key
	*/
	@Override
	public Object getCachedObject(Object cacheKey) 
	{
		return objectCache.get(cacheKey);
	}

	/**
		Store an auxilliary object in the overview cache. This is a service
		provided by the overview implementation that can be used by node
		loaders and/or validators to store information that may be needed
		again in the future and is	expensive to retreive (eg. from the database). 
		This is for example used to store annotation types that have been marked 
		as "Required by MIAME", parents to annotatable items and more.
		@param cacheKey The key to store the value under
		@param value The value to store in the cache, or null to
			remove the entry
	*/
	@Override
	public void setCachedObject(Object cacheKey, Object value) 
	{
		if (value == null)
		{
			objectCache.remove(cacheKey);
		}
		else
		{
			objectCache.put(cacheKey, value);
		}
	}

	
	@Override
	public NodeCache<Object> getNodeCache()
	{
		return nodeCache;
	}
	
	/**
		Register a failure of a validation rule.
		@param validator The validation rule that fails
		@param node The node that was validated
		@param message An optional message, if null the message from
			{@link Validator#getFailureSummary()} is used
		@param fixes Optional fixes that may be used to correct the problem
	*/
	@Override
	public Failure createFailure(Validator validator, Node node, String message, Fix... fixes)
	{
		Failure f = new Failure(validator, node, message, fixes);
		failures.add(f);
		if (debug) log.debug(f.getMessage() + " on node " + node.getName());
		return f;
	}

	/**
		Initialise a query with "global options" and optionally a sort
		order. Note! The "global options" currently includes setting
		{@link ItemQuery#include(java.util.Collection)} to {@link Include#ALL},
		but this may change/be configurable in the future.
		
		@param query The query that should be initialised
		@param sortby An array with Order objects (may be null)
		@return The same query object
	*/
	@Override
	public <I extends BasicItem> ItemQuery<I> initQuery(ItemQuery<I> query, Order... sortby)
	{
		query.include(Include.ALL);
		if (sortby != null)
		{
			for (Order o : sortby)
			{
				if (o != null) query.order(o);
			}
		}
		return query;
	}

	/**
		Same as {@link #initQuery(ItemQuery, Order...)} but easier to use when
		sorting the items in ascending order by a single column.
		@param query The query that should be initialised
		@param sortby The property name to sort by (may be null)
		@return The same query object
	*/
	@Override
	public <I extends BasicItem> ItemQuery<I> initQuery(ItemQuery<I> query, String sortby)
	{
		Order o = sortby == null ? null : Orders.asc(Hql.property(sortby));
		return initQuery(query, o);
	}
	
	/**
		Get the root item of this overview.
	*/
	public BasicItem getRootItem()
	{
		return rootItem;
	}
	
	/**
		Get the root node.
	*/
	public Node getRootNode()
	{
		return rootNode;
	}
	
	/**
		Load children of a node. If the children has already been loaded
		({@link Node#isChildrenLoaded()} == true) this method does nothing.
		
		@param dc The DbControl to use for database access
		@param node The node to expand
		@param recursive TRUE if children should be loaded recursively, FALSE to
			only load the immediate children
	*/
	public void expand(DbControl dc, Node node, boolean recursive)
	{
		if (node == null) return;
		if (!node.isChildrenLoaded())
		{
			if (debug) log.debug("Expanding: " + node + (recursive ? " (recursive)" : ""));
			nodeLoaderFactory.setAutoLoadChildren(recursive);
			NodeLoader loader = nodeLoaderFactory.createNodeLoader(node.getItemType());
			loader.loadChildNodes(dc, this, node);
			cacheChildNodes(node);
		}
		else
		{
			if (debug) log.debug("Already expanded: " + node);
		}
		nodeLoaderFactory.setAutoLoadChildren(false);
	}

	
	/**
		Get the validation options currently in use.
	*/
	public ValidationOptions getValidationOptions()
	{
		return validationOptions;
	}

	/**
		Updates the failure count (warnings and errors) on all nodes.
		This method should be called if the validation options has 
		changed. If the struture or information of the experiment has
		changed use the {@link #reset(DbControl)} method instead.
	*/
	public void updateFailureCountOnNodes()
	{
		rootNode.clearFailures();
		if (failures == null) return;
		for (Failure f : failures)
		{
			Severity s = validationOptions.getSeverity(f.getValidator());
			Node n = f.getNode();
			if (s == Severity.ERROR)
			{
				n.addErrors(1);
			}
			else if (s == Severity.WARNING)
			{
				n.addWarnings(1);
			}
		}
	}
	
	/**
		Reset the overview and re-create the root node. 
		@param dc A DbControl to use for database access.
	*/
	public void reset(DbControl dc)
	{
		this.allNodes = new HashMap<String, Node>();
		//this.nodeCache = new NodeCache<Object>();
		this.objectCache = new HashMap<Object, Object>();
		this.failures = new LinkedList<Failure>();
		this.rootNode = createRootNode(dc, rootItem);
		allNodes.put(rootNode.getId(), rootNode);
	}

	/**
		Get a list containing all validation failures.
		@return A list (may be null)
	*/
	public List<Failure> getFailures()
	{
		return failures;
	}
	
	/**
		Get the node with the given ID.
		@param nodeId The ID of the node
		@return A node or null if no node with the ID exists
	*/
	public Node getNode(String nodeId)
	{
		return allNodes.get(nodeId);
	}
	
	@SuppressWarnings("unchecked")
	private Node createRootNode(DbControl dc, BasicItem root)
	{
		if (debug) log.debug("Creating root node for item: " + root);
		NodeLoader loader = nodeLoaderFactory.createNodeLoader(root.getType());
		Node node = loader.createRootNode(dc, this, root);
		return node;
	}
	
	private void cacheChildNodes(Node node)
	{
		if (node.getChildren() != null)
		{
			for (Node child : node.getChildren())
			{
				allNodes.put(child.getId(), child);
				cacheChildNodes(child);
			}
		}
	}

}
