/**
	$Id: Validator.java 4789 2009-02-24 10:07:07Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview;

import net.sf.basedb.util.overview.validator.NodeValidator;

/**
	Placeholder for a validation rule. An object instance contains a 
	name and description of the validation and a default severity level.
	<p>
	The actual validation takes place in the overview generating code.
	See implementations of {@link NodeValidator} interface.
	
	@author Nicklas
	@version 2.2
	@base.modified $Date: 2009-02-24 11:07:07 +0100 (Tue, 24 Feb 2009) $
*/
public class Validator
{

	/**
		Validator rule that checks for missing protocols. Default severity level
		is WARNING.
	*/
	public static final Validator MISSING_PROTOCOL = new Validator(
			"missing.protocol", "Missing protocol", 
			"Checks if a protocol has been specified for items that can have a protocol",
			"Missing protocol", Severity.WARNING
		);

	/**
		Validator rule that checks for denied access to protocols. Default severity level
		is WARNING.
	*/
	public static final Validator DENIED_PROTOCOL = new Validator(
			"denied.protocol", "Access denied to protocol", 
			"Checks if the user is denied access to a protocol",
			"Access denied to protocol", Severity.WARNING
		);

	/**
		Validator rule that checks for missing scan. Default severity level
		is WARNING.
	*/
	public static final Validator MISSING_SCAN = new Validator(
			"missing.scan", "Missing scan", 
			"Checks if a scan has been specified for a raw bioassay",
			"Missing scan", Severity.WARNING
		);
	
	/**
		Validator rule that checks for denied access to scans. Default severity level
		is WARNING.
	*/
	public static final Validator DENIED_SCAN = new Validator(
			"denied.scan", "Access denied to scan", 
			"Checks if the user is denied access to a scan",
			"Access denied to scan", Severity.WARNING
		);
	
	/**
		Validator rule that checks for denied access to platform. 
		Default severity level is WARNING.
	*/
	public static final Validator DENIED_PLATFORM = new Validator(
			"denied.platform", "Access denied to platform", 
			"Checks if the user is denied access to a platform",
			"Access denied to platform", Severity.WARNING
		);

	/**
		Validator rule that checks for denied access to (data) files. 
		Default severity level is WARNING.
	*/
	public static final Validator DENIED_FILE = new Validator(
			"denied.file", "Access denied to file", 
			"Checks if the user is denied access to a file",
			"Access denied to file", Severity.WARNING
		);
	
	/**
		Validator rule that checks for denied access to data file types. 
		Default severity level is WARNING.
	*/
	public static final Validator DENIED_DATAFILETYPE = new Validator(
			"denied.datafiletype", "Access denied to data file type", 
			"Checks if the user is denied access to a data file type",
			"Access denied to data file type", Severity.WARNING
		);
	
	/**
		Validator rule that checks for missing software. Default severity level
		is WARNING.
	*/
	public static final Validator MISSING_SOFTWARE = new Validator(
			"missing.software", "Missing software", 
			"Checks if a software has been specified for a raw bioassay",
			"Missing software", Severity.WARNING
		);
	
	/**
		Validator rule that checks for denied access to software. Default severity level
		is WARNING.
	*/
	public static final Validator DENIED_SOFTWARE = new Validator(
			"denied.software", "Access denied to software", 
			"Checks if the user is denied access to a software",
			"Access denied to software", Severity.WARNING
		);
	
	/**
		Validator rule that checks for missing hardware. Default severity level
		is WARNING.
	*/
	public static final Validator MISSING_HARDWARE = new Validator(
			"missing.hardware", "Missing hardware", 
			"Checks if a hardware (scanner, print robot, etc.) has been specified for " +
			"item that can have a hardware.",
			"Missing hardware", Severity.WARNING
		);

	/**
		Validator rule that checks for denied access to hardware. Default severity level
		is WARNING.
	*/
	public static final Validator DENIED_HARDWARE = new Validator(
			"denied.hardware", "Access denied to hardware", 
			"Checks if the user is denied access to a hardware",
			"Access denied to hardware", Severity.WARNING
		);

	/**
		Validator rule that checks for missing hybridization. Default severity level
		is WARNING.
	*/
	public static final Validator MISSING_HYBRIDIZATION = new Validator(
			"missing.hybridization", "Missing hybridization", 
			"Checks if a hybridization has been specified for a scan or array slide",
			"Missing hybridization", Severity.WARNING
		);

	/**
		Validator rule that checks for denied access to hybridization. Default severity level
		is WARNING.
	*/
	public static final Validator DENIED_HYBRIDIZATION = new Validator(
			"denied.hybridization", "Access denied to hybridization", 
			"Checks if the user is denied access to a hybridization",
			"Access denied to hybridization", Severity.WARNING
		);

	/**
		Validator rule that checks for missing array slide. Default severity level
		is WARNING.
	*/
	public static final Validator MISSING_ARRAYSLIDE = new Validator(
			"missing.slide", "Missing array slide", 
			"Checks if an array slide has been specified for a hybiridzation",
			"Missing array slide", Severity.WARNING
		);

	/**
		Validator rule that checks for denied access to array slides. Default severity level
		is WARNING.
	*/
	public static final Validator DENIED_ARRAYSLIDE = new Validator(
			"denied.slide", "Access denied to array slide", 
			"Checks if the user is denied access to an array slide",
			"Access denied to array slide", Severity.WARNING
		);

	/**
		Validator rule that checks for missing labeled extract. Default severity level
		is WARNING.
	*/
	public static final Validator MISSING_LABELED_EXTRACT = new Validator(
			"missing.labeledextract", "Missing labeled extract", 
			"Checks if at least one labeled extract has been specified as parent to a " +
			"pooled labeled extract",
			"Missing labeled extract", Severity.WARNING
		);
	
	/**
		Validator rule that checks for missing extract. Default severity level
		is WARNING.
	*/
	public static final Validator MISSING_EXTRACT = new Validator(
			"missing.extract", "Missing extract", 
			"Checks if an extract has been specified for a labled extract or pooled extract",
			"Missing extract", Severity.WARNING
		);

	/**
		Validator rule that checks for denied access to extracts. Default severity level
		is WARNING.
	*/
	public static final Validator DENIED_EXTRACT = new Validator(
			"denied.extract", "Access denied to extract", 
			"Checks if the user is denied access to a extract",
			"Access denied to extract", Severity.WARNING
		);

	
	/**
		Validator rule that checks for missing sample. Default severity level
		is WARNING.
	*/
	public static final Validator MISSING_SAMPLE = new Validator(
			"missing.sample", "Missing sample", 
			"Checks if a sample has been specified for an extract or pooled sample",
			"Missing sample", Severity.WARNING
		);

	/**
		Validator rule that checks for denied access to samples. Default severity level
		is WARNING.
	*/
	public static final Validator DENIED_SAMPLE = new Validator(
			"denied.sample", "Access denied to sample", 
			"Checks if the user is denied access to a sample",
			"Access denied to sample", Severity.WARNING
		);

	
	/**
		Validator rule that checks for missing biosource. Default severity level
		is WARNING.
	*/
	public static final Validator MISSING_BIOSOURCE = new Validator(
			"missing.biosource", "Missing biosource", 
			"Checks if a biosource has been specified for a non-pooled sample",
			"Missing biosource", Severity.WARNING
		);

	/**
		Validator rule that checks for denied access to biosource. Default severity level
		is WARNING.
	*/
	public static final Validator DENIED_BIOSOURCE = new Validator(
			"denied.biosource", "Access denied to biosource", 
			"Checks if the user is denied access to a biosource",
			"Access denied to biosource", Severity.WARNING
		);

	/**
		Validator rule that checks for denied access to labels. Default severity level
		is WARNING.
	*/
	public static final Validator DENIED_LABEL = new Validator(
			"denied.label", "Access denied to label", 
			"Checks if the user is denied access to a label",
			"Access denied to label", Severity.WARNING
		);
	
	/**
		Validator rule that checks for missing array design. Default severity level
		is WARNING.
	*/
	public static final Validator MISSING_ARRAYDESIGN = new Validator(
			"missing.arraydesign", "Missing array design", 
			"Checks if an array design has been specified for a raw bioassay",
			"Missing array design", Severity.WARNING
		);

	/**
		Validator rule that checks for denied access to an array design. Default severity level
		is WARNING.
	*/
	public static final Validator DENIED_ARRAYDESIGN = new Validator(
			"denied.arraydesign", "Access denied to array design", 
			"Checks if the user is denied access to an array design",
			"Access denied to array design", Severity.WARNING
		);

	/**
		Validator rule that checks for missing array batch. Default severity level
		is WARNING.
	*/
	public static final Validator MISSING_ARRAYBATCH = new Validator(
			"missing.arraybatch", "Missing array batch", 
			"Checks if an array batch has been specified for an array slide",
			"Missing array batch", Severity.WARNING
		);
	
	/**
		Validator rule that checks for denied access to an array batch. Default severity level
		is WARNING.
	*/
	public static final Validator DENIED_ARRAYBATCH = new Validator(
			"denied.arraybatch", "Access denied to array batch", 
			"Checks if the user is denied access to an array batch",
			"Access denied to array batch", Severity.WARNING
		);

	
	/**
		Validator rule that checks for missing experimental factor value. 
		Default severity level is ERROR.
	*/
	public static final Validator MISSING_FACTORVALUE = new Validator(
			"missing.factorvalue", "Missing factor value", 
			"Checks if a value has been specified for all experimental factors for a raw bioassay",
			"Missing factor value", Severity.ERROR
		);

	/**
		Validator rule that checks for missing protocol parameter values. 
		Default severity level is ERROR.
	*/
	public static final Validator MISSING_PARAMETER = new Validator(
			"missing.parameter", "Missing parameter value", 
			"Checks if a parameter value has been specified for all protocol parameters",
			"Missing parameter value", Severity.ERROR
		);

	/**
		Validator rule that checks for missing "Required for MIAME" annotation. 
		Default severity level is ERROR.
	*/
	public static final Validator MISSING_MIAME_ANNOTATION = new Validator(
			"missing.miameannotation", "Missing MIAME annotation value", 
			"Checks if an annotation value has been specified for a 'Required for MIAME' annotation type",
			"Missing MIAME annotation value", Severity.ERROR
		);

	/**
		Validator rule that checks for denied access to an annotation type. 
		Default severity level is WARNING.
	*/
	public static final Validator DENIED_ANNOTATIONTYPE = new Validator(
			"denied.annotationtype", "Access denied to annotation type", 
			"Checks if the user is denied access to an annotation type",
			"Access denied to annotation type", Severity.WARNING
		);
	
	/**
		Validator rule that checks if an item uses a different platform / variant than
		specified by the project default. Default severity level is WARNING.
	*/
	public static final Validator NONDEFAULT_PLATFORM = new Validator(
			"nondefault.platform", "Use of non-default platform",
			"Checks if an item uses a different platform than the default for the current project",
			"Use of non-default platform", Severity.WARNING
		);

	/**
		Validator rule that checks if an item uses a different protocol than
		specified by the project default. Default severity level is WARNING.
	*/
	public static final Validator NONDEFAULT_PROTOCOL = new Validator(
			"nondefault.protocol", "Use of non-default protocol",
			"Checks if an item uses a different protocol than the default for the current project",
			"Use of non-default protocol", Severity.WARNING
		);

	/**
		Validator rule that checks if an item uses a different hardware than
		specified by the project default. Default severity level is WARNING.
	*/
	public static final Validator NONDEFAULT_HARDWARE = new Validator(
			"nondefault.hardware", "Use of non-default hardware",
			"Checks if an item uses a different hardware (scanner, print robot, etc.) " +
			"than the default for the current project",
			"Use of non-default hardware", Severity.WARNING
		);
	
	/**
		Validator rule that checks if an item uses a different software than
		specified by the project default. Default severity level is WARNING.
	*/
	public static final Validator NONDEFAULT_SOFTWARE = new Validator(
			"nondefault.software", "Use of non-default software",
			"Checks if an item uses a different software (feature extraction) " +
			"than the default for the current project",
			"Use of non-default software", Severity.WARNING
		);

	/**
		Validator rule that checks if an item uses a different array design than
		specified by the project default. Default severity level is WARNING.
	*/
	public static final Validator NONDEFAULT_ARRAYDESIGN = new Validator(
			"nondefault.arraydesign", "Use of non-default array design",
			"Checks if an item uses a different array design " +
			"than the default for the current project",
			"Use of non-default array design", Severity.WARNING
		);
	
	/**
		Validator rule that checks if an item uses a different raw data type than
		specified by the project default. Default severity level is WARNING.
	*/
	public static final Validator NONDEFAULT_RAWDATATYPE = new Validator(
			"nondefault.rawdatatype", "Use of non-default raw data type",
			"Checks if an experiment uses the same raw data type as specified by the " +
			"default for the current project",
			"Use of non-default raw data type", Severity.WARNING
		);
	
	/**
		Validator rule that checks if all raw bioassays in an experiment
		uses the same array design. Default severity level is WARNING.
	*/
	public static final Validator MULTIPLE_ARRAYDESIGNS = new Validator(
			"misc.multiplearraydesigns", "Multiple array designs",
			"Checks if all raw bioassays in an experiment uses the same " +
			"array design.",
			"Multiple array designs", Severity.WARNING
		);

	/**
		Validator rule that checks if the array design for a raw bioassay
		matches the array design of the associated hybridization.
		Default severity level is WARNING.
	*/
	public static final Validator NONMATHCING_ARRAYDESIGN = new Validator(
			"misc.nonmatchingarraydesigns", "Array design mismatch",
			"Checks if the array design for a raw bioassay is the same " +
			"array design as of the associated hybridization.",
			"Array design mismatch", Severity.WARNING
		);

	/**
		Validator rule that checks if the protocol attached to an item
		has a correct protocol type. Default severity level is WARNING.
	*/
	public static final Validator INCORRECT_PROTOCOLTYPE = new Validator(
			"misc.incorrectprotocoltype", "Incorrect protocol type",
			"Checks if the protocol attached to an item has a correct protocol type.",
			"Incorrect protocol type", Severity.WARNING
		);

	/**
		Validator rule that checks if the software attached to an item
		has a correct software type. Default severity level is WARNING.
	*/
	public static final Validator INCORRECT_SOFTWARETYPE = new Validator(
			"misc.incorrectsoftwaretype", "Incorrect software type",
			"Checks if the software attached to an item has a correct software type.",
			"Incorrect software type", Severity.WARNING
		);

	/**
		Validator rule that checks if the hardware attached to an item
		has a correct hardware type. Default severity level is WARNING.
	*/
	public static final Validator INCORRECT_HARDWARETYPE = new Validator(
			"misc.incorrecthardwaretype", "Incorrect hardware type",
			"Checks if the hardware attached to an item has a correct hardware type.",
			"Incorrect hardware type", Severity.WARNING
		);
	
	/**
		Validator rule that checks if the sub-array index on a raw bioassay
		or labeled extract is between 1 and the number of sub-arrays on the 
		hybridization. Default severity level is WARNING.
	*/
	public static final Validator INCORRECT_ARRAY_NUM = new Validator(
			"misc.incorrectarraynum", "Incorrect sub-array index",
			"Checks if the sub-array index on a raw bioassay or labeled extract " +
			"is between 1 and the number of sub-arrays on the hybridization.",
			"Incorrect sub-array index", Severity.WARNING
		);

	/**
		Validator rule that checks if the same sub-array index is used
		on more than one raw bioassay from the same scan. 
		Default severity level is WARNING.
	*/
	public static final Validator DUPLICATE_ARRAY_NUM = new Validator(
			"misc.duplicatearraynum", "Duplicate sub-array index",
			"Checks if the same sub-array index is used on more " +
			"than one raw bioassay from the same scan.",
			"Duplicate sub-array index", Severity.WARNING
		);
	
	/**
		Validator rule that checks if the number of labeled extracts
		attached to a hybridization matches the number of channels in the
		experiment. Default severity level is ERROR.
	*/
	public static final Validator INCORRECT_NUM_LABELEDEXTRACTS = new Validator(
			"misc.incorrectnumlabeledextracts", "Incorrect number of labeled extracts",
			"Checks if the the number of labeled extracts attached to a hybridzation matches " +
			"the number of channels in the experiment.",
			"Incorrect number of labeled extracts", Severity.ERROR
		);
	
	/**
		Validator rule that checks if a hybridization has more than one
		labeled extract with the same label. Default severity level is WARNING.
	*/
	public static final Validator NONUNIQUE_LABEL = new Validator(
			"misc.nonuniquelabel", "Multiple labeled extracts with same label",
			"Checks if a hybridization has more than one labeled extract with the same label",
			"Multiple labeled extracts with same label", Severity.WARNING
		);

	/**
		Validator rule that checks if a parent to a pooled labeled extract has the
		same lebel. Default severity level is WARNING.
	*/
	public static final Validator DIFFERENT_LABEL = new Validator(
			"misc.differentlabel", "Labeled extract with different label",
			"Checks if the sources to a pooled labeled extract has the same " +
			"label as the product",
			"Labeled extract with different label", Severity.WARNING
		);
	
	/**
		Validator rule that checks if two items of the same type has the same
		name. Default severity level is ERROR.
	*/
	public static final Validator NONUNIQUE_NAME_FOR_TYPE = new Validator(
			"misc.nonuniquenamefortype", "Non-unique name",
			"Checks if more than one item of the same type has the same name.",
			"Non-unique name", Severity.ERROR
		);

	/**
		Validator rule that checks if two items of different type has the same
		name. Default severity level is IGNORE.
	*/
	public static final Validator NONUNIQUE_NAME_GLOBAL = new Validator(
			"misc.nonuniquenameglobal", "Global non-unique name",
			"Checks if more than one item (of any type) has the same name.",
			"Global non-unique name", Severity.IGNORE
		);

	/**
		Validator rule that checks for circular references among pooled items. 
		Default severity level is ERROR.
	*/
	public static final Validator CIRCULAR_REFERENCE = new Validator(
			"misc.circular", "Circular reference to pooled item",
			"Checks if there are circular references among pooled items.",
			"Circular reference to pooled item", Severity.ERROR
		);

	
	/**
		Validator rule that checks if an annotation is a protocol parameter.
		Default severity level is WARNING.
	*/
	public static final Validator ANNOTATION_IS_PARAMETER = new Validator(
			"annotation.isparameter", "Annotation is protocol parameter",
			"Checks if an item has annotations that are protocol parameters.",
			"Annotation is protocol parameter", Severity.WARNING
		);

	/**
		Validator rule that checks if an annotation is not valid for the
		current item type. Default severity level is WARNING.
	*/
	public static final Validator ANNOTATION_INVALID_ITEM = new Validator(
			"annotation.invaliditem", "Annotation invalid for item",
			"Checks if an annotation is invalid for the type of item.",
			"Annotation is invalid for item", Severity.WARNING
		);

	/**
		Validator rule that checks if an annotation value is invalid
		according to annotation type specification. Default severity level is ERROR.
	*/
	public static final Validator ANNOTATION_INVALID_VALUE = new Validator(
			"annotation.invalidvalue", "Annotation has invalid value",
			"Checks if an annotation has an invalid value.",
			"Annotation has invalid value", Severity.ERROR
		);
	
	/**
		Validator rule that checks if an inherited annotation value is inherited
		from a parent item. Default severity level is ERROR.
	*/
	public static final Validator ANNOTATION_INHERIT_FROM_NONPARENT = new Validator(
			"annotation.inhertitfromnonparent", "Inheriting annotation from non-parent",
			"Checks if an item inherits an annotation that is is not from a parent item.",
			"Inherting annotation from non-parent", Severity.ERROR
		);
	
	/**
		Validator rule that checks if the logged in user has access to
		the parent item of an inherited annotation value. Default severity 
		level is WARNING.
		@since 2.10
	*/
	public static final Validator ANNOTATION_INHERIT_FROM_DENIEDPARENT = new Validator(
			"annotation.inhertitfromdeniedparent", "Inheriting annotation from denied parent",
			"Checks if an item inherits an annotation from a parent item to which the " +
			"logged in user doesn't have access.",
			"Inherting annotation from denied parent", Severity.WARNING
		);

	
	/**
		Validator rule that checks if two or more inherited annotations have the
		same annotation type. Default severity level is ERROR.
	*/
	public static final Validator ANNOTATION_INHERIT_MULTIPLE = new Validator(
			"annotation.inhertitmultiple", "Inheriting multiple annotation of same type",
			"Checks if an item inherits more than one annotation of the same annotation type.",
			"Inheriting multiple annotation of same type", Severity.ERROR
		);
	
	/**
		Validator rule that checks if a data file can be used on the
		current item type. Default severity level is WARNING.
	*/
	public static final Validator DATAFILE_INVALID_ITEM = new Validator(
			"datafile.invaliditem", "Data file can't be used on item",
			"Checks if a data file can be used for a given item type.",
			"Data file can't be used on item", Severity.WARNING
		);

	/**
		Validator rule that checks if a data file has been validated
		or not. Default severity level is WARNING, but this rule is not 
		checked if the data file type doesn't have a validator.
	*/
	public static final Validator DATAFILE_NOTVALIDATED = new Validator(
			"datafile.notvalidated", "Data file has not been validated",
			"Checks if a data file has been validated or not.",
			"Data file has not been validated", Severity.WARNING
		);
	
	/**
		Validator rule that checks if a data file is valid.
		Default severity level is ERROR, but this rule is not checked if
		the data file hasn't been validated.
	*/
	public static final Validator DATAFILE_INVALID = new Validator(
			"datafile.invalid", "Data file is invalid",
			"Checks if a data file is valid or not.",
			"Data file is invalid", Severity.ERROR
		);

	/**
		Validator rule that checks if all files required by the platform
		are present or not. Default severity level is ERROR.
	*/
	public static final Validator MISSING_DATAFILE = new Validator(
			"datafile.missing", "Missing a required data file",
			"Checks if all files required by a platform are present.",
			"Missing a required data file", Severity.ERROR
		);

	/**
		Validator rule that checks if all files required by the platform
		are present or not. Default severity level is WARNING.
	*/
	public static final Validator DATAFILE_NOTPLATFORM = new Validator(
			"datafile.notplatform", "File is not used by platform",
			"Checks that all data files are used by the current platform.",
			"File is not used by platform", Severity.WARNING
		);
	
	private String id;
	private String title;
	private String description;
	private String failureSummary;
	private Severity defaultSeverity;
	
	/**
		Create a new validator.
		@param id A unique ID among all validators
		@param title The title for the validation rule 
		@param description A lengthier description of the validation that takes place
		@param failureSummary A short summary that is displayed when the validation
			fails
		@param defaultSeverity The default severity level of the validation
	*/
	public Validator(String id, String title, String description, String failureSummary, Severity defaultSeverity)
	{
		this.id = id;
		this.title = title;
		this.description = description;
		this.failureSummary = failureSummary;
		this.defaultSeverity = defaultSeverity;
	}
	
	/**
		Get the unique ID of this validator.
	*/
	public String getId()
	{
		return id;
	}
	
	/**
		Get the title of this validator.
	*/
	public String getTitle()
	{
		return title;
	}
	
	/**
		Get a lengthier description of the validation that takes place
	*/
	public String getDescription()
	{
		return description;
	}
	
	/**
		Get a short summary that is displayed when the validation fails
	*/
	public String getFailureSummary()
	{
		return failureSummary;
	}
	
	/**
		Get the default severity level.
	*/
	public Severity getDefaultSeverity()
	{
		return defaultSeverity;
	}
}
