/*
	$Id: CompoundMapper.java 4515 2008-09-11 20:08:14Z jari $
	
	Copyright (C) 2006 Nicklas Nordborg
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.
	
	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.parser;

import java.text.NumberFormat;
import java.util.List;

import net.sf.basedb.core.Type;
import net.sf.basedb.util.parser.FlatFileParser.Data;

/**
	A compound mapper that gets its value by invoking multiple other mappers
	which are set in the constructor. This type of mapper is used to combine
	constant strings with column values.

	@author nicklas
	@version 2.0
	@base.modified $Date: 2008-09-11 22:08:14 +0200 (Thu, 11 Sep 2008) $
*/
public class CompoundMapper
	implements Mapper
{
	private final List<Mapper> mappers;
	private final NumberFormat parser;
	private final boolean nullIfException;
	
	/**
		Create a new compound mapper.
		@param mappers A list of other mappers that will be invoked in 
			the order they appear in the list
	*/
	public CompoundMapper(List<Mapper> mappers)
	{
		this(mappers, null, false);
	}
	
	/**
		Create a new compound mapper, using a specific number formatter as it's parser.
		@param mappers A list of other mappers that will be invoked in 
			the order they appear in the list
		@param parser The number format to use or null to use Float.valueOf()
		@since 2.2.3
		@deprecated Use {@link CompoundMapper#CompoundMapper(List, NumberFormat, boolean)}
	*/
	public CompoundMapper(List<Mapper> mappers, NumberFormat parser)
	{
		this(mappers, parser, false);
	}
	
	/**
		Create a new compound mapper, using a specific number formatter as it's parser.
		@param mappers A list of other mappers that will be invoked in 
			the order they appear in the list
		@param parser The number format to use or null to use Float.valueOf()
		@param nullIfException If TRUE, the mapper returns null for unparsable numeric
			values, otherwise an excption is thrown
		@since 2.4
	*/
	public CompoundMapper(List<Mapper> mappers, NumberFormat parser, boolean nullIfException)
	{
		this.mappers = mappers;
		this.parser = parser;
		this.nullIfException = nullIfException;
	}
	
	/*
		From the Mapper interface
		-------------------------------------------
	*/
	/**
		Get a value by invoking each mapper in the order they appear in
		the list and concatenating the result.
		@return The concatenated value
	*/
	public String getValue(Data data)
	{
		StringBuilder sb = new StringBuilder();
		for (Mapper m : mappers)
		{
			sb.append(m.getValue(data));
		}
		return sb.toString();
	}
	public Integer getInt(Data data)
	{
		String stringValue = getValue(data);
		Integer intValue = null;
		if (stringValue != null)
		{
			if (parser != null)
			{
				intValue = (Integer)Type.INT.parseString(stringValue, parser, nullIfException);
			}
			else
			{
				try
				{
					intValue = Double.valueOf(stringValue).intValue();
				}
				catch (RuntimeException rex)
				{
					if (!nullIfException) throw rex;
				}
			}
		}
		return intValue;
	}
	public Float getFloat(Data data)
	{
		String stringValue = getValue(data);
		Float floatValue = null;
		if (stringValue != null)
		{
			if (parser != null)
			{
				floatValue = (Float)Type.FLOAT.parseString(stringValue, parser, nullIfException);
			}
			else
			{
				try
				{
					floatValue = Float.parseFloat(stringValue);
				}
				catch (RuntimeException rex)
				{
					if (!nullIfException) throw rex;
				}
			}
		}
		return floatValue;
	}
	// -------------------------------------------
	
	/*
		From the Object class
		-------------------------------------------
	*/
	public String toString()
	{
		StringBuilder sb = new StringBuilder();
		for (Mapper m : mappers)
		{
			sb.append(m.toString());
		}
		return sb.toString();
	}
	// -------------------------------------------
}

