/*
	$Id: ScatterPlot.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.plot;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.query.SqlResult;
import net.sf.basedb.core.query.SqlResultIterator;
import net.sf.basedb.core.signal.ThreadSignalHandler;

import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.plot.SeriesRenderingOrder;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.XYDotRenderer;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;


/**
	A simple plot utility for generating scatter plots without
	the need to delve deep into the JFreePlot package. Note! This class
	is experimental and it is likely that the implementation will change 
	in the future.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2009-04-06 14:52:39 +0200 (Mon, 06 Apr 2009) $
*/
public class ScatterPlot
{
	
	private XYPlot plot;
	private XYSeriesCollection allSeries;
	private JFreeChart chart;

	/**
		Create a new ScatterPlot.
		@param nameX The label on the X-axis
		@param nameY The label on the Y-axis
	*/
	public ScatterPlot(String nameX, String nameY)
	{
		allSeries = new XYSeriesCollection();
		NumberAxis domainAxis = new NumberAxis(nameX);
      domainAxis.setAutoRangeIncludesZero(false);

      NumberAxis rangeAxis = new NumberAxis(nameY);
		rangeAxis.setAutoRangeIncludesZero(true);

      plot = new XYPlot(allSeries, domainAxis, rangeAxis, new XYDotRenderer());
      plot.setSeriesRenderingOrder(SeriesRenderingOrder.FORWARD);
		chart = new JFreeChart(plot);
		chart.setAntiAlias(false);
	}
	
	/**
		Add data to the plot. The <code>SqlResultIterator</code> should return
		the x value as a float in the first position, ie. <code>data.getFloat(1)</code>
		and y values in the positions 2 and up.

		@param data The data to plot
		@param names An array of series names, x-value for series 'i' is found
			in <code>data.getFloat(2+i)</code>
		@throws SQLException
	*/
	public void addData(SqlResultIterator data, String... names)
		throws SQLException
	{
		XYSeries[] series = new XYSeries[names.length];
		for (int i = 0; i < names.length; ++i)
		{
			XYSeries s = new XYSeries(names[i]);
			s.setNotify(false);
			series[i] = s;
		}
		while (data.hasNext())
		{
			ThreadSignalHandler.checkInterrupted();
			SqlResult r = data.next();
			float x = r.getFloat(1);
			for (int i = 0; i < series.length; ++i)
			{
				series[i].add(x, r.getFloat(2+i));
			}
		}
		for (int i = 0; i < series.length; ++i)
		{
			allSeries.addSeries(series[i]);
		}
	}
	
	/**
		Add data to the plot. The SqlResultIterator must return at least 3 columns.
		The first column in the result is used to determine which annotation/series 
		the spot belongs to. The second column is the x value and the third column 
		is the y value. Additional columns are not used.
		<p>
		The annotations collection contains the mappings from the actual value to
		the annotation. Ie. if the result contains a 1 in the first column we
		check which annotation returns a 1 in it's {@link PlotAnnotation#getValues()}
		method. One value may not be used by more than one annotation.
		
		@param data The data to plot
		@param annotations Annotation information that maps each data spot to 
			an annotation
		@throws SQLException
	*/
	public void addData(SqlResultIterator data, Collection<PlotAnnotation> annotations)
		throws SQLException
	{
		Map<Integer, XYSeries> series = new HashMap<Integer, XYSeries>(annotations.size());
		List<XYSeries> newSeries = new ArrayList<XYSeries>(annotations.size());
		plot.getRenderer().setSeriesVisibleInLegend(null);
		for (PlotAnnotation pa : annotations)
		{
			XYSeries s = new XYSeries(pa.getName());
			s.setNotify(false);
			newSeries.add(s);
			for (Integer value : pa.getValues())
			{
				if (series.containsKey(value))
				{
					throw new BaseException("Value " + value + " has already been mapped to another series");
				}
				series.put(value, s);
			}
		}
		while (data.hasNext())
		{
			ThreadSignalHandler.checkInterrupted();
			SqlResult r = data.next();
			XYSeries s = series.get(r.getInt(1));
			s.add(r.getFloat(2), r.getFloat(3));
		}
		for (XYSeries s : newSeries)
		{
			allSeries.addSeries(s);
			plot.getRenderer().setSeriesVisibleInLegend(allSeries.getSeriesCount()-1, !s.getKey().equals(""));
		}
	}
	
	/**
	 	Get the underlying JFreeChar object. Use this method to polish up 
	 	everything and save the image.
	*/
	public JFreeChart getChart()
	{
		return chart;
	}
}
