/**
	$Id$

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.units;

import net.sf.basedb.core.Quantity;
import net.sf.basedb.core.Unit;

/**
	A linear unit converter is a converter implementation were all units
	can be converted to another unit using a multiplication factor and/or an
	offset. The formula used in this converter is:
	<p>
	<code>value(ref) = value(specific) * factor + offset</code>
	
	@author Nicklas
	@version 2.9
	@base.modified $Date$
*/
public class LinearUnitConverter
	implements UnitConverter
{

	private final double factor;
	private final double offset;
	
	/**
		Create a new unit converter.
		@param factor The factor to use when converting values
		@param offset The offset to use when converting values
	*/
	public LinearUnitConverter(double factor, double offset)
	{
		this.factor = factor;
		this.offset = offset;
	}
	
	/**
		Create a unit converter that converts between two units that 
		are both related to a third "common" unit by given factors and offsets.
		The formula used is:
		<code>value(ref) * factor(ref) + offset(ref) == 
		value(specific) * factor(specfic) + offset(specific)</code>
		which is equivalent to :
		<pre class="code">
factor = factor(specific) / factor(ref)
offset = (offset(specific) - offset(ref)) / factor(ref)
</pre>

		@param factor The factor of the specific unit
		@param offset The offset of the specific unit
		@param referenceFactor The factor of the reference unit
		@param referenceOffset The offset of the specific unit
	 */
	public LinearUnitConverter(double factor, double offset, 
		double referenceFactor, double referenceOffset)
	{
		this(factor / referenceFactor, (offset - referenceOffset) / referenceFactor);
	}
	
	/**
		Create a new unit converter that converts to/from a
		given unit. The reference unit is the unit given by 
		{@link Quantity#getReferenceUnit()} for the given specific 
		unit's quantity.
		
		@param unit The specific unit
	*/
	public LinearUnitConverter(Unit unit)
	{
		this(unit.getReferenceFactor(), unit.getReferenceOffset());
	}
	
	/**
		Create a new unit converter that converts to/from 
		given specific and reference unit.
		
		@param specific The specific unit
		@param reference The reference unit
	*/
	public LinearUnitConverter(Unit specific, Unit reference)
	{
		this(specific.getReferenceFactor(), specific.getReferenceOffset(),
			reference.getReferenceFactor(), reference.getReferenceOffset());
	}

	/**
		From the UnitConverter interface
		--------------------------------
	*/
	/**
		value(ref) = value(specific) * factor + offset
	*/
	public double convertToReferenceUnit(double specific)
	{
		return specific * factor + offset;
	}
	/**
		value(specific) = (value(ref) - offset) / factor
	*/
	public double convertToSpecificUnit(double ref)
	{
		return (ref - offset)/factor;
	}
	// -----------------------------------
	
}
