/**
	$Id: TarUtil.java 4515 2008-09-11 20:08:14Z jari $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.zip;

import java.io.ByteArrayOutputStream;
import java.io.IOException;

import net.sf.basedb.util.FileUtil;

import com.ice.tar.TarEntry;
import com.ice.tar.TarInputStream;
import com.ice.tar.TarOutputStream;

/**
	Contains utility functions for TAR archives. Most importantly functions
	for handling entries with too long file names to fit in the header section.

	@author nicklas
	@version 2.5.1
	@base.modified $Date: 2008-09-11 22:08:14 +0200 (Thu, 11 Sep 2008) $
*/
public class TarUtil
{
	/**
		TAR has a limit of 100 characters in file names. Names that are longer
		store the actual filename in an entry with this name just before
		the real file (which usually has a truncated name entry).
	*/
	public static final String LONG_LINK_NAME = "././@LongLink";
	
	/**
		Maximum file name length.
		@see #LONG_LINK_NAME
	*/
	public static final int MAX_NAME_LENGTH = 100;

	/**
		The TAR header flag that must be set for @LongLink entries.
	*/
	private static byte LONG_LINK_FLAG = (byte)'L';

	/**
		Get the next entry from a TAR input stream. This method has support for
		long filenames. After calling this method the actual file data can be
		read from the stream.
		@param tarStream The TAR stream to read from
		@return The next entry or null if the end of stream has been reached
		@throws IOException
	*/
	public static final TarEntry getNextEntry(TarInputStream tarStream)
		throws IOException
	{
		TarEntry entry = tarStream.getNextEntry();
		if (entry != null)
		{
			if (LONG_LINK_NAME.equals(entry.getName()))
			{
				ByteArrayOutputStream out = new ByteArrayOutputStream((int)entry.getSize());
				FileUtil.copy(tarStream, out);
				byte[] temp = out.toByteArray();
				int usedBytes = temp[temp.length-1] == 0x00 ? temp.length-1 : temp.length;
				String filename = new String(temp, 0, usedBytes);
				entry = tarStream.getNextEntry();
				entry.setName(filename);
			}
		}
		return entry;
	}
	
	/**
		Put the next entry to a TAR stream. This method has support for 
		long filenames and will automatically create all neccesary extra information
		inside the TAR stream. After calling this method the actual file data
		should be written to the stream.
		
		@param entry The next entry
		@param tarStream The tar stream to write to
		@throws IOException
	*/
	public static final void putNextEntry(TarEntry entry, TarOutputStream tarStream)
		throws IOException
	{
		String name = entry.getName();
		if (name.length() > MAX_NAME_LENGTH)
		{
			TarEntry longLink = new TarEntry(LONG_LINK_NAME);
			if (entry.getModTime() != null) longLink.setModTime(entry.getModTime());
			longLink.setSize(name.length()+1);
			longLink.getHeader().linkFlag = LONG_LINK_FLAG;
			tarStream.putNextEntry(longLink);
			tarStream.write(name.getBytes());
			tarStream.write(0);
			tarStream.closeEntry();
			entry.setName(name.substring(0, MAX_NAME_LENGTH));
		}
		tarStream.putNextEntry(entry);
	}
	
}
