/**
	$Id: BioAssaySetExporter.java 4928 2009-05-13 11:01:46Z nicklas $

	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Johan Enell, Jari Häkkinen, Peter Johansson, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2008 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.plugins;

import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.BooleanParameterType;
import net.sf.basedb.core.Config;
import net.sf.basedb.core.DataFileType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DynamicPositionQuery;
import net.sf.basedb.core.DynamicResultIterator;
import net.sf.basedb.core.DynamicSpotQuery;
import net.sf.basedb.core.ExtendedProperties;
import net.sf.basedb.core.ExtendedProperty;
import net.sf.basedb.core.ExtraValue;
import net.sf.basedb.core.ExtraValueType;
import net.sf.basedb.core.File;
import net.sf.basedb.core.FileSetMember;
import net.sf.basedb.core.FileStoreUtil;
import net.sf.basedb.core.Formula;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.IntensityTransform;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.Path;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PluginDefinition;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.RawDataProperty;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.Version;
import net.sf.basedb.core.VirtualColumn;
import net.sf.basedb.core.VirtualTable;
import net.sf.basedb.core.Formula.AverageMethod;
import net.sf.basedb.core.data.ReporterData;
import net.sf.basedb.core.plugin.About;
import net.sf.basedb.core.plugin.AboutImpl;
import net.sf.basedb.core.plugin.AbstractExporterPlugin;
import net.sf.basedb.core.plugin.ExportOutputStream;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.Permissions;
import net.sf.basedb.core.plugin.Plugin;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.JoinType;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Selects;
import net.sf.basedb.core.query.SqlResult;
import net.sf.basedb.core.signal.SignalHandler;
import net.sf.basedb.core.signal.SignalTarget;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.BioAssaySetUtil;
import net.sf.basedb.util.Enumeration;
import net.sf.basedb.util.export.spotdata.AbstractBioAssaySetExporter;
import net.sf.basedb.util.export.spotdata.MatrixBaseFileExporter;
import net.sf.basedb.util.export.spotdata.SerialBaseFileExporter;

import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;


/**
	 Exporter for bioassay sets.

	 Supported formats:

	 - Tab Delimited, Multiple Sample Files (TDMS Format) for
	   use in MeV (available from http://www.tm4.org).
	 - BASEfile.
	 - Plain Matrix, see documentation for exportPlainMatrix method for
     a description of the format.

	 @author enell, jari, peter
	 @version 2.1
	 @base.modified $Date: 2009-05-13 13:01:46 +0200 (Wed, 13 May 2009) $
	 @deprecated In 2.12. Use: {@link SerialBaseFileExporter} or {@link MatrixBaseFileExporter}
	 	for BASEfile export and the extension package at http://baseplugins.thep.lu.se/wiki/net.sf.basedb.mev
	 	for MeV export
*/
public class BioAssaySetExporter
	extends AbstractExporterPlugin
	implements InteractivePlugin, SignalTarget
{

	private ThreadSignalHandler signalHandler;
	
	public void configure(GuiContext context, Request request, Response response)
	{
		try
		{
			String command = request.getCommand();
			if (command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				RequestInformation ri = getConfigureJobParameters (context, !request.isAllowedImmediateExecution());
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found " +
														"in the request", errors);
					return;
				}
				if (request.getParameterValue(SAVE_AS) == null)
				{
					if (!request.isAllowedImmediateExecution())
					{
						response.setError("Immediate download is not allowed. Please specify a filename", null);
						return;
					}
				}
				else
				{
					Object parameterValue = request.getParameterValue(OVERWRITE); 
					boolean overwrite = parameterValue != null ? (Boolean)parameterValue : false;
					if (!pathCanBeUsed((String)request.getParameterValue(ri.getParameter(SAVE_AS).getName()), overwrite))
					{
						response.setError("File exists: " + (String)request.getParameterValue(ri.getParameter(SAVE_AS).getName()), null);
						return;
					}
				}

				storeValue(job, request, bioAssaySetParameter);
				storeValue(job, request, formatParameter);
				storeValue(job, request, ri.getParameter(SAVE_AS));
				storeValue(job, request, ri.getParameter(OVERWRITE));
				storeValue(job, request, ri.getParameter("attachToBioAssaySet"));
				
				String fileFormat = (String)request.getParameterValue(formatParameter.getName());
				if (FORMAT_BASEFILES.contains(fileFormat))
				{
					response.setContinue(COMMAND_CONFIGURE_BASEFILE);
				}
				else
				{
					if (job.getValue(SAVE_AS) == null)
					{
						BioAssaySet bas = (BioAssaySet)request.getParameterValue("bioAssaySet");
						response.setDownloadImmediately("Export bioassay set " + 
								bas.getName(), Job.ExecutionTime.SHORT, true);
					}
					else
					{
						response.setDone("Job configuration complete", Job.ExecutionTime.SHORT);
					}
				}
			}
			else if (command.equals(COMMAND_CONFIGURE_BASEFILE))
			{
				List<Throwable> errors = validateRequestParameters
					(getConfigureBasefileParameters(context).getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found " +
														"in the request", errors);
					return;
				}
				
				storeValue(job, request, geneAveragesParameter);
				storeValues(job, request, usedFieldsParameter);
				storeValues(job, request, usedColumnsParameter);

				if (job.getValue(SAVE_AS) == null)
				{
					BioAssaySet bas = (BioAssaySet)job.getValue("bioAssaySet");
					response.setDownloadImmediately("Export bioassay set " + bas.getName() , Job.ExecutionTime.SHORT, true);
				}
				else
				{
					response.setDone("Job configuration complete", Job.ExecutionTime.SHORT);
				}
			}
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
	}
	
	/*
		From the SignalTarget interface
		-------------------------------------------
	*/
	public SignalHandler getSignalHandler()
	{
		signalHandler = new ThreadSignalHandler();
		return signalHandler;
	}
	// -------------------------------------------

	/**
		@deprecated In 2.11 and 2.12, use {@link MatrixBaseFileExporter} instead
	*/
	public void exportBaseFileMatrix(BioAssaySet bas, File file, Map<String, String> parameters, 
			List<String> reporterFields, List<String> spotFields, Boolean mergeReporters)
			throws IOException, SQLException
	{
		exportBaseFileMatrix(bas, file, parameters, reporterFields, spotFields, mergeReporters, null);
	}
	
	/**
		Export the data from a bioassayset as a serial BASEfile. 
		The <code>reporterFields</code> and <code>spotFields</code> parameter
		must identify the data to export. 
		<p>
		The <code>reporterFields</code> parameter supports the following:

		<table>
		<tr>
			<td>rep.<i>&lt;property&gt;</i></td>
			<td>
				The reporter property with the specified name, for example:
				rep.externalId, rep.symbol or a property defined in the extended properties 
				file.
			</td>
		</tr>
		</table>

		The <code>spotFields</code> parameter supports the following:
		<table>
		<tr>
			<td>intensityN</td>
			<td>The intensity in channel N</td>
		</tr>
		<tr>
			<td>position</td>
			<td>The position number for this spot in the data matrix</td>
		</tr>
		<tr>
			<td>raw.<i>&lt;property&gt;</td>
			<td>
				The raw data property with the specified name, for example:
				raw.x, raw.row or a property defined by the raw data type.
			</td>
		</tr>
		<tr>
			<td>frm.<i>&lt;id&gt;</td>
			<td>The value of the formula with the specified ID.</td>
		</tr>
		<tr>
			<td>xtra.<i>&lt;externalId&gt;</td>
			<td>The value of the extra value with the specified external ID.</td>
		</tr>
		</table>
		
		@param bas The bioassayset to export
		@param file The file to export to
		@param parameters Header information that should go at the top of
			the BASEfile
		@param reporterFields The reporter fields that should be exportered
		@param spotFields The spot fields that should be exported
		@param mergeReporters TRUE to merge spot data having the same reporter
			(currently not supported)
		@param progress An optional progress reporter
		@throws IOException If there is an IO error
		@throws SQLException If there is an error reading from the database
		@since 2.11
		@deprecated In 2.12, use {@link MatrixBaseFileExporter} instead
	*/
	public void exportBaseFileMatrix(BioAssaySet bas, File file, Map<String, String> parameters, 
		List<String> reporterFields, List<String> spotFields, Boolean mergeReporters, 
		ProgressReporter progress)
		throws IOException, SQLException
	{
		OutputStream os = file.getUploadStream(false);
		exportBaseFileMatrix(bas, os, parameters, reporterFields, spotFields, mergeReporters);
		os.close();
	}

	/**
		@deprecated In 2.11 and 2.12, use {@link MatrixBaseFileExporter} instead.
	*/
	public void exportBaseFileMatrix(BioAssaySet bas, OutputStream os, Map<String, String> parameters, 
			List<String> reporterFields, List<String> spotFields, boolean mergeReporters)
			throws IOException, SQLException
	{
		exportBaseFileMatrix(bas, os, parameters, reporterFields, spotFields, mergeReporters, null);
	}
	
	/**
		@since 2.11
		@deprecated In 2.11 and 2.12, use {@link MatrixBaseFileExporter} instead.
	*/
	public void exportBaseFileMatrix(BioAssaySet bas, OutputStream os, Map<String, String> parameters, 
		List<String> reporterFields, List<String> spotFields, boolean mergeReporters, ProgressReporter progress)
		throws IOException, SQLException
	{
		DbControl dc = bas.getDbControl();
		PrintWriter out = new PrintWriter(new OutputStreamWriter(os, Config.getCharset()));

		Map<String, QueryItem> qis = getBaseFileQIMap(dc, bas);

		checkInterrupted();
		if (progress != null) progress.display(0, "Exporting headers...");
		exportBaseFileSectionHead(out, parameters);
		exportBaseFileSectionAssays(out, dc, bas);

		DynamicSpotQuery query = bas.getSpotData();
		if (mergeReporters) query.joinReporters(JoinType.INNER);
		for (String field : spotFields)
		{
			QueryItem qi = qis.get(field);
			if (qi == null)
			{
				throw new BaseException("Unknown field in spot fields: "+field+" (you can add this field using a formula)");
			}
			query.select(Selects.expression(qi.getExpression(dc, bas, mergeReporters), qi.getQueryAlias()));
		}
	
		checkInterrupted();
		if (progress != null) progress.display(5, "Counting spots...");
		long count = getBaseFileSpotSectionCount(dc, bas.getSpotData(), mergeReporters);
		checkInterrupted();
		if (progress != null) progress.display(10, "Loading reporter information...");
		Map<Integer, String> reporterLine = getBaseFileReporterFields(dc, bas, bas.getPositionData(), 
				qis, reporterFields, mergeReporters);
		checkInterrupted();
		exportBaseFileSectionSpots(out, dc, bas, query, reporterFields, spotFields, bas.getBioAssays().list(dc), 
				qis, count, mergeReporters, reporterLine, progress);
		out.flush();
		if (progress != null) progress.display(100, "Export complete. " + count + " spots done");
	}
	
	private void exportBaseFileSectionAssays(PrintWriter out, DbControl dc, BioAssaySet bas)
	{
		checkInterrupted();
		out.println("section\tassays");
		out.println("count\t" + bas.getBioAssays().count(dc));
		String annotationString = "";
		for (AnnotationType at : getExperimentalFactors(dc, bas))
		{
			annotationString += "\t" + at.getName();
		}
		out.println("columns\tid\tname" + annotationString);
		out.println("annotationColumns" + (annotationString.length() == 0 ? "\t" : annotationString));
		out.println("%");
		for (BioAssay ba : bas.getBioAssays().list(dc))
		{
			checkInterrupted();
			out.print(ba.getId() + "\t" + ba.getName());
			for (AnnotationType at : getExperimentalFactors(dc, bas))
			{
				out.print("\t");
				Set<?> values = BioAssaySetUtil.getAnnotationValues(dc, ba, at);
				if (!values.isEmpty()) out.print(values.iterator().next());
			}
			out.println();
		}
		out.println();
	}
	
	private void exportBaseFileSectionHead(PrintWriter out, Map<String, String> parameters)
	{
		checkInterrupted();
		out.println("BASEfile");
		
		if (!parameters.isEmpty())
		{
			List<String> keys = new ArrayList<String>(parameters.keySet());
			if ("section".equals(keys.get(0)))
			{
				out.println("section\t"+parameters.get("section"));
				keys = keys.subList(1, keys.size());
			}
			else
			{
				out.println("section\tsettings");
			}
			if (keys.contains("section"))
			{
				throw new BaseException("The parameters can't have a parameter 'section' in the middle of the list.");
			}
			for (String key : keys)
			{
				checkInterrupted();
				out.println(key+"\t"+parameters.get(key));
			}
			out.println("%");
			out.println();
		}
	}
	
	/**
		@deprecated In 2.11 and 2.12, use {@link SerialBaseFileExporter} instead.
	*/
	public void exportBaseFileSerial(BioAssaySet bas, File file, Map<String, String> parameters, 
			List<String> reporterFields, List<String> spotFields, Boolean mergeReporters)
			throws IOException, SQLException
	{
		exportBaseFileSerial(bas, file, parameters, reporterFields, spotFields, mergeReporters, null);
	}

	/**
		Export the data from a bioassayset as a serial BASEfile.
		See {@link #exportBaseFileMatrix(BioAssaySet, File, Map, List, List, Boolean)}
		for information about the naming of parameters.
		
		@param bas The bioassayset to export
		@param file The file to export to
		@param parameters Header information that should go at the top of
			the BASEfile
		@param reporterFields The reporter fields that should be exportered
		@param spotFields The spot fields that should be exported
		@param mergeReporters TRUE to merge spot data having the same reporter
			(currently not supported)
		@param progress An optional progress reporter
		@throws IOException If there is an IO error
		@throws SQLException If there is an error reading from the database
		@since 2.11
		@deprecated In 2.12, use {@link SerialBaseFileExporter} instead.
	*/
	public void exportBaseFileSerial(BioAssaySet bas, File file, Map<String, String> parameters, 
		List<String> reporterFields, List<String> spotFields, Boolean mergeReporters, ProgressReporter progress)
		throws IOException, SQLException
	{
		OutputStream os = file.getUploadStream(false);
		exportBaseFileSerial(bas, os, parameters, reporterFields, spotFields, mergeReporters, progress);
		os.close();
	}
	
	/**
		@deprecated In 2.11 and 2.12, use {@link SerialBaseFileExporter} instead.
	*/
	public void exportBaseFileSerial(BioAssaySet bas, OutputStream os, Map<String, String> parameters, 
			List<String> reporterFields, List<String> spotFields, boolean mergeReporters)
			throws IOException, SQLException
	{
		exportBaseFileSerial(bas, os, parameters, reporterFields, spotFields, mergeReporters, null);
	}

	/**
		@since 2.11
		@deprecated In 2.12, use {@link SerialBaseFileExporter} instead.
	*/
	public void exportBaseFileSerial(BioAssaySet bas, OutputStream os, Map<String, String> parameters, 
		List<String> reporterFields, List<String> spotFields, boolean mergeReporters, ProgressReporter progress)
		throws IOException, SQLException
	{
		checkInterrupted();
		DbControl dc = bas.getDbControl();
		PrintWriter out = new PrintWriter(new OutputStreamWriter(os, Config.getCharset()));
		
		Map<String, QueryItem> qis = getBaseFileQIMap(dc, bas);

		if (progress != null) progress.display(0, "Exporting headers...");
		exportBaseFileSectionHead(out, parameters);
		exportBaseFileSectionAssays(out, dc, bas);
		
		List<BioAssay> assays = bas.getBioAssays().list(dc);
		int numAssays = assays.size();
		int assaysDone = 0;
		checkInterrupted();
		if (progress != null) progress.display(10, "Loading reporter information...");
		Map<Integer, String> reporterLine = getBaseFileReporterFields(dc, bas, bas.getPositionData(), 
				qis, reporterFields, mergeReporters);
		for (BioAssay ba : assays)
		{
			checkInterrupted();
			if (progress != null)
			{
				int percent = 10+(int)((90L * assaysDone) / numAssays);
				progress.display(percent, "Exporting bioassay '" + ba.getName() + 
					"' (" + (assaysDone+1)+" of " + numAssays + ")");
			}
			DynamicSpotQuery query = ba.getSpotData();
			if (mergeReporters) query.joinReporters(JoinType.INNER);
			for (String field : spotFields)
			{
				QueryItem qi = qis.get(field);
				if (qi == null)
				{
					throw new BaseException("Unknown field: "+field+" (you can add this field using a formula)");
				}
				query.select(Selects.expression(qi.getExpression(dc, bas, mergeReporters), qi.getQueryAlias()));
			}

			checkInterrupted();
			long count = getBaseFileSpotSectionCount(dc, ba.getSpotData(), mergeReporters);
			exportBaseFileSectionSpots(out, dc, bas, query, reporterFields, spotFields, 
					Collections.singletonList(ba), qis, count, mergeReporters, reporterLine, null);
			assaysDone++;
			out.println();
		}
		out.flush();
	}


	private void exportBaseFileSectionSpots(PrintWriter out, DbControl dc, BioAssaySet bas, DynamicSpotQuery query,
		List<String> reporterFields, List<String> spotFields, List<BioAssay> assays, Map<String, QueryItem> qi, 
		long count, boolean mergeOnReporters, Map<Integer, String> reporterLine, ProgressReporter progress)
		throws SQLException
	{
		checkInterrupted();
		char[] tabs = new char[spotFields.size()-1];
		Arrays.fill(tabs, '\t');
		
		out.println("section\tspots");
		out.println("channels\t" + bas.getRawDataType().getChannels());
		out.print("assayFields");
		for (String field : spotFields)
		{
			out.print("\t" + qi.get(field).getExportName());
		}
		out.println();
		out.print("columns");
		StringBuilder noReporter = new StringBuilder();
		for (String field : reporterFields)
		{
			QueryItem item = qi.get(field);
			out.print("\t" + item.getExportName());
			if ("position".equals(item.getExportName())) noReporter.append("{pos}");
			noReporter.append("\t");
		}
		out.println("\tassayData");
		out.print("assays");
		for (BioAssay ba : assays)
		{
			out.print("\t" + ba.getId());
		}
		out.println();
		out.println("count\t" + count);
		out.println("%");
		
		ArrayList<Short> columnNo = new ArrayList<Short>(assays.size());
		for (BioAssay ba : assays)
		{
			columnNo.add(ba.getDataCubeColumnNo());
		}
		
		Expression posExpression = null;
		Expression colExpression = null;
		if (mergeOnReporters)
		{
			posExpression = Dynamic.column(VirtualTable.POSITION, VirtualColumn.REPORTER_ID);
			colExpression = Dynamic.column(VirtualColumn.COLUMN);
			query.group(colExpression);
			query.group(posExpression);
		}
		else
		{
			posExpression = Dynamic.column(VirtualColumn.POSITION);
			colExpression = Dynamic.column(VirtualColumn.COLUMN);
		}
		query.select(Selects.expression(posExpression, "_position"));
		query.select(Selects.expression(colExpression, "_column"));
		query.order(Orders.asc(posExpression));
		query.order(Orders.asc(colExpression));
		
		if (progress != null) progress.display(10, "Loading spot data...");
		checkInterrupted();
		DynamicResultIterator iterator = query.iterate(dc);
		List<Integer> spotIndex = new ArrayList<Integer>(spotFields.size());
		for (String field : spotFields)
		{
			spotIndex.add(iterator.getIndex(qi.get(field).getQueryAlias()));
		}
		int posIndex = iterator.getIndex("_position");
		int colIndex = iterator.getIndex("_column");
		
		Integer prevPos = null;
		HashMap<Short, String> spotValues = new HashMap<Short, String>();
		long spotsDone = 0;
		long progressInterval = count / 100;
		while (iterator.hasNext())
		{
			checkInterrupted();
			SqlResult r = iterator.next();
			Integer currentPos = r.getInt(posIndex);
			
			if (!currentPos.equals(prevPos))
			{
				// Print spot fields
				if (!spotValues.isEmpty())
				{
					spotsDone++;
					if (progress != null && spotsDone % progressInterval == 0)
					{
						int percent = 10+(int)((90L * spotsDone) / count);
						progress.display(percent, "Exporting spot data: " + spotsDone + " of " + count + " done");
					}
					String separator = "";
					for (Short col : columnNo)
					{
						String spot = spotValues.get(col);
						out.print(separator);
						if (spot != null)
						{
							out.print(spot);
						}
						else
						{
							out.print(tabs);
						}
						separator = "\t";
					}
					out.println();
					spotValues.clear();
				}
				String reporter = reporterLine.get(currentPos);
				if (reporter == null) 
				{
					reporter = noReporter.toString().replace("{pos}", Integer.toString(currentPos));
				}
				out.print(reporter);
			}
			
			String spot = "";
			for (int i = 0; i < spotIndex.size(); ++i)
			{
				String field = r.getString(spotIndex.get(i));
				if (i > 0) spot += "\t";
				spot += field == null ? "" : field;
			}
			spotValues.put(r.getShort(colIndex), spot);
			prevPos = currentPos;
		}
		
		// Print the last spot fields
		if (!spotValues.isEmpty())
		{
			String separator = "";
			for (Short col : columnNo)
			{
				String spot = spotValues.get(col);
				out.print(separator);
				if (spot != null)
				{
					out.print(spot);
				}
				else
				{
					out.print(tabs);
				}
				separator = "\t";
			}
			out.println();
			spotValues.clear();
		}
	}
	
	/**
		Export bioassayset data in MeV format. The export includes all
		reporter information and all experimental factor information. If
		the expriment is a two-channel experiment the ch1/ch2 ratio
		(non-log value!) is exported as the expression value otherwise the
		raw intensity of channel 1 is exported.
		
		@param bioassayset The bioassayset to export
		@param to The stream to write the exported data to
		@param progress An optional progress reporter
		@throws IOException If there is an IO error
		@throws SQLException If there is an error reading from the database
		@deprecated In 2.12, use extension package found at 
			http://baseplugins.thep.lu.se/wiki/net.sf.basedb.mev instead
	*/
	public void exportMeV(BioAssaySet bioassayset, OutputStream to,
												 ProgressReporter progress)
		throws IOException, SQLException
	{
		checkInterrupted();
		DbControl dc=bioassayset.getDbControl();
		PrintWriter out = new PrintWriter(new OutputStreamWriter(to, Config.getCharset()));

		// comment header
		/* Not supported by MeV
		out.write("# Created with " +  Application.getVersionString() + "\n");
		out.write
			( "# Date created: " +
				new SimpleDateFormat("d MMM yyyy HH:mm:ss 'UTC'Z").format(new Date()) +
				"\n");
		*/

		List<ExtendedProperty> reporterProperties =
			ExtendedProperties.getProperties("ReporterData");

		// Data header
		// Reporter headers
		out.write("id\texternalId\tsymbol");
		for (ExtendedProperty ep : reporterProperties)
		{
			out.write("\t"+ep.getName());
		}
		// Number 2 in below statement comes from using two reporter
		// entries from ReporterData. Below there is dependencies on this
		// '2' such as the selectReporter("") in below query.select()
		// statements (and more).
		int nof_reporter_columns=2+reporterProperties.size();
		// Bioassay headers
		out.write("\texperimentalFactor");
		short nof_bioassays=0;
		for (BioAssay bioassay : bioassayset.getBioAssays().list(dc))
		{
			out.write("\t" + bioassay.getName());
			++nof_bioassays;
		}

		// Experimental factors
		for (AnnotationType at : getExperimentalFactors(dc, bioassayset))
		{
			checkInterrupted();
			out.write("\n");
			for (int i=0; i < nof_reporter_columns; ++i)
			{
				out.write("\t");
			}
			out.write("\t"+at.getName());
			for (BioAssay bioassay : bioassayset.getBioAssays().list(dc))
			{
				Set<?> values = BioAssaySetUtil.getAnnotationValues(dc,bioassay,at);
				out.print("\t");
				out.print(!values.isEmpty() ? values.iterator().next() : "N/A" );
			}
		}

		int nofchannels=bioassayset.getRawDataType().getChannels();
		
 		// Reporter data query
		if (progress != null) progress.display(0, "Loading reporter information...");
		checkInterrupted();
		DynamicPositionQuery posQuery = bioassayset.getPositionData();
		posQuery.joinReporters(JoinType.INNER);
		posQuery.select(Dynamic.select(VirtualColumn.POSITION));
		posQuery.select(Dynamic.selectReporter("externalId"));
		posQuery.select(Dynamic.selectReporter("symbol"));
		for (ExtendedProperty ep : reporterProperties)
		{
			posQuery.select(Dynamic.selectReporter(ep.getName()));
		}
		Map<Integer, String> reporterLine = new HashMap<Integer, String>(bioassayset.getNumReporters());
		DynamicResultIterator posData = posQuery.iterate(dc);
		int column_position = posData.getIndex(VirtualColumn.POSITION.getName());
		int[] column_reporter = new int[nof_reporter_columns];
		column_reporter[0] = posData.getIndex("externalId");
		column_reporter[1] = posData.getIndex("symbol");
		int idx=2;
		StringBuilder noReporter = new StringBuilder("\t\t\t");
		for (ExtendedProperty ep : reporterProperties)
		{
			column_reporter[idx++] = posData.getIndex(ep.getName());
			noReporter.append("\t");
		}
		while (posData.hasNext())
		{
			checkInterrupted();
			SqlResult row = posData.next();
			int position = row.getInt(column_position);
			StringBuilder reporter = new StringBuilder();
			for (int i=0; i<nof_reporter_columns; ++i)
			{
				String str = row.getString(column_reporter[i]);
				reporter.append("\t");
				if (str != null) reporter.append(str);
			}
			reporter.append("\t"); // experimentalFactor is always empty for reporters
			reporterLine.put(position, reporter.toString());
		}
		
		// Spot data query
		if (progress != null) progress.display(5, "Loading spot information...");
		checkInterrupted();
		DynamicSpotQuery query=bioassayset.getSpotData();
		query.select(Dynamic.select(VirtualColumn.POSITION));
		query.select(Dynamic.select(VirtualColumn.COLUMN));
		for (int i=1; i<=nofchannels; ++i)
		{
			query.select(Dynamic.select(VirtualColumn.channelRaw(i)));
		}
 		query.order(Orders.asc(Dynamic.column(VirtualColumn.POSITION)));
 		query.order(Orders.asc(Dynamic.column(VirtualColumn.COLUMN)));

 		DynamicResultIterator spotData = query.iterate(dc);
		column_position = spotData.getIndex(VirtualColumn.POSITION.getName());
		int column_column = spotData.getIndex(VirtualColumn.COLUMN.getName());
		int[] column_channel = new int[nofchannels];
		for (int i=1; i<=nofchannels; ++i)
		{
			column_channel[i-1] = spotData.getIndex(VirtualColumn.channelRaw(i).getName());
		}
		int position = -1;
		short column = nof_bioassays;
		int nof_reporters = bioassayset.getNumReporters();
		int progress_report_interval = nof_reporters/10;
		long nof_processed_reporters = 0;
		IntensityTransform transform = bioassayset.getIntensityTransform();
		boolean isLogged = transform == IntensityTransform.LOG10 || 
			transform == IntensityTransform.LOG2;
		while (spotData.hasNext())
		{
			checkInterrupted();
			if (progress != null && (nof_processed_reporters%progress_report_interval == 0))
			{
				progress.display(10 + (int)(90L * nof_processed_reporters / nof_reporters),
						"Exporting ... " + nof_processed_reporters + " of " +
						nof_reporters + " done.");
			}

			SqlResult item = spotData.next();
			int thisposition=item.getInt(column_position);
			short thiscolumn=item.getShort(column_column);

			if (position != thisposition)	// new reporter
			{
				++nof_processed_reporters;
				// fill missing values before next spot
				while (nof_bioassays>column++) 
				{
					out.write("\t");
				}
				column = 0;
				position = thisposition;
				out.write("\n" + position);
				String reporter = reporterLine.get(position);
				if (reporter == null) reporter = noReporter.toString();
				out.write(reporter);
			}
			// fill missing values up to next available column (bioassay)
			while (++column < thiscolumn) 
			{
				out.write("\t");
			}

			// Calculate expression value
			double ch1 = item.getFloat(column_channel[0]);
			boolean goodvalue = true;
			if (nofchannels == 2)
			{
				double ch2 = item.getFloat(column_channel[1]);
				if (isLogged)
				{
					ch1 -= ch2;
				}
				else
				{
					if (ch2 != 0)
					{
						ch1 /= ch2;
						if (ch1 <= 0) goodvalue = false;
					}
					else
					{
						goodvalue = false;
					}
				}
			}

			out.write("\t" + (goodvalue ? ch1 : "NaN"));
		}
		out.write("\n");
		out.flush();
	}


	/**
		Export bioassayset data in plain matrix format.

		Plain matrix format is simply a matrix with no header nor reporter
		information. Each column in the matrix corresponds to an assay,
		and each row corresponds to a reporter.

		Missing values are reported as entries with no number and 'NaN' is
		reported for undefined values (such as division with zero). Note,
		reporters with no measurement within a bioassay set are not
		reported at all, i.e., a row is not exported with misssing values
		only. To be clear, we state this once more: Rows (reporters) are
		only exported if one or more columns (assays) have a value. The
		value can be 'NaN'.

		The data exported depends on the underlying microarray platform;
		 - 1-channel: raw ch1 intensity
		 - 2-channel: the fraction ch1/ch2, note! non-log value
		The justification for exporting non-log values for 2-channel data
		is to follow the MeV export format where non-log data is exported
		for 2-channel data. 'NaN' is returned for zero ch2 values.

		The plug-in or user that uses this method must keep track of the
		row and column information themselves. The row is sorted in
		ascending position order, the columns are sorted in ascending
		column order.

		@param bioassayset The bioassayset to export
		@param os The stream to write the exported data to
		@param progress An optional progress reporter
		@throws IOException If there is an IO error
		@throws SQLException If there is an error reading from the database
		@deprecated In 2.12, no direct replacement, write a subclass of 
			{@link AbstractBioAssaySetExporter} instead
	*/
	public void exportPlainMatrix(BioAssaySet bioassayset, OutputStream os,
																ProgressReporter progress)
		throws IOException, SQLException
	{
		checkInterrupted();
		DbControl dc=bioassayset.getDbControl();
		PrintWriter out = new PrintWriter(new OutputStreamWriter(os, Config.getCharset()));

		int nof_bioassays=0;
		nof_bioassays=bioassayset.getBioAssays().list(dc).size();
		int nof_channels=bioassayset.getRawDataType().getChannels();

		if (progress != null)
		{
			progress.display(0, "Exporting ... performing database query");
		}

		DynamicSpotQuery query=bioassayset.getSpotData();
		query.select(Dynamic.select(VirtualColumn.POSITION));
		query.select(Dynamic.select(VirtualColumn.COLUMN));
		for (int i=1; i<=nof_channels; ++i)
		{
			query.select(Dynamic.select(VirtualColumn.channelRaw(i)));
		}

 		query.order(Orders.asc(Dynamic.column(VirtualColumn.POSITION)));
 		query.order(Orders.asc(Dynamic.column(VirtualColumn.COLUMN)));

		checkInterrupted();
		DynamicResultIterator spotData = query.iterate(dc);
		int column_position=spotData.getIndex(VirtualColumn.POSITION.getName());
		int column_column=spotData.getIndex(VirtualColumn.COLUMN.getName());
		int[] column_channel=new int[nof_channels];
		for (int i=0; i<nof_channels; ++i)
		{
			column_channel[i]=spotData.getIndex(VirtualColumn.channelRaw(i+1).getName());
		}

		int position=-1;
		int column=nof_bioassays;
		int nof_spots=bioassayset.getNumSpots();
		int progress_spot_interval=nof_spots/10;
		long nof_processed_spots=0;
		IntensityTransform transform = bioassayset.getIntensityTransform();
		boolean isLogged = transform == IntensityTransform.LOG10 || 
			transform == IntensityTransform.LOG2;
		while (spotData.hasNext())
		{
			checkInterrupted();
			SqlResult item = spotData.next();
			int thisposition=item.getInt(column_position);
			short thiscolumn=item.getShort(column_column);

			if (progress != null && 
					(nof_processed_spots%progress_spot_interval == 0))
				{
					progress.display((int)(100L * nof_processed_spots / nof_spots),
													 "Exporting ..." + nof_processed_spots + " of " +
													 nof_spots + " done.");
				}
				++nof_processed_spots;
	
			if (position!=thisposition)	// new position
			{
				// fill missing values before next spot
				while (column < nof_bioassays) {
					out.write("\t");
					++column;
				}
				column=1;
				if (position != -1) // no leading newline in output 
				{
					out.write("\n");
				}
				position=thisposition;
			}

			// fill missing values up to next available column (bioassay)
			while (column < thiscolumn) 
			{
				++column;
				out.write("\t");
			}

			// Calculate expression value
			double exValue=item.getFloat(column_channel[0]);
			boolean goodvalue=true;
			if (nof_channels==2)
			{
				double ch2=item.getFloat(column_channel[1]);
				if (isLogged)
				{
					exValue -= ch2;
				}
				else
				{
					if (ch2!=0)
					{
						exValue /= ch2;
					}
					else
					{
						goodvalue=false;
					}
				}
			}

			out.write( "" + (goodvalue ? exValue : "NaN"));
		}
		out.close();
	}


	public About getAbout()
	{
		return about;
	}
	
	
	private long getBaseFileSpotSectionCount(DbControl dc, DynamicSpotQuery count, boolean mergeReporters)
	{
		count.setDistinct(true);
		if (mergeReporters)
		{
			count.joinReporters(JoinType.INNER);
			count.select(Selects.expression(Dynamic.column(VirtualTable.POSITION, VirtualColumn.REPORTER_ID), "rep", true));
		}
		else
		{
			count.select(Selects.expression(Dynamic.column(VirtualColumn.POSITION), "pos", true));
		}
		checkInterrupted();
		return count.count(dc);
	}

	private Map<Integer, String> getBaseFileReporterFields(DbControl dc, BioAssaySet bas,
			DynamicPositionQuery query, 
			Map<String, QueryItem> queryMap, List<String> reporterFields, boolean mergeOnReporters)
		throws SQLException
	{
		query.joinReporters(JoinType.INNER);
		for (String field : reporterFields)
		{
			QueryItem qi = queryMap.get(field);
			if (qi == null)
			{
				throw new BaseException("Unknown column: "+field+" (you can add this field using a formula)");
			}
			query.select(Selects.expression(qi.getExpression(dc, bas, mergeOnReporters, AverageMethod.MAX), qi.getQueryAlias()));
		}
		Expression posExpression = null;
		if (mergeOnReporters)
		{
			posExpression = Dynamic.column(VirtualTable.POSITION, VirtualColumn.REPORTER_ID);
			query.group(posExpression);
		}
		else
		{
			posExpression = Dynamic.column(VirtualColumn.POSITION);
		}
		query.select(Selects.expression(posExpression, "_position"));

		Map<Integer, String> reporterLine = new HashMap<Integer, String>();
		DynamicResultIterator it = query.iterate(dc);
		List<Integer> reporterIndex = new ArrayList<Integer>(reporterFields.size());
		for (String column : reporterFields)
		{
			reporterIndex.add(it.getIndex(queryMap.get(column).getQueryAlias()));
		}
		int posIndex = it.getIndex("_position");
		
		while (it.hasNext())
		{
			SqlResult r = it.next();
			Integer currentPos = r.getInt(posIndex);
			
			StringBuilder line = new StringBuilder();
			for (int i = 0; i < reporterIndex.size(); ++i)
			{
				String field = r.getString(reporterIndex.get(i));
				line.append((field == null ? "" : field) + "\t");
			}
			reporterLine.put(currentPos, line.toString());
		}
		return reporterLine;
	}

	
	private List<QueryItem> getBaseFileExtraQI(DbControl dc, BioAssaySet bas)
	{
		List<QueryItem> items = new ArrayList<QueryItem>();
		ItemQuery<ExtraValue> query = bas.getExtraValues();
		query.order(Orders.asc(Hql.property("extraValueType.name")));
		for (ExtraValue ev : query.list(dc))
		{
			ExtraValueType evt = ev.getExtraValueType();
			items.add(new QueryItem("xtra."+evt.getExternalId(), "[Extra] "+evt.getName(), 
				"_xc_"+evt.getExternalId(), "xtra('" + evt.getExternalId() + "')", evt.getAverageMethod()));
		}
		Collections.sort(items, new QueryItemTitleComparator());
		return items;
	}
	
	
	private List<QueryItem> getBaseFileFormulaQI(DbControl dc, BioAssaySet bas)
	{
		List<QueryItem> items = new ArrayList<QueryItem>();
		ItemQuery<Formula> queryFormula = Formula.getQuery(Formula.Type.COLUMN_EXPRESSION, 
			bas.getRawDataType(), bas.getIntensityTransform());
		queryFormula.include(Include.MINE, Include.SHARED, Include.IN_PROJECT, Include.OTHERS);
		queryFormula.order(Orders.asc(Hql.property("name")));
		for (Formula f : queryFormula.list(dc))
		{
			items.add(new QueryItem("frm." + f.getId(), "[Formula] "+f.getName(), f.getName(), f.getFormula(0), f.getAverageMethod()));
		}
		Collections.sort(items, new QueryItemTitleComparator());
		return items;
	}
	
	
	private Map<String, QueryItem> getBaseFileQIMap(DbControl dc, BioAssaySet bas)
	{
		Map<String, QueryItem> items = new HashMap<String, QueryItem>();
		for (QueryItem qi : getBaseFileExtraQI(dc, bas)) items.put(qi.getName(), qi);
		for (QueryItem qi : getBaseFileFormulaQI(dc, bas)) items.put(qi.getName(), qi);
		for (QueryItem qi : getBaseFileRawQI(bas)) items.put(qi.getName(), qi);
		for (QueryItem qi : getBaseFileReporterQI()) items.put(qi.getName(), qi);
		for (QueryItem qi : getBaseFileSpotQI(bas)) items.put(qi.getName(), qi);
		return items;
	}
	
	
	private List<QueryItem> getBaseFileRawQI(BioAssaySet bas)
	{
		List<QueryItem> items = new ArrayList<QueryItem>();
		items.add(new QueryItem("raw.position", "[Raw] Position", "rawPosition", "raw('position')", Formula.AverageMethod.NONE));
		items.add(new QueryItem("raw.block", "[Raw] Array block", "block", "raw('block')", Formula.AverageMethod.NONE));
		items.add(new QueryItem("raw.metaGridX", "[Raw] Meta grid X", "metaGridX", "raw('metaGridX')", Formula.AverageMethod.NONE));
		items.add(new QueryItem("raw.metaGridY", "[Raw] Meta grid Y", "metaGridY", "raw('metaGridY')", Formula.AverageMethod.NONE));
		items.add(new QueryItem("raw.row", "[Raw] Row in block", "row", "raw('row')", Formula.AverageMethod.NONE));
		items.add(new QueryItem("raw.col", "[Raw] Column in block", "col", "raw('column')", Formula.AverageMethod.NONE));
		items.add(new QueryItem("raw.x", "[Raw] X coordinate", "x", "raw('x')", Formula.AverageMethod.NONE));
		items.add(new QueryItem("raw.y", "[Raw] Y coordinate", "y", "raw('y')", Formula.AverageMethod.NONE));
		for (RawDataProperty rdp : bas.getRawDataType().getProperties())
		{
			items.add(new QueryItem("raw." + rdp.getName(), "[Raw] "+rdp.getTitle(), rdp.getName(), "raw('"+rdp.getName()+"')", rdp.getAverageMethod()));
		}
		Collections.sort(items, new QueryItemTitleComparator());
		return items;
	}
	
	
	private List<QueryItem> getBaseFileReporterQI()
	{
		List<QueryItem> items = new ArrayList<QueryItem>();
		items.add(new QueryItem("rep.id", "Reporter (internal id)", "reporter", "rep('id')", null));
		items.add(new QueryItem("rep.position", "Position", "position", "pos()", Formula.AverageMethod.NONE));
		items.add(new QueryItem("rep.name", "Reporter Name", "name", "rep('name')", Formula.AverageMethod.MAX));
		items.add(new QueryItem("rep.externalId", "External ID", "externalId", "rep('externalId')", Formula.AverageMethod.MAX));
		items.add(new QueryItem("rep.symbol", "Gene symbol", "symbol", "rep('symbol')", Formula.AverageMethod.MAX));
		items.add(new QueryItem("rep.description", "Description", "description", "rep('description')", Formula.AverageMethod.MAX));
		items.add(new QueryItem("rep.lastUpdate", "Last updated", "lastUpdate", "rep('lastUpdate')", Formula.AverageMethod.MAX));
		for (ExtendedProperty ep : ExtendedProperties.getProperties(ReporterData.class.toString()))
		{
			items.add(new QueryItem("rep." + ep.getName(), ep.getTitle(), ep.getName(), "rep('"+ep.getName()+"')", Formula.AverageMethod.MAX));
		}
		Collections.sort(items, new QueryItemTitleComparator());
		return items;
	}	
	
	
	private List<QueryItem> getBaseFileSpotQI(BioAssaySet bas)
	{
		List<QueryItem> items = new ArrayList<QueryItem>();
		int channels = bas.getRawDataType().getChannels();
		IntensityTransform transform = bas.getIntensityTransform();
		for (int channel = 1; channel <= channels; ++channel)
		{
			items.add(new QueryItem("intensity"+channel, 
					"Int ch"+channel, "intensity" + channel, 
					"rawCh("+channel+")", transform.getAverageMethod()));
		}
		Collections.sort(items, new QueryItemTitleComparator());
		return items;
	}

	
	private RequestInformation getConfigureBasefileParameters(GuiContext context)
	{
		if (configureJobBasefile == null)
		{
			DbControl dc = null;
			try
			{
				dc = sc.newDbControl();
				List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
				geneAveragesParameter = new PluginParameter<Boolean>
					(	"geneAverages",
						"Average on reporters",
						"Average on reporters within each assay.",
						new BooleanParameterType(false, true));
							
				Enumeration<String, String> fields = new Enumeration<String, String>();
				BioAssaySet bas = (BioAssaySet) job.getValue(bioAssaySetParameter.getName());
				for (QueryItem qi : getBaseFileSpotQI(bas))
				{
					fields.add(qi.getName(), qi.getTitle());
				}
				for (QueryItem qi : getBaseFileFormulaQI(dc, bas))
				{
					fields.add(qi.getName(), qi.getTitle());
				}
				for (QueryItem qi : getBaseFileExtraQI(dc, bas))
				{
					fields.add(qi.getName(), qi.getTitle());
				}
				for (QueryItem qi : getBaseFileRawQI(bas))
				{
					fields.add(qi.getName(), qi.getTitle());
				}
				//Limsfields???
				usedFieldsParameter = new PluginParameter<String>(
					"usedFields",
					"Fields",
					"The spot information that is exporter. The position is allways exported when avarage on reporter is false.",
					new StringParameterType(null, null, true, 0, 0, 0, fields));

				Enumeration<String, String> columns = new Enumeration<String, String>();
				for (QueryItem qi : getBaseFileReporterQI())
				{
					columns.add(qi.getName(), qi.getTitle());
				}
				usedColumnsParameter = new PluginParameter<String>(
					"usedColumns",
					"Reporter",
					"The reporter information that is exported. An internal reporter id is allways exported.",
					new StringParameterType(null, null, false, 0, 0, 0, columns));
				
				parameters.add(geneAveragesParameter);
				parameters.add(usedFieldsParameter);
				parameters.add(usedColumnsParameter);
	
				configureJobBasefile = new RequestInformation
					( COMMAND_CONFIGURE_BASEFILE,
						"Configure BASEfile export.",
						"Here you configure the parameter for the BASEfile format.",
						parameters );
			}
			finally
			{
				if (dc != null)
				{
					dc.close();
				}
			}
		}
		return configureJobBasefile;
	}


	
	private RequestInformation getConfigureJobParameters(GuiContext context, Boolean requireFile)
	{
		if (configureJob == null)
		{
			if (requireFile == null)
			{
				PluginDefinition pd = job.getPluginDefinition();
				requireFile = pd == null ? false : !pd.getAllowImmediateExecution();
			}
			String defaultPath = null;
			if (requireFile)
			{
				int currentBioAssaySetId = sc.getCurrentContext(Item.BIOASSAYSET).getId();
				if (currentBioAssaySetId != 0)
				{
					DbControl dc = sc.newDbControl();
					try
					{
						BioAssaySet base = BioAssaySet.getById(dc, currentBioAssaySetId);
						defaultPath = "~/" + Path.makeSafeFilename(base.getName(), "").toLowerCase() + "txt";
					}
					finally
					{
						if (dc != null)dc.close();
					}
				}
			}
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
			bioAssaySetParameter = new PluginParameter<BioAssaySet>
				( "bioAssaySet",
					"Bioassay set",
					"The bioassay set to export",
					new ItemParameterType<BioAssaySet>(BioAssaySet.class, null, true, 1, null) );
			
			Enumeration<String, String> formats = new Enumeration<String, String>();
			formats.add(FORMAT_BASEFILE_SERIAL, "Serial BASEfile");
			formats.add(FORMAT_BASEFILE_MATRIX, "Matrix BASEfile");
			formats.add(FORMAT_MEV_TDMS, "MeV TDMS");
			/**
				 Plain matrix export is currently not supported from the web
				 interface. Simply move the code line below outside this
				 comment block to enable plain matrix export in the web
				 interface. If the export is enabled the About information
				 should be updated to reflect the change.

				 formats.add(FORMAT_PLAIN_MATRIX, "Plain Matrix");
			*/
			formatParameter = new PluginParameter<String>
				( "fileformat",
					"Fileformat",
					"Please specify what format to export.",
					new StringParameterType(255, null, true, 1, 0, 0, formats) );

			PluginParameter<Boolean> attachToBasParameter = 
				new PluginParameter<Boolean>(
				"attachToBioAssaySet",
				"Attach file to bioassay set",
				"Deselect this option if you don't want the generated file to " +
				"become attached to the current bioassay set. This option is ignored " +
				"if the file isn't saved in the BASE file system.",
				new BooleanParameterType(true, false)
				);
			
			parameters.add(bioAssaySetParameter);
			parameters.add(formatParameter);
			parameters.add(getSaveAsParameter(null, null, defaultPath, requireFile));
			parameters.add(getOverwriteParameter(null, null));
			parameters.add(attachToBasParameter);

			configureJob = new RequestInformation
				( Request.COMMAND_CONFIGURE_JOB,
					"Select a place to save the exported data.",
					"Here you select where to save exported data and " +
					"how to treat existing files.",
					parameters );
		}
		return configureJob;
	}
	
	
	public List<AnnotationType> getExperimentalFactors(DbControl dc, BioAssaySet bas)
	{
		ItemQuery<AnnotationType> query = bas.getExperiment().getExperimentalFactors();
		query.include(Include.MINE, Include.SHARED, Include.IN_PROJECT, Include.OTHERS);
		query.order(Orders.asc(Hql.property("name")));
		return query.list(dc);
	}

	
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}


	@Override
	public Plugin.MainType getMainType()
	{
		return Plugin.MainType.EXPORT;
	}


	/**
	 	Request create and write access to File:s and Directory:s and read access to
	 	Experiment:s and User:s.
	*/
	@Override
	public Collection<Permissions> getPermissions()
	{
		if (permissions.size() == 0)
		{
			permissions.add(new Permissions(Item.ANNOTATIONTYPE, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.DIRECTORY, null, EnumSet.of(Permission.CREATE, Permission.WRITE)));
			permissions.add(new Permissions(Item.EXPERIMENT, null, EnumSet.of(Permission.USE)));
			permissions.add(new Permissions(Item.FILE, null, EnumSet.of(Permission.CREATE, Permission.WRITE)));
			permissions.add(new Permissions(Item.FORMULA, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.RAWBIOASSAY, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.USER, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.EXTRAVALUETYPE, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.DATAFILETYPE, null, EnumSet.of(Permission.READ)));
		}
		return permissions;
	}

	
	public RequestInformation getRequestInformation(GuiContext context, String command)
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			requestInformation = getConfigureJobParameters(context, null);
		}
		else if (command.equals(COMMAND_CONFIGURE_BASEFILE))
		{
			requestInformation = getConfigureBasefileParameters(context);
		}
		return requestInformation;
	}

	
	public String isInContext(GuiContext context, Object item)
	{
		String message = null;	
		if (item == null)
		{
			message = "The object is null";
		}
		else if (!(item instanceof BioAssaySet))
		{
			message = "The object is not a BioAssaySet: " + item;
		}
		else
		{
			BioAssaySet bas = (BioAssaySet)item;
			if (bas.getNumSpots() <= 0)
			{
				message = "This plug-in requires spot data in the database";
			}
		}
		return message;
	}

	
	@Override
	public boolean requiresConfiguration()
	{
		return false;
	}

	private DbControl dc;
	private BioAssaySet bas;
	@Override
	protected void begin(DbControl dc)
	{
		this.dc = dc;
		bas = (BioAssaySet)job.getValue("bioAssaySet");
	}	
	
	@Override
	@SuppressWarnings("unchecked")
	protected void performExport(ExportOutputStream out, ProgressReporter progress)
		throws IOException
	{
		if (signalHandler != null) signalHandler.setWorkerThread(null);
		try
		{
			BioAssaySet bas = (BioAssaySet)job.getValue("bioAssaySet");
			bas = BioAssaySet.getById(dc, bas.getId());
			boolean attachToBioAssaySet = Boolean.TRUE.equals(job.getValue("attachToBioAssaySet"));
			out.setMimeType("text/plain");
			String filename = Path.makeSafeFilename(bas.getName() + ".export.", "").toLowerCase() + "txt";
			out.setFilename(filename);
			out.setCharacterSet(Config.getCharset());
			File exportFile = out.getFile();
			String dataFileTypeId = null;
			
			if (progress != null) progress.display(0, "Exporting ...");
	
			// Add other export formats here
			String format = (String) job.getValue("fileformat");
			if (FORMAT_MEV_TDMS.equals(format))
			{
				exportMeV(bas, out, progress);
				dataFileTypeId = DataFileType.MEV_TDMS;
			}
			else if (FORMAT_BASEFILES.contains(format)) 
			{
				List<String> reporterFields = (List<String>)job.getValues("usedColumns");
				List<String> spotFields = (List<String>)job.getValues("usedFields");
				
				if (reporterFields == null) reporterFields = Collections.emptyList();
				if (spotFields == null) spotFields = Collections.emptyList();
				
				boolean geneAverage = (Boolean) job.getValue("geneAverages");
				
				if (FORMAT_BASEFILE_SERIAL.equals(format))
				{
					exportBaseFileSerial(bas, out, Collections.EMPTY_MAP, reporterFields, spotFields, geneAverage, progress);
					dataFileTypeId = DataFileType.BASEFILE_SERIAL;
				}
				else 
				{
					exportBaseFileMatrix(bas, out, Collections.EMPTY_MAP, reporterFields, spotFields, geneAverage, progress);
					dataFileTypeId = DataFileType.BASEFILE_MATRIX;
				}
			}
			else if (FORMAT_PLAIN_MATRIX.equals(format))
			{
				exportPlainMatrix(bas, out, progress);
			}
			else
			{
				throw new BaseException("Unknown format. "+format);
			}
			
			if (attachToBioAssaySet && exportFile != null && dataFileTypeId != null)
			{
				FileSetMember file = FileStoreUtil.setDataFile(dc, bas, dataFileTypeId, exportFile);
				file.setValid(true, null);
				bas.setNumFileSpots(bas.getNumSpots());
				bas.setNumFileReporters(bas.getNumReporters());
				for (BioAssay ba : bas.getBioAssays().list(dc))
				{
					ba.setNumFileSpots(ba.getNumSpots());
				}
			}
		}
		catch (SQLException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	@Override
	protected void end(boolean success)
	{
		this.dc = null;
	}
	
	@Override
	protected String getSuccessMessage()
	{
		return "Bioassay set '" + bas.getName() + "' exported successfully";
	}
	
	@Override
	public boolean supportsConfigurations()
	{
		return false;
	}

	
	private static final About about =
		new AboutImpl
		(
			"Bioassay set file exporter (deprecated)",
			"This plug-in has been deprecated and has been replaced with the " +
			"BASEfile exporter for serial and matrix BASEfile formats and with" +
			"the MeV extension for MeV TDMS format " +
			"(see http://baseplugins.thep.lu.se/wiki/net.sf.basedb.mev)\n",
			Version.getMajor() + "." + Version.getMinor() + "." + Version.getMaintenance(),
			"2006, Base 2 development team",
			null,
			null,
			"http://base.thep.lu.se"
		);
	private RequestInformation configureJob;
	private RequestInformation configureJobBasefile;
	private static final Set<GuiContext> guiContexts =
		Collections.singleton(new GuiContext(Item.BIOASSAYSET, GuiContext.Type.ITEM));
 	private static final Set<Permissions> permissions =
		new HashSet<Permissions>();
 	
 	//General parameters
 	private PluginParameter<BioAssaySet> bioAssaySetParameter;
 	private PluginParameter<String> formatParameter;
 	
 	//BaseFile parameters
 	private PluginParameter<Boolean> geneAveragesParameter;
 	private PluginParameter<String> usedFieldsParameter;
 	private PluginParameter<String> usedColumnsParameter;
 	
 	// Constants
 	private static final String COMMAND_CONFIGURE_BASEFILE = "configure_basefile";
 	private static final String FORMAT_BASEFILE_SERIAL = "basefile.serial";
 	private static final String FORMAT_BASEFILE_MATRIX = "basefile.matrix";
 	private static final String FORMAT_MEV_TDMS = "mev.tdms";
 	private static final String FORMAT_PLAIN_MATRIX = "plain.matrix";
 	
 	private static final Set<String> FORMAT_BASEFILES = new HashSet<String>(
 			Arrays.asList(FORMAT_BASEFILE_SERIAL, FORMAT_BASEFILE_MATRIX));

	public class QueryItem
	{
		private String name;
		
		private String title;
		
		private String exportName;
		
		private String jep;

		private String queryAlias;
		
		private Formula.AverageMethod avg;
		
		QueryItem(String name, String title, String exportName, String jep, Formula.AverageMethod avg)
		{
			this.name = name;
			this.title = title;
			this.exportName = exportName;
			this.jep = jep;
			this.queryAlias = "a" + Math.abs(System.identityHashCode(name));
			this.avg = avg;
		}

		public final Expression getExpression(DbControl dc, BioAssaySet bas, boolean average)
		{
			return getExpression(dc, bas, average, avg);
		}
		
		public final Expression getExpression(DbControl dc, BioAssaySet bas, boolean average, AverageMethod forceAvgMethod)
		{
			Expression e = BioAssaySetUtil.createJepExpression(dc, bas, jep, true);
			if (average && forceAvgMethod != null)
			{
				e = forceAvgMethod.getAverageExpression(e);
				if (e == null)
				{
					e = Expressions.string("");
				}
			}
			return e;
		}

		public final String getName()
		{
			return name;
		}

		public final String getTitle()
		{
			return title;
		}
		
		public final String getExportName()
		{
			return exportName;
		}

		public final String getQueryAlias()
		{
			return queryAlias;
		}
	}
	
	
	private class QueryItemTitleComparator
		implements Comparator<QueryItem>
	{
		public int compare(QueryItem qi1, QueryItem qi2)
		{
			return qi1.getTitle().compareTo(qi2.getTitle());
		}
	}
}
