/**
	$Id: CdfFileReporterImporter.java 4513 2008-09-11 20:06:36Z jari $

	Copyright (C) 2007 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins;

import java.io.BufferedOutputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import affymetrix.fusion.cdf.FusionCDFData;

import net.sf.basedb.core.Affymetrix;
import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DataFileType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.File;
import net.sf.basedb.core.FileParameterType;
import net.sf.basedb.core.FileStoreUtil;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemContext;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.Location;
import net.sf.basedb.core.Operator;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.Platform;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.PropertyFilter;
import net.sf.basedb.core.ReporterBatcher;
import net.sf.basedb.core.ReporterType;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.Version;
import net.sf.basedb.core.filehandler.CdfFileHandler;
import net.sf.basedb.core.plugin.About;
import net.sf.basedb.core.plugin.AboutImpl;
import net.sf.basedb.core.plugin.AbstractPlugin;
import net.sf.basedb.core.plugin.AutoDetectingImporter;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.Permissions;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.signal.SignalHandler;
import net.sf.basedb.core.signal.SignalTarget;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.FileUtil;

/**
	This plug-in can import probesets from Affymetrix CDF files as reporters.
	The plug-in uses the {@link Affymetrix} class for the actual work. 
	
	@base.developer
	In order to support the <code>AutoDetectingImporter</code> interface
	we needed to duplicate the file format checks from the Fusion SDK
	in this class. See {@link #isImportable(InputStream)}.

	@author nicklas
	@version 2.4
	@base.modified $Date: 2008-09-11 22:06:36 +0200 (Thu, 11 Sep 2008) $
*/
public class CdfFileReporterImporter
	extends AbstractPlugin
	implements InteractivePlugin, AutoDetectingImporter, SignalTarget
{

	/**
	   Plugin description
	*/
	private static final About about = 
		new AboutImpl
		(
			"Affymetrix CDF probeset importer",
			"This plug-in is used to import probesets from an Affymetrix CDF file as reporters. " +
			"It can be used in import mode from the reporter list view and from the array design view " +
			"and in verification mode from the array design view. " +
			"The plug-in can only set the name and ID of the reporters, since the CDF file doesn't " +
			"contains any annotation information. Probesets already in BASE will not " +
			"be affected by the import",
			Version.getMajor() + "." + Version.getMinor() + "." + Version.getMaintenance(),
			"2007, Base 2 development team",
			null,
			null,
			"http://base.thep.lu.se"
		);
	
	private static final Set<GuiContext> guiContexts = 
		Collections.unmodifiableSet(new HashSet<GuiContext>(
			Arrays.asList(
				new GuiContext(Item.REPORTER, GuiContext.Type.LIST),
				new GuiContext(Item.ARRAYDESIGN, GuiContext.Type.ITEM)
			)
		));
	
	private static final Set<Permissions> permissions = new HashSet<Permissions>();

	private RequestInformation configureJob;
	
	private ThreadSignalHandler signalHandler;
	
	/**
	   Constructor should be empty. 
	*/
	public CdfFileReporterImporter()
	{}
	
	/*
		From the Plugin interface
		-------------------------------------------
	*/
	public About getAbout()
	{
		return about;
	}
	public boolean supportsConfigurations()
	{
		return false;
	}
	public boolean requiresConfiguration()
	{
		return false;
	}
	public MainType getMainType()
	{
		return MainType.IMPORT;
	}

	/**
	 	Request create and write access to Reporter:s
	 	and read access to File:s ArrayDesign:s, ReporterType:s,
	 	Platform:s and DataFileType:s
	*/
	public Collection<Permissions> getPermissions()
	{
		if (permissions.size() == 0)
		{
			permissions.add(new Permissions(Item.REPORTER, null, EnumSet.of(Permission.CREATE, Permission.WRITE)));
			permissions.add(new Permissions(Item.REPORTERTYPE, EnumSet.of(Permission.USE), null));
			permissions.add(new Permissions(Item.FILE, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.ARRAYDESIGN, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.PLATFORM, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.DATAFILETYPE, null, EnumSet.of(Permission.READ)));
		}
		return permissions;
	}

	public void run(Request request, Response response, ProgressReporter progress)
	{
		if (signalHandler != null) signalHandler.setWorkerThread(null);
		DbControl dc = sc.newDbControl();
		try
		{
			// Get the cdf file
			File cdfFile = (File)job.getValue("file");
			cdfFile = File.getById(dc, cdfFile.getId());
			checkInterrupted();
			FusionCDFData cdf = new CdfFileHandler().loadCdfFile(cdfFile);
	
			// Import data from it
			int numCreated = importFromCdf(dc, cdf, progress);
			int numTotal = cdf.getHeader().getNumProbeSets();
			int numExisted = numTotal - numCreated;
			
			// Commit and return
			dc.commit();
			response.setDone("Created " + numCreated + " reporters; " + numExisted + " skipped because the already existed");
		}
		catch (Throwable t)
		{
			response.setError(t.getMessage(), Arrays.asList(t));
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	// -------------------------------------------
	/*
		From the InteractivePlugin interface
		-------------------------------------------
	*/
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	/**
		If used from an array design context, verify that the array design
		is an affymetrix chip.
	*/
	public String isInContext(GuiContext context, Object item)
	{
		String message = null;
		if (context.getItem() == Item.ARRAYDESIGN)
		{
			if (item == null)
			{
				message = "The object is null";
			}
			else if (!(item instanceof ArrayDesign))
			{
				message = "The object is not an array design: " + item;
			}
			ArrayDesign design = (ArrayDesign)item;
			if (!design.isPlatform(Platform.AFFYMETRIX))
			{
				message = "The array design is not an Affymetrix chip";
			}
			else
			{
				if (!FileStoreUtil.hasDataFile(design.getDbControl(), design, DataFileType.AFFYMETRIX_CDF, true))
				{
					throw new InvalidDataException("The array design doesn't have a valid CDF file");
				}
			}
		}
		return message;
	}

	public RequestInformation getRequestInformation(GuiContext context, String command) 
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			requestInformation = getConfigureJobParameters(context);
			ItemContext fileContext = sc.getCurrentContext(Item.FILE);
			fileContext.setPropertyFilter(new PropertyFilter("name", Operator.LIKE, 
					"%.cdf", Type.STRING));
			fileContext.setPropertyFilter(new PropertyFilter("location", Operator.EQ, 
					Integer.toString(Location.PRIMARY.getValue()), Type.INT));
		}
		return requestInformation;
	}

	public void configure(GuiContext context, Request request, Response response)
	{
		String command = request.getCommand();
		try
		{
			if (command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				RequestInformation ri = getConfigureJobParameters(context);
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found in the request", errors);
					return;
				}
				
				// Verify that the file is a CDF file
				File cdfFile = (File)request.getParameterValue("file");
				FusionCDFData cdf = new CdfFileHandler().loadCdfFile(cdfFile);
				
				storeValue(job, request, ri.getParameter("file"));
				storeValue(job, request, ri.getParameter("mode"));
				storeValue(job, request, ri.getParameter("reporterType"));
				response.setDone("Job configuration complete", Job.ExecutionTime.SHORT);
				// TODO - maybe check file size to make a better estimate
			}
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
	}
	// -------------------------------------------
	/*
		From the AutoDetctingImporter interface
		-------------------------------------------
	*/
	public final boolean isImportable(InputStream in)
		throws BaseException
	{
		try
		{
			// We don't want to download the entire file, so
			// we read the first 8 bytes and try to mimic the
			// same checks made by the Fusion SDK
			byte[] first8bytes = new byte[8];
			in.read(first8bytes);
			
			// Check if XDA format
			ByteBuffer buffer = ByteBuffer.wrap(first8bytes);
			buffer.order(ByteOrder.LITTLE_ENDIAN);
			int magicNumber = buffer.getInt();
			int version = buffer.getInt();
			if (magicNumber == 67 || version == 1)
			{
				// This is an CDF file in XDA format
				return true;
			}
			else
			{
				String start = new String(first8bytes);
				if (start.startsWith("[CDF]"))
				{
					// This is a CDF File in text format
					return true;
				}
			}
			return false;
		}
		catch (IOException ex)
		{
			throw new BaseException(ex);
		}
	}

	public void doImport(InputStream in, ProgressReporter progress)
		throws BaseException
	{
		DbControl dc = null;
		try
		{
			// Download input stream to temp file
			java.io.File tempFile = java.io.File.createTempFile("affymetrix", "cdf");
			tempFile.deleteOnExit();			
			OutputStream out = new BufferedOutputStream(new FileOutputStream(tempFile));
			FileUtil.copy(in, out);
			out.flush();
			out.close();
			
			FusionCDFData cdf = new FusionCDFData();
			cdf.setFileName(tempFile.getAbsolutePath());
			
			dc = sc.newDbControl();
			importFromCdf(dc, cdf, progress);
			tempFile.delete();
			dc.commit();
		}
		catch (IOException ex)
		{
			throw new BaseException(ex);
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	// -------------------------------------------
	
	/*
		From the SignalTarget interface
		-------------------------------------------
	*/
	public SignalHandler getSignalHandler()
	{
		signalHandler = new ThreadSignalHandler();
		return signalHandler;
	}
	// -------------------------------------------

	private RequestInformation getConfigureJobParameters(GuiContext context)
	{
		DbControl dc = null;
		try
		{
			if (configureJob == null)
			{
				// RequestInformation for CONFIGURE_JOB
				List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
	
				dc = sc.newDbControl();
				ArrayDesign design = null;
				File cdfFile = null;
				List<File> presets = null;
				if (context.getItem() == Item.ARRAYDESIGN)
				{
					int currentArrayDesignId = sc.getCurrentContext(Item.ARRAYDESIGN).getId();
					design = ArrayDesign.getById(dc, currentArrayDesignId);
					cdfFile = FileStoreUtil.getDataFile(dc, design, DataFileType.AFFYMETRIX_CDF);
					presets = Arrays.asList(new File[] { cdfFile } );
				}
				
				// The CDF file to import from
				// The FileParameterType can't have a list of presets to select from
				PluginParameter<File> fileParameter = new PluginParameter<File>(
					"file",
					"CDF file",
					"The CDF file to import probesets IDs from",
					presets != null ?
						new ItemParameterType<File>(File.class, cdfFile, true, 1, presets) :
						new FileParameterType(cdfFile, true, 1)
				);
				parameters.add(fileParameter);
				
				if (context.getItem() == Item.ARRAYDESIGN)
				{
					boolean hasCreatePermission = (sc.hasPermission(Permission.CREATE, Item.REPORTER));
					List<String> modes = new ArrayList<String>();
					if (hasCreatePermission) modes.add("import");
					modes.add("verify");
					StringParameterType modeType = 
						new StringParameterType(255, hasCreatePermission ? "import" : "verify", true, 1, 0, 0, modes);					
					PluginParameter<String> modeParameter = new PluginParameter<String>(
						"mode",
						"Mode",
						"Select the mode the plug-in should use:\n\n"+
						"import = Import missing reporters"+
						(!hasCreatePermission ? " (you don't have this permission)" : "")+
						"\nverify = Verify that the reporters already exists",
						modeType
					);
					parameters.add(modeParameter);
				}
				
				// Load reporter types and initialise the reporterTypeParameter
				ItemQuery<ReporterType> query = ReporterType.getQuery();
				query.order(Orders.asc(Hql.property("name")));
				List<ReporterType> reporterTypes = new ArrayList<ReporterType>(query.list(dc));
				
				if (reporterTypes.size() > 0)
				{
					ItemParameterType<ReporterType> reporterTypeType = 
						new ItemParameterType<ReporterType>(ReporterType.class, null, false, 1, reporterTypes);
					PluginParameter<ReporterType> reporterTypeParameter = new PluginParameter<ReporterType>(
						"reporterType",
						"Reporter type",
						"The reporter type assigned to new reporters. Already existing reporters " +
						"are not modified.",
						reporterTypeType);
					parameters.add(reporterTypeParameter);
				}
				
				configureJob = new RequestInformation
				(
					Request.COMMAND_CONFIGURE_JOB,
					"Select CDF file",
					"Select a CDF file to import reporters from.",
					parameters
				);
			}
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return configureJob;
	}	
	
	private int importFromCdf(DbControl dc, FusionCDFData cdf, ProgressReporter progress)
	{
		checkInterrupted();

		ReporterBatcher batcher = ReporterBatcher.getNew(dc);
		if (job.getValue("file") != null)
		{
			batcher.setUpdateSource("File: " + ((File)job.getValue("file")).getName());
		}
		ReporterType type = (ReporterType)job.getValue("reporterType");
		String mode = (String)job.getValue("mode");
		int numCreated = Affymetrix.checkIfReportersExists(cdf, batcher, !"verify".equals(mode), 
			type, progress);
		return numCreated;
	}

}
