/*
	$Id: TestGroup.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
import net.sf.basedb.core.*;
import java.util.Date;
import java.util.EnumSet;

public class TestGroup
{
	static boolean ok = true;

	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		write("++Testing groups");
		write_header();
		// Standard tests: create, load, list
		int id = test_create(true);
		int id2 = test_create(false);
		test_load(id);
		test_list(-1);

		// Extra tests: add, list and remove users
		test_listusers(id, 0);
		test_adduser(id, SystemItems.getId(User.ROOT));
		test_listusers(id, 1);
		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
		test_removeuser(id, SystemItems.getId(User.ROOT));
		test_listusers(id, 0);

		// Extra tests: add, list and remove groups
		test_listgroups(id, 0);
		test_addgroup(id, id2);
		test_listgroups(id, 1);
		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
		test_removegroup(id, id2);
		test_listgroups(id, 0);

		// Extra tests: add, list and remove projects
		int projectId = TestProject.test_create(true);
		test_listprojects(id, 0);
		TestProject.test_addgroup(projectId, id, EnumSet.of(Permission.WRITE));
		test_listprojects(id, 1);
		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
		TestProject.test_removegroup(projectId, id);
		test_listprojects(id, 0);
		TestProject.test_delete(projectId);

		// Standard test: Delete
		test_delete(id2);
		test_delete(id);
		write("++Testing groups "+(ok ? "OK" : "Failed")+"\n");
		return ok;
	}

	static int test_create(boolean setAll)
	{
		if (!TestUtil.hasPermission(Permission.CREATE, Item.GROUP)) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Group g = Group.getNew(dc);
			if (setAll)
			{
				g.setName("Test group");
				g.setDescription("Added at "+new Date());
			}
			dc.saveItem(g);
			dc.commit();
			id = g.getId();
			write_item(0, g);
			write("--Create group OK");
		}
		catch (Throwable ex)
		{
			write("--Create group FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	static void test_load(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Group g = Group.getById(dc, id);
			write_item(0, g);
			write("--Load group OK");
		}
		catch (Throwable ex)
		{
			write("--Load group FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_list(int expectedResults)
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemResultList<Group> l = Group.getQuery().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List groups OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List groups FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_delete(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Group g = Group.getById(dc, id);
			dc.deleteItem(g);
			dc.commit();
			write("--Delete group OK");
		}
		catch (Throwable ex)
		{
			write("--Delete group FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void write_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tName      \tDescription");
			write("-- \t-- \t--------- \t-----------");
		}
	}
	static void write_item(int i, Group g)
		throws BaseException
	{
		if (!TestUtil.getSilent()) System.out.println(i+":\t"+g.getId()+"\t"+g.getName()+"\t"+g.getDescription());
	}
	static void write_item(int i, User u)
		throws BaseException
	{
		if (!TestUtil.getSilent()) System.out.println(i+":\t"+u.getId()+"\t"+u.getName()+"\t"+u.getDescription());
	}
	static void write_item(int i, Project p)
		throws BaseException
	{
		if (!TestUtil.getSilent()) System.out.println(i+":\t"+p.getId()+"\t"+p.getName()+"\t"+p.getDescription());
	}
	static void write(String message)
	{
		System.out.println(message);
	}

	static void test_adduser(int id, int userId)
	{
		if (id == 0 || userId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Group g = Group.getById(dc, id);
			User u = User.getById(dc, userId);
			g.addUser(u);
			dc.commit();
			write("--Add user to group OK");
		}
		catch (Throwable ex)
		{
			write("--Add user to group FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_listusers(int id, int expectedResults)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Group g = Group.getById(dc, id);
			ItemResultList<User> l = g.getUsers().list(dc);
			for (int i=0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List members of group OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List members of group FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_removeuser(int id, int userId)
	{
		if (id == 0 || userId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Group g = Group.getById(dc, id);
			User u = User.getById(dc, userId);
			g.removeUser(u);
			dc.commit();
			write("--Remove user from group OK");
		}
		catch (Throwable ex)
		{
			write("--Remove user from group FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_addgroup(int id, int groupId)
	{
		if (id == 0 || groupId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Group g = Group.getById(dc, id);
			Group g2 = Group.getById(dc, groupId);
			g.addGroup(g2);
			dc.commit();
			write("--Add group to group OK");
		}
		catch (Throwable ex)
		{
			write("--Add group to group FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_listgroups(int id, int expectedResults)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Group g = Group.getById(dc, id);
			ItemResultList<Group> l = g.getGroups().list(dc);
			for (int i=0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List groups in group OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List groups in group FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_removegroup(int id, int groupId)
	{
		if (id == 0 || groupId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Group g = Group.getById(dc, id);
			Group g2 = Group.getById(dc, groupId);
			g.removeGroup(g2);
			dc.commit();
			write("--Remove group from group OK");
		}
		catch (Throwable ex)
		{
			write("--Remove group from group FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static void test_listprojects(int id, int expectedResults)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Group g = Group.getById(dc, id);
			ItemResultList<Project> l = g.getProjects().list(dc);
			for (int i=0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List projects for group OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List projects for group FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
}

