/*
	$Id: TestPluginConfiguration.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ParameterInfo;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PluginConfiguration;
import net.sf.basedb.core.PluginDefinition;
import net.sf.basedb.core.PluginConfigurationRequest;
import net.sf.basedb.core.PluginResponse;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultList;
import net.sf.basedb.core.plugin.Response;

import java.util.List;
import java.util.Arrays;

public class TestPluginConfiguration
{

	static boolean ok = true;
	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		write("++Testing plugin configuration");
		write_header();
		// Standard tests: create, load, list
		int plugindefinitionId = TestPluginDefinition.test_create("NullPlugin", null);
		int id = test_create(plugindefinitionId);
		test_list();
		test_load(id);

		// Extra tests:
		test_configure(id);

		// Standard test: Delete
		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
		test_delete(id);
		TestPluginDefinition.test_delete(plugindefinitionId);
		write("++Testing plugin configuration "+(ok ? "OK" : "Failed")+"\n");
		return ok;
	}

	static int test_create(int plugindefinitionId)
	{
		if (plugindefinitionId == 0 || !TestUtil.hasPermission(Permission.CREATE, Item.PLUGINCONFIGURATION)) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			PluginDefinition pd = PluginDefinition.getById(dc, plugindefinitionId);
			PluginConfiguration pc = PluginConfiguration.getNew(dc, pd);
			pc.setName(pd.getName());
			dc.saveItem(pc);
			dc.commit();
			id = pc.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(pc);
			write_item(0, pc);
			write("--Create plugin configuration OK");
		}
		catch (Throwable ex)
		{
			write("--Create plugin configuration FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	static void test_load(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			PluginConfiguration pc = PluginConfiguration.getById(dc, id);
			write_item(0, pc);
			write_values(pc);
			write("--Load plugin configuration OK");
		}
		catch (Throwable ex)
		{
			write("--Load plugin configuration FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_list()
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemQuery<PluginConfiguration> q = PluginConfiguration.getQuery();
			ItemResultList<PluginConfiguration> l = q.list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			write("--List plugin configurations OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List plugin configurations FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_delete(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			PluginConfiguration pc = PluginConfiguration.getById(dc, id);
			dc.deleteItem(pc);
			dc.commit();
			write("--Delete plugin configuration OK");
		}
		catch (Throwable ex)
		{
			write("--Delete plugin configuration FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	

	static void test_configure(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			PluginConfiguration pc = PluginConfiguration.getById(dc, id);
			
			PluginConfigurationRequest pr = pc.configure();
			pr.getRequestInformation();

			pr.setParameterValue("oneString", "aString");
			pr.setParameterValues("manyStrings", Arrays.asList("A", "C"));
			pr.setParameterValue("selectOneString", "B");
			pr.setParameterValues("selectMultipleStrings", Arrays.asList("One", "Three"));
			PluginResponse response = pr.invoke();
			if (response.getStatus() == Response.Status.DONE)
			{
				response.saveParameters(dc);
				dc.commit();
			}
			else
			{
				throw new BaseException(response.getMessage(), response.getErrorList().get(0));
			}

			dc = TestUtil.getDbControl();
			dc.reattachItem(pc);
			
			List<?> value = pc.getParameterValues("oneString");
			if (value == null || value.size() != 1 || !value.get(0).equals("aString"))
			{
				throw new Exception("Cant find the correct value for plugin parameter 'oneString'");
			}
			value = pc.getParameterValues("manyStrings");
			if (value == null || value.size() != 2 || !value.contains("A") || !value.contains("C"))
			{
				throw new Exception("Cant find the correct value for plugin parameter 'manyStrings'");
			}
			value = pc.getParameterValues("selectOneString");
			if (value == null || value.size() != 1 || !value.get(0).equals("B"))
			{
				throw new Exception("Cant find the correct value for plugin parameter 'oneString'");
			}
			value = pc.getParameterValues("selectMultipleStrings");
			if (value == null || value.size() != 2 || !value.contains("One") || !value.contains("Three"))
			{
				throw new Exception("Cant find the correct value for plugin parameter 'manyStrings'");
			}

			write_item(0, pc);
			write_values(pc);
			write("--Configure plugin configuration OK");
		}
		catch (Throwable ex)
		{
			write("--Configure plugin configuration FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void write_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tName      \tDescription\tDefinition");
			write("-- \t-- \t--------- \t-----------\t----------");
		}
	}
	
	static void write_item(int i, PluginConfiguration pc)
		throws BaseException
	{
		if (!TestUtil.getSilent())
		{
			System.out.println(i+":\t"+pc.getId()+"\t"+pc.getName()+"\t"+pc.getDescription()+
					"\t"+pc.getPluginDefinition());
		}
	}
	
	static void write_values(PluginConfiguration pc)
		throws BaseException
	{
		if (!TestUtil.getSilent())
		{
			System.out.println("Configuration values:");
			for(String name : pc.getParameterNames())
			{
				ParameterInfo pi = pc.getParameterInfo(name);
				System.out.println((pi.getLabel() == null ? name : pi.getLabel()) + "=" + pi.getValues());
			}
		}
		
	}
	
	static void write(String message)
	{
		System.out.println(message);
	}
}

